(*  Title:      HOL/Tools/BNF/bnf_nu_fp.ML
    Author:     Dmitriy Traytel, ETH Zuerich
    Author:     Fabian Meier, ETH Zuerich
    Copyright   2016

Nonuniform datatype construction.
*)

signature BNF_NU_FP =
sig
  val construct_nu_fp: BNF_Util.fp_kind -> mixfix list -> binding list -> binding list ->
    binding list -> binding option list list -> binding list -> (string * sort) list -> typ list ->
    ((typ list * BNF_Def.bnf) * (typ list * BNF_Def.bnf) list list) list ->
    BNF_Comp.absT_info list -> local_theory -> BNF_NU_FP_Util.nu_fp_result * local_theory
end;

structure BNF_NU_FP : BNF_NU_FP =
struct

open Ctr_Sugar
open BNF_Def
open BNF_Util
open BNF_Tactics
open BNF_FP_Util
open BNF_FP_Def_Sugar
open BNF_LFP_Rec_Sugar
open BNF_GFP_Rec_Sugar
open BNF_NU_FP_Util
open BNF_NU_FP_Tactics
open BNF_NU_FP_Def_Sugar

fun target_ctr_sugar_of_fp_sugar lthy fpT ({T, fp_ctr_sugar = {ctr_sugar, ...}, ...} : fp_sugar) =
  let
    val thy = Proof_Context.theory_of lthy;
    val rho = Vartab.fold (cons o apsnd snd) (Sign.typ_match thy (T, fpT) Vartab.empty) [];
    val phi = Morphism.term_morphism "BNF" (Term.subst_TVars rho);
  in
    morph_ctr_sugar phi ctr_sugar : ctr_sugar
  end;

fun wit_apply (arg_I, arg_wit) (fun_I, fun_wit) =
  (union (op =) arg_I fun_I, fun_wit $ arg_wit);

fun mk_ctor_witss decode f m Dss As Ts bnfs ctors lthy =
  let
    val ((ys, ys'), _) = lthy |> mk_Frees' "z" As
    val witss = @{map 2} (fn Ds => fn bnf => mk_wits_of_bnf
      (replicate (nwits_of_bnf bnf) Ds)
      (replicate (nwits_of_bnf bnf) Ts) bnf) Dss bnfs;
    fun close_wit (I, wit) = fold_rev Term.absfree (map (nth ys') I) wit;

    fun gen_arg support js i =
      if i < m then map (pair [i]) (f js ys i)
      else
        let
          val j = i - m;
          val j' = decode j;
        in
          maps (mk_wit support (nth ctors j') (j :: js) j') (nth support j')
        end
    and mk_wit support ctor js i (I, wit) =
      let
        val args = map (gen_arg (nth_map i (remove (op =) (I, wit)) support) js) I;
      in
        (args, [([], wit)])
        |-> fold (map_product wit_apply)
        |> map (apsnd (fn t => ctor $ t))
        |> minimize_wits
      end;
  in
    @{map 3} (fn ctor => fn i => map close_wit o minimize_wits o maps (mk_wit witss ctor [] i))
      ctors (0 upto length ctors - 1) witss
  end;

(*FIXME for Jasmin*)
fun primcorec int opts fixes specs lthy: (term list * thm list * thm list list) * local_theory =
  let
(*val _ = map (writeln o Syntax.string_of_term lthy o Syntax.check_term lthy o snd o #1) specs*)
    val fixes' = map_filter (fn (b, topt, m) => if is_some topt then SOME ((b, the topt), m) else NONE) fixes;
    val decl = fixes' |> Free o apfst Binding.name_of o fst o hd;
    val specs' = map (pair (Binding.empty, []) o snd o #1) specs;
  in
    primcorec_ursive int true opts fixes' specs' (replicate (length specs) NONE) lthy
    |> (fn (goalss, after_qed, lthy) => lthy
      |> after_qed (map (fn [] => [] | _ => error "\"auto\" failed") goalss))
    |> `(fn lthy: local_theory => Spec_Rules.retrieve lthy decl
      |> hd |> snd
      |> (fn (trms, simps) => (trms, [], map single simps)))
  end;

val context_transfer_add = fold (snd oo Thm.proof_attributes [Transfer.transfer_add]);
fun transfer_prover_add_tac ctxt thms =
  Transfer.transfer_prover_tac (context_transfer_add thms ctxt);
fun mk_transfer_by_transfer_prover_tac ctxt defs transfers =
  unfold_thms_tac ctxt defs THEN HEADGOAL (transfer_prover_add_tac ctxt transfers);

(*FIXME for Jasmin: how to retrieve transfer property from prim(co)rec (transfer)*)
fun derive_transfer_by_transfer_prover ctxt Ys Zs Rs const const_defs transfers =
  let
    val constZ = Term.subst_atomic_types (Ys ~~ Zs) const;

    val goal = mk_parametricity_goal ctxt Rs const constZ;
  in
    Variable.add_free_names ctxt goal []
    |> (fn vars => Goal.prove (*no sorry*) ctxt vars [] goal (fn {context = ctxt, prems = _} =>
      mk_transfer_by_transfer_prover_tac ctxt const_defs transfers))
    |> Thm.close_derivation
  end;

fun flip f a b = f b a;

fun fill_None (SOME x :: xs) ys = x :: fill_None xs ys
  | fill_None (NONE :: xs) (y::ys) = y :: fill_None xs ys
  | fill_None [] _ = []
  | fill_None _ [] = error "resBs has wrong size";

fun mk_internal internal ctxt f =
  if internal andalso not (Config.get ctxt bnf_internals) then f else I;

fun postproc_thms lthy Ts T's_opt ctors ctor's_opt mk_pos thms raw_simps =
  let
    val ((xs, xs'), names_lthy) = lthy
      |> mk_Frees "x" Ts
      ||>> mk_Frees "x'" (the_default Ts T's_opt);
  in
    @{map 5} (fn ctor => fn ctor' => fn x => fn x' =>
      let
        val ct = Thm.cterm_of names_lthy (ctor $ x) |> SOME;
        val ct' = Thm.cterm_of names_lthy (ctor' $ x') |> SOME;
        val ct_ctor = Thm.cterm_of names_lthy ctor' |> SOME;
      in
        infer_instantiate' names_lthy (mk_pos (ct, ct', ct_ctor))
        #> unfold_thms names_lthy thms
      end)
    ctors (the_default ctors ctor's_opt) xs xs' raw_simps
    |> Variable.export names_lthy lthy
  end;

fun mk_default_tyargs Ds Ts =
  map (fn D => (NONE, (D, @{sort type}))) Ds @
  map (fn T => (SOME Binding.empty, (T, @{sort type}))) Ts;
val mk_n_of_Type = fst o dest_Type;
val mk_Frees_of = @{map 2} (Free oo pair o Binding.name_of);
val mk_Freess_of = @{map 2} (@{map 2} (Free oo pair o Binding.name_of));
fun mk_id_lambda T = Abs (Name.uu_, T, Bound 0);
fun mk_undefined T = Const (@{const_name undefined}, T);
val Sfn = flip (curry op $);
val NoBind = (Binding.empty, Binding.empty, Binding.empty);
val discs_sels = false;
fun mk_decl b T = (b, SOME T, NoSyn);
val mk_const_val = Const oo pair o fst o dest_Const;
val mk_Trueprop_list_conj = HOLogic.mk_Trueprop o Library.foldr1 HOLogic.mk_conj;
fun name_of (Const (x, _)) = x
  | name_of (Free (x, _)) = x
  | name_of _ = error "not const or free";

fun reorder_deads resDs As T bnf =
  let
    val Ts = T |> dest_Type |> snd;
    val Us = mk_T_of_bnf resDs As bnf |> dest_Type |> snd;
    val subst = Term.typ_subst_atomic (Ts ~~ Us);
    val resDs' = map subst resDs;
  in
    permute_deads (permute_like_unique (op =) resDs' resDs) bnf
  end;

fun update_bnf_in_sugar f ({T, BT, X, fp, fp_res, fp_res_index, pre_bnf, fp_bnf, absT_info,
    fp_nesting_bnfs, live_nesting_bnfs, fp_ctr_sugar, fp_bnf_sugar,
    fp_co_induct_sugar} : fp_sugar) =
  {T = T,
   BT = BT,
   X = X,
   fp = fp,
   fp_res = fp_res,
   fp_res_index = fp_res_index,
   pre_bnf = pre_bnf,
   fp_bnf = f fp_bnf,
   absT_info = absT_info,
   fp_nesting_bnfs = fp_nesting_bnfs,
   live_nesting_bnfs = live_nesting_bnfs,
   fp_ctr_sugar = fp_ctr_sugar,
   fp_bnf_sugar = fp_bnf_sugar,
   fp_co_induct_sugar = fp_co_induct_sugar} : fp_sugar;

val reorder_deads_in_sugar = update_bnf_in_sugar ooo reorder_deads;

fun define_shape bs resDs As A's Dsss Fss lthy =
  let
    val nr_param = length Fss;
    val nr_occ = length (hd Fss);
    fun rep_par x = replicate nr_param x;

    val b_names = map Binding.name_of bs;
    val b_name = mk_common_name b_names;
    val b = Binding.name b_name;

    fun mk_internal_of_b qualified name =
      Binding.prefix_name (name ^ "_") #> qualified ? Binding.prefix true b_name #> Binding.concealed;
    fun mk_internal_b qualified name = mk_internal_of_b qualified name b;

    (* shape and label *)

    val shape_bs = map (mk_internal_b false o mk_shapeN) (1 upto nr_param);
    val shape_names = map (Local_Theory.full_name lthy) shape_bs;
    val shape_As = map (Type o rpair (resDs @ As)) shape_names;
    val label_b = mk_internal_b false labelN;
    val label_name = Local_Theory.full_name lthy label_b;
    val labelT = Type (label_name, []);
    val F_Tss' = @{map 2} (@{map 2} (flip mk_T_of_bnf shape_As)) Dsss Fss;

    val leaf_b = mk_internal_b true shape_LeafN;
    val mk_node_b = mk_internal_b true o mk_shape_NodeN;
    val mk_label_b = mk_internal_b true o mk_label_CtorN;

    val shapes_ctor_specs = @{map 2} (fn T => fn F_Ts =>
      (((Binding.empty, leaf_b), [(Binding.empty, T)]), NoSyn) :: map_index (fn (i, F_T) =>
        (((Binding.empty, mk_node_b i), [(Binding.empty, F_T)]), NoSyn))
        F_Ts)
      As F_Tss';
    val Depth_ctor_specs = map (fn i => (((Binding.empty, mk_label_b i), []), NoSyn))
      (1 upto nr_occ);

    val shapes_specs = @{map 2} (fn b => fn ctor =>
        (((((mk_default_tyargs resDs As, b), NoSyn), ctor), NoBind), []))
      shape_bs shapes_ctor_specs;
    val Depth_spec = ((((([], label_b), NoSyn), Depth_ctor_specs), NoBind), []);
    val plugins = Plugin_Name.make_filter lthy (K (curry (op =) transfer_plugin));

    val lthy = lthy
      |> Local_Theory.open_target |> snd
      |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
      |> co_datatypes Least_FP BNF_LFP.construct_lfp ((plugins, discs_sels), shapes_specs)
      |> co_datatypes Least_FP BNF_LFP.construct_lfp ((plugins, discs_sels), [Depth_spec])
      |> Local_Theory.close_target;

    val shape_sugars = map (the o fp_sugar_of lthy) shape_names;
    val shape_ctor_sugars = @{map 2} (target_ctr_sugar_of_fp_sugar lthy) shape_As shape_sugars;
    val shape_ctorss = map #ctrs shape_ctor_sugars;
    val shape_Leaves = map hd shape_ctorss;
    val shape_Nodess = map tl shape_ctorss;
    val shape_Nodess' = transpose shape_Nodess;

    val label_sugar = (the o fp_sugar_of lthy) label_name;
    val label_ctor_sugar = target_ctr_sugar_of_fp_sugar lthy labelT label_sugar;
    val label_ctors = #ctrs label_ctor_sugar;

    val depthT = HOLogic.listT labelT;
    val depth_sugar = (the o fp_sugar_of lthy o mk_n_of_Type) depthT;
    val depth_ctor_sugar = target_ctr_sugar_of_fp_sugar lthy depthT depth_sugar;
    val depth_Nil = HOLogic.nil_const labelT;
    val depth_Cons = HOLogic.cons_const labelT;

    (* Types *)

    val F_Ass = @{map 2} (@{map 2} (flip mk_T_of_bnf As)) (transpose Dsss) (transpose Fss);
    val F_shape_Ass = @{map 2} (@{map 2} (flip mk_T_of_bnf shape_As)) Dsss Fss;

    val shape_sugars = map
        (reorder_deads_in_sugar resDs As (hd shape_Leaves |> fastype_of |> range_type))
      shape_sugars;
    val shape_Fss = map (fn F_As => map (fn sugar =>
      mk_T_of_bnf resDs F_As (#fp_bnf sugar)) shape_sugars) F_Ass;

    (* Terms *)

    val ((((((((A_vals, shape_vals), shape_F_valss), F_shape_valss), f_vals), u0), x), xs), _) = lthy
      |> mk_Frees "a" As
      ||>> mk_Frees "s" shape_As
      ||>> mk_Freess "s_f" shape_Fss
      ||>> mk_Freess "f_s" F_shape_Ass
      ||>> mk_Frees "f" (@{map 2} (curry (op -->)) As A's)
      ||>> yield_singleton (mk_Frees "u0") depthT
      ||>> yield_singleton (mk_Frees "l") labelT
      ||>> yield_singleton (mk_Frees "ls") depthT;

    val snoc_val = Const (@{const_name snoc}, labelT --> depthT --> depthT);
    val map_Fss = @{map 2} (@{map 2} (fn Ds => mk_map_of_bnf Ds As A's)) Dsss Fss |> transpose;
    val set_Fsss = @{map 2} (@{map 2} (fn Ds => mk_sets_of_bnf (rep_par Ds) (rep_par As))) Dsss Fss
      |> transpose;
    val rel_Fss = @{map 2} (@{map 2} (fn Ds => mk_rel_of_bnf Ds As A's)) Dsss Fss |> transpose;

    (* Destructor definition *)

    fun mk_unLeaves Ds As shape_As shape_ctor_sugars F_shape_Ass lthy =
      let
        val undefineds = map mk_undefined As;
        val lambda_undefinedss = @{map 2} (fn undef => map (flip absdummy undef))
          undefineds F_shape_Ass;
        val (shape_vals, _) = mk_Frees "x" shape_As lthy;
      in
        @{map 4} (fn T => fn sugar => fn lambda_undefineds => fn x =>
          lambda x ((list_comb ((mk_case (Ds @ As) T (#casex sugar)) $ (mk_id_lambda T),
          lambda_undefineds)) $ x))
          As shape_ctor_sugars lambda_undefinedss shape_vals
      end;

    (* Invariant for shape *)

    val invar_shape_Ts = map (fn shape => depthT --> shape --> HOLogic.boolT) shape_As;
    val invar_shape_bs = map (mk_internal_b true o mk_invar_shapeN) (1 upto nr_param);
    val invar_shape_terms = mk_Frees_of invar_shape_bs invar_shape_Ts;
    val case_Depth = mk_case [labelT] HOLogic.boolT (#casex depth_ctor_sugar);
    val pred_Fss = @{map 2} (@{map 2} (flip mk_pred_of_bnf shape_As)) Dsss Fss;

    val case_x_xs1 = lambda x (lambda xs @{term False});
    fun mk_case_x_xs2 f label pred invars =
      lambda x (lambda xs (mk_If (HOLogic.mk_eq (x, label))
      (list_comb (pred, map (Sfn xs) invars) $ f) @{term False}));
    fun mk_invar_shape_eq1 invar leaf u =
      mk_Trueprop_eq (invar $ u0 $ (leaf $ u),
      case_Depth $ @{term True} $ case_x_xs1 $ u0)
    fun mk_invar_shape_eq2 invar label node f pred_F =
      mk_Trueprop_eq (invar $ u0 $ (node $ f),
      case_Depth $ @{term False} $ (mk_case_x_xs2 f label pred_F invar_shape_terms) $ u0);

    val invar_shapes_eq1s = @{map 3} mk_invar_shape_eq1
      invar_shape_terms shape_Leaves A_vals;
    val invar_shapes_eq2ss = @{map 4} (fn invar => @{map 4} (mk_invar_shape_eq2 invar) label_ctors)
      invar_shape_terms shape_Nodess F_shape_valss pred_Fss;

    val invar_shapes_eqs = invar_shapes_eq1s @ flat invar_shapes_eq2ss;
    val invar_shapes_decls = @{map 2} mk_decl invar_shape_bs invar_shape_Ts;

    val ((invar_shapes_free, _, invar_shape_simps_free), (lthy, lthy_old)) = lthy
      |> Local_Theory.open_target |> snd
      |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
      |> primrec false [] invar_shapes_decls
        (map (fn eq => ((Binding.empty_atts, eq), [], [])) invar_shapes_eqs)
      ||> `Local_Theory.close_target;

    val phi = Proof_Context.export_morphism lthy_old lthy;
    val invar_shapes = map (Morphism.term phi) invar_shapes_free;
    val invar_shape_simps = map (Morphism.fact phi) invar_shape_simps_free |> flat;
    val invar_shape_vals = @{map 2} mk_const_val invar_shapes invar_shape_Ts;

    val invar_shape_F_Tss = map (map (fn shape => depthT --> shape --> HOLogic.boolT)) shape_Fss;
    val invar_shape_F_valss = map (@{map 2} mk_const_val invar_shapes) invar_shape_F_Tss;

    (* Flat and Unflat definitions *)

    fun mk_flat_unflat_shapes lthy =
      let
        fun is_compatible Leaf_F f =
          let
            val F_T1 = Term.argument_type_of Leaf_F 0;
            val F_T2 = fastype_of f;
          in
            equal (fst (dest_Type F_T1)) (fst (dest_Type F_T2))
          end;
        val flat_shape_Tss = map (@{map 2} (flip (curry (op -->))) shape_As) shape_Fss;
        val flat_shape_bss = map (fn j => map (mk_internal_b true o mk_flat_shapeN j)
          (1 upto nr_param)) (1 upto nr_occ)
        val unflat_shape_Tss =
          map (@{map 2} (fn shape => fn shape_F => depthT --> shape --> shape_F) shape_As) shape_Fss;
        val unflat_shape_bss = map (fn j => map (mk_internal_b true o mk_unflat_shapeN j)
          (1 upto nr_param)) (1 upto nr_occ)
        val flat_shape_termss = mk_Freess_of flat_shape_bss flat_shape_Tss;
        val unflat_shape_termss = mk_Freess_of unflat_shape_bss unflat_shape_Tss;

        val F_shape_Fsss = map (fn shape_Fs =>
          @{map 2} (@{map 2} (flip mk_T_of_bnf shape_Fs)) Dsss Fss) shape_Fss;

        val shape_F_ctor_sugarss = map (@{map 2} (fn fp_sugar => fn shape =>
              target_ctr_sugar_of_fp_sugar lthy shape fp_sugar)
              shape_sugars) shape_Fss;

        val shape_F_ctorsss = map (map #ctrs) shape_F_ctor_sugarss;
        val shape_F_Leavess = map (map hd) shape_F_ctorsss;
        val shape_F_Nodesss = map (map tl) shape_F_ctorsss;

        val (((f1ss, f2ss), f3sss), _) = lthy
          |> mk_Freess "f1_" F_Ass
          ||>> mk_Freess "f2_" F_shape_Ass
          ||>> mk_Freesss "f3_" F_shape_Fsss;

        val map_F1ss = @{map 2} (@{map 2} (fn Ds =>
          mk_map_of_bnf Ds As shape_As)) (transpose Dsss) (transpose Fss);
        val map_F2sss = map (fn shape_Fs =>
          @{map 2} (@{map 2} (fn Ds => mk_map_of_bnf Ds shape_Fs shape_As)) Dsss Fss) shape_Fss;

        fun mk_flat_shape_eq1 flat Leaf Node map_F1 f1 =
          mk_Trueprop_eq (flat $ (Leaf $ f1), Node $
          (list_comb (map_F1, shape_Leaves) $ f1));
        val flat_shape_eq1ss = @{map 5} (@{map 5} mk_flat_shape_eq1)
          flat_shape_termss shape_F_Leavess shape_Nodess' map_F1ss f1ss

        fun mk_flat_shape_eq2 flats flat shape_Node shape_F_Node map_F2 f3 =
          mk_Trueprop_eq (flat $ (shape_F_Node $ f3),
          shape_Node $ (list_comb (map_F2, flats) $ f3));
        val flat_shape_eq2sss = @{map 4} (fn flats => @{map 5} (fn flat =>
          @{map 4} (mk_flat_shape_eq2 flats flat)) flats shape_Nodess)
          flat_shape_termss shape_F_Nodesss map_F2sss f3sss;

        val flat_shape_eqss = @{map 2} (flip append o flat) flat_shape_eq2sss flat_shape_eq1ss;
        val flat_shape_declss = @{map 2} (@{map 2} mk_decl) flat_shape_bss flat_shape_Tss;

        (* unflat shapes *)

        val case_Depthss = map (map (fn shape_F =>
          mk_case [labelT] shape_F (#casex depth_ctor_sugar))) shape_Fss;
        val undefinedss = map (map mk_undefined) shape_Fss;

        val map_F1ss = @{map 2} (@{map 2} (fn Ds =>
          mk_map_of_bnf Ds shape_As As)) (transpose Dsss) (transpose Fss);
        val map_F2sss = map (fn shape_Fs =>
          @{map 2} (@{map 2} (fn Ds => mk_map_of_bnf Ds shape_As shape_Fs)) Dsss Fss) shape_Fss;

        val unLeaves = mk_unLeaves resDs As shape_As shape_ctor_sugars F_shape_Ass lthy;

        fun mk_unflat_shape_eq unflat unflats case_Depth undef Leaf_F map_F1 Node f2 Node_F map_F2 =
          mk_Trueprop_eq (unflat $ u0 $ (Node $ f2), case_Depth $
            (if is_compatible Leaf_F f2
             then (Leaf_F $ (list_comb (map_F1, unLeaves) $ f2))
             else undef) $
            (lambda x (lambda xs (Node_F $ (list_comb (map_F2, map (Sfn xs) unflats) $ f2)))) $ u0);

        val unflat_shape_eqsss =
          @{map 7} (fn unflats =>
            @{map 9} (fn unflat => fn nodes => fn f2s => fn case_Depth => fn undef => fn leaf =>
              fn map_F1 =>
                @{map 4} (mk_unflat_shape_eq unflat unflats case_Depth undef leaf map_F1)
                nodes f2s)
            unflats shape_Nodess f2ss)
          unflat_shape_termss case_Depthss undefinedss shape_F_Leavess map_F1ss shape_F_Nodesss
          map_F2sss;

        val unflat_shape_eqss = map flat unflat_shape_eqsss;
        val unflat_shape_declss = @{map 2} (@{map 2} mk_decl) unflat_shape_bss unflat_shape_Tss;

        fun mk_flat_unflat flat_decls flat_eqs flat_Ts unflat_decls unflat_eqs unflat_Ts
            (ress, lthy) =
          let
            val ((flat_shapes_free, flat_defs_free, flat_shapes_simps_free), (lthy, lthy_old)) =
              lthy
              |> Local_Theory.open_target |> snd
              |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
              |> primrec false [BNF_LFP_Rec_Sugar.Transfer_Option] flat_decls
                (map (fn eq => ((Binding.empty_atts, eq), [], [])) flat_eqs)
              ||> `Local_Theory.close_target;

            val phi = Proof_Context.export_morphism lthy_old lthy;
            val flat_shapes = map (Morphism.term phi) flat_shapes_free;
            val flat_shapes_simps = map (Morphism.fact phi) flat_shapes_simps_free |> flat;
            val flat_shapes_defs = Morphism.fact phi flat_defs_free;
            val flat_shapes_vals = @{map 2} mk_const_val flat_shapes flat_Ts;

            val ((unflat_shapes_free, _, unflat_shapes_simps_free), (lthy, lthy_old)) = lthy
              |> Local_Theory.open_target |> snd
              |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
              |> primrec false [Nonexhaustive_Option] unflat_decls
                (map (fn eq => ((Binding.empty_atts, eq), [], [])) unflat_eqs)
              ||> `Local_Theory.close_target;

            val phi = Proof_Context.export_morphism lthy_old lthy;
            val unflat_shapes = map (Morphism.term phi) unflat_shapes_free;
            val unflat_shapes_simps = map (Morphism.fact phi) unflat_shapes_simps_free |> flat;
            val unflat_shapes_vals = @{map 2} mk_const_val unflat_shapes unflat_Ts;
          in
            (ress @ [(flat_shapes, flat_shapes_simps, flat_shapes_defs, flat_shapes_vals,
            unflat_shapes, unflat_shapes_simps, unflat_shapes_vals)], lthy)
          end;

        val (flat_unflat_shape_resultss, lthy) = @{fold 6} mk_flat_unflat flat_shape_declss
          flat_shape_eqss flat_shape_Tss unflat_shape_declss unflat_shape_eqss unflat_shape_Tss
          ([], lthy);
      in
        (@{split_list 7} flat_unflat_shape_resultss, lthy)
      end;

    val ((flat_shapess, flat_shape_simpss, flat_shape_defss, flat_shape_valss, _,
      unflat_shape_simpss, unflat_shape_valss), lthy) = mk_flat_unflat_shapes lthy;

    val shape_shape_subst = As ~~ shape_As;
    val shape_shape_As = map (Term.typ_subst_atomic shape_shape_subst) shape_As;
    val ((shape_shape_vals, shape_shape_vals'), _) = lthy
      |>  mk_Frees "s" shape_shape_As
      ||>>  mk_Frees "s'" shape_shape_As;
    val invar_shape_shape_vals = map (Term.subst_atomic_types shape_shape_subst) invar_shape_vals;
    val flat_shape_shape_valss =
      map (map (Term.subst_atomic_types shape_shape_subst)) flat_shape_valss;
    val unflat_shape_shape_valss =
      map (map (Term.subst_atomic_types shape_shape_subst)) unflat_shape_valss;
    fun mk_immerse_shapes shape_sugars Nodes =
      let
        val (Ts, Us) = map_split (dest_funT o fastype_of) Nodes;
        val map_Nodes = map (fn sugar =>
          list_comb (mk_map_of_bnf resDs Ts Us (#fp_bnf sugar), Nodes)) shape_sugars;
      in
        @{map 2} (fn map => fn unflat_shape =>
          HOLogic.mk_comp (map, unflat_shape $ (mk_rev u0))
          |> absfree (dest_Free u0))
        map_Nodes
      end;
    val immerse_shapess =
      @{map 2} (mk_immerse_shapes shape_sugars) shape_Nodess' unflat_shape_shape_valss
      |> transpose;
    val cimmerse_shape_bs = map (mk_internal_b true o mk_cimmerse_shapeN) (1 upto nr_param);
    val cimmerse_shape_Ts = map (fn shape => depthT --> shape --> shape) shape_shape_As;
    val cimmerse_shape_terms = mk_Frees_of cimmerse_shape_bs cimmerse_shape_Ts;
    fun mk_cimmerse_shape_eqs cimmerse immerse_shapes sh =
      let
        val cimmerse_nil = mk_Trueprop_eq (cimmerse $ depth_Nil $ sh, sh);
        val T = fastype_of sh;
        val case_Label = mk_case [] (depthT --> T --> T) (#casex label_ctor_sugar);
        val cimmerse_cons = mk_Trueprop_eq (cimmerse $ (depth_Cons $ x $ xs) $ sh,
          cimmerse $ xs $ (list_comb (case_Label, immerse_shapes) $ x $ xs $ sh));
      in
        [cimmerse_nil, cimmerse_cons]
      end;
    val cimmerse_shape_eqss =
      @{map 3} mk_cimmerse_shape_eqs cimmerse_shape_terms immerse_shapess shape_shape_vals;
    val cimmerse_shape_decls = @{map 2} mk_decl cimmerse_shape_bs cimmerse_shape_Ts;

    val ((cimmerse_shapess_free, _, cimmerse_shape_simpss_free), (lthy, lthy_old)) = lthy
      |> Local_Theory.open_target |> snd
      |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
      |> @{fold_map 2} (fn decl => fn eqs =>
          primrec false [] [decl] (map (fn eq => ((Binding.empty_atts, eq), [], [])) eqs))
        cimmerse_shape_decls cimmerse_shape_eqss
      |>> @{split_list 3}
      ||> `Local_Theory.close_target;

    val phi = Proof_Context.export_morphism lthy_old lthy;
    val cimmerse_shapes = map (Morphism.term phi) (flat cimmerse_shapess_free);
    val cimmerse_shape_simpss = map (Morphism.fact phi) (flat cimmerse_shape_simpss_free);
    val cimmerse_shape_vals = @{map 2} mk_const_val cimmerse_shapes cimmerse_shape_Ts;

    (* Most useful theorems *)

    val F_pred_map_thms = map (map pred_map_of_bnf) Fss |> flat;
    val F_pred_True_thms = map (map pred_True_of_bnf) Fss |> flat;
    val F_pred_set_thms = map (map pred_set_of_bnf) Fss |> flat;
    val F_pred_cong_thms = map (map pred_cong_of_bnf) Fss |> flat;
    val F_pred_mono_strong_thms = map (map pred_mono_strong_of_bnf) Fss |> flat;
    val F_map_comp_thms = map (map map_comp_of_bnf) Fss |> flat;
    val F_map_cong_thms = map (map map_cong_of_bnf) Fss |> flat;
    val F_map_cong_pred_thms = map (map map_cong_pred_of_bnf) Fss |> flat;
    val F_map_ident_thms = map (map map_ident_of_bnf) Fss |> flat;
    val F_map_transfer_thms = map (map map_transfer_of_bnf) Fss |> flat;
    val F_set_map_thms = map (map set_map_of_bnf) Fss |> flat |> flat;
    val F_rel_map_thms = map (map rel_map_of_bnf) Fss |> flat |> flat;
    val F_rel_mono_strong_thms = map (map rel_mono_strong_of_bnf) Fss |> flat;
    val shape_induct_thm = hd shape_sugars
      |> co_induct_of o #common_co_inducts o the o #fp_co_induct_sugar;
    val shape_exhaust_thms = map #exhaust shape_ctor_sugars;
    val shape_case_thms = map #case_thms shape_ctor_sugars |> flat;
    val shape_inject_thmss = map #injects shape_ctor_sugars;
    val shape_inject_thms = shape_inject_thmss |> flat;
    val shape_distinct_thmss = map #distincts shape_ctor_sugars;
    val shape_inj_map_strong_thms = map (inj_map_strong_of_bnf o #fp_bnf) shape_sugars;
    val shape_map_thms = map (#map_thms o #fp_bnf_sugar) shape_sugars |> flat;
    val shape_set_thms = map (#set_thms o #fp_bnf_sugar) shape_sugars |> flat;
    val shape_rel_inject_thms = map (#rel_injects o #fp_bnf_sugar) shape_sugars |> flat;
    val Label_exhaust_thm = #exhaust label_ctor_sugar;
    val Label_case_thms = #case_thms label_ctor_sugar;
    val Label_split_thms = [#split label_ctor_sugar, #split_asm label_ctor_sugar];
    val Label_distinct_thms = #distincts label_ctor_sugar;

    (* ~~~~Lemmas~~~~ *)

    (* invar_shape_depth_iff *)
    fun mk_invar_shape_depth_iff1 invar shape T Leaf a =
      HOLogic.mk_eq (invar $ depth_Nil $ shape,
      HOLogic.mk_exists (name_of a, T, HOLogic.mk_eq (shape, Leaf $ a)))
    val invar_shape_depth_iff1_goals = @{map 5} mk_invar_shape_depth_iff1
      invar_shape_vals shape_vals As shape_Leaves A_vals;
    val invar_shape_depth_iff1_vars = map (single o name_of) shape_vals;
    val invar_shape_depth_iff1_tactics = @{map 5} (fn exhaust => fn P => fn u =>
      fn injects => fn distincts =>
      fn {context = ctxt, prems = _} =>
      mk_invar_shape_depth_iff_leaf_tac ctxt exhaust
      (Thm.cterm_of lthy P) (Thm.cterm_of lthy u) invar_shape_simps injects distincts)
      shape_exhaust_thms invar_shape_depth_iff1_goals shape_vals
      shape_inject_thmss shape_distinct_thmss;

    val invar_shape_depth_iff1_thms = @{map 3} (fn vars => fn goal => fn tac =>
      Goal.prove_sorry lthy vars [] (HOLogic.mk_Trueprop goal) tac
      |> Thm.close_derivation)
      invar_shape_depth_iff1_vars invar_shape_depth_iff1_goals invar_shape_depth_iff1_tactics;

    val pred_Fss = @{map 2} (@{map 2} (flip mk_pred_of_bnf shape_As)) Dsss Fss;
    fun mk_invar_shape_depth_iff2 invar shape label f_S Node pred_F =
      HOLogic.mk_eq (invar $ (depth_Cons $ label $ u0) $ shape,
      HOLogic.mk_exists (name_of f_S, fastype_of f_S, HOLogic.mk_conj (HOLogic.mk_eq (shape, Node $ f_S),
      (fold (fn invar => fn agg => agg $ (invar $ u0)) invar_shape_vals pred_F) $ f_S)));
    val invar_shape_depth_iff2_goalss = @{map 5} (fn invar => fn shape =>
      @{map 4} (mk_invar_shape_depth_iff2 invar shape) label_ctors)
      invar_shape_vals shape_vals F_shape_valss shape_Nodess pred_Fss;
    val invar_shape_depth_iff2_varss = map (fn shape => map name_of [u0, shape]) shape_vals;

    val invar_shape_depth_iff2_tacticss = @{map 3} (fn exhaust => fn u => map (fn P =>
      fn {context = ctxt, prems = _} => mk_invar_shape_depth_iff_node_tac ctxt exhaust
      (Thm.cterm_of ctxt P) (Thm.cterm_of ctxt u) invar_shape_simps shape_inject_thms
      (shape_distinct_thmss |> flat) F_pred_map_thms Label_case_thms Label_distinct_thms))
      shape_exhaust_thms shape_vals invar_shape_depth_iff2_goalss;

    val invar_shape_depth_iff2_thmss = @{map 3} (fn vars => @{map 2} (fn tac => fn goal =>
      Goal.prove_sorry lthy vars [] (HOLogic.mk_Trueprop goal) tac
      |> Thm.close_derivation))
      invar_shape_depth_iff2_varss invar_shape_depth_iff2_tacticss invar_shape_depth_iff2_goalss;

    val invar_shape_depth_iff_thms = @{map 2} cons invar_shape_depth_iff1_thms invar_shape_depth_iff2_thmss |> flat;

    (* invar_shape_closed *)

    fun mk_invar_map_shape_closed f_vals invar invar' inst map_T =
      lambda inst (HOLogic.mk_all (name_of u0, depthT,
      HOLogic.mk_eq (invar' $ u0 $ (list_comb (map_T, f_vals) $ inst),
      invar $ u0 $ inst)));

    val shape's = map (mk_T_of_bnf resDs A's o #fp_bnf) shape_sugars;
    val invar_shape_T's = map (fn shape => depthT --> shape --> HOLogic.boolT) shape's;
    val invar_shape'_vals = @{map 2} mk_const_val invar_shapes invar_shape_T's;
    val map_shapes = map (mk_map_of_bnf resDs As A's o #fp_bnf) shape_sugars;

    val invar_map_shape_closed_subgoals = @{map 4} (mk_invar_map_shape_closed f_vals)
      invar_shape_vals invar_shape'_vals shape_vals map_shapes;
    val invar_map_shape_closed_goals = @{map 2} Sfn shape_vals invar_map_shape_closed_subgoals;
    val invar_map_shape_closed_goal = mk_Trueprop_list_conj invar_map_shape_closed_goals;
    val invar_map_shape_closed_vars = map name_of (f_vals @ shape_vals);

    val invar_map_shape_closed_tactic = fn {context = ctxt, prems = _} =>
        mk_invar_map_shape_closed_tac ctxt shape_induct_thm
        (map (Thm.cterm_of lthy) invar_map_shape_closed_subgoals)
        (map (Thm.cterm_of lthy) shape_vals) shape_map_thms invar_shape_simps
        F_pred_map_thms F_pred_cong_thms Label_split_thms;

    val invar_map_shape_closed_thms =
      let
        val raw_thm =
          Goal.prove_sorry lthy invar_map_shape_closed_vars []
            invar_map_shape_closed_goal invar_map_shape_closed_tactic
          |> Thm.close_derivation;
      in
        map (fn thm => thm RS spec) (split_conj_thm raw_thm)
      end;

    (* Lemmas about Flat and unFlat *)

    (* flat_shape_unflat_shape *)

    fun mk_flat_shape_unflat_shape label invar shape flat_shape unflat_shape =
      lambda shape (HOLogic.mk_all (name_of u0, depthT,
      HOLogic.mk_imp (invar $ (snoc_val $ label $ u0) $ shape,
      HOLogic.mk_eq (flat_shape $ (unflat_shape $ u0 $ shape), shape))));
    val flat_shape_unflat_shape_subgoalss = @{map 3} (fn label =>
      @{map 4} (mk_flat_shape_unflat_shape label) invar_shape_vals shape_vals)
      label_ctors flat_shape_valss unflat_shape_valss;
    val flat_shape_unflat_shape_goalss = map (@{map 2} Sfn shape_vals) flat_shape_unflat_shape_subgoalss;
    val flat_shape_unflat_shape_goals = map mk_Trueprop_list_conj flat_shape_unflat_shape_goalss;
    val flat_shape_unflat_shape_vars = map name_of shape_vals;

    val flat_shape_unflat_shape_tactics = @{map 3} (fn unflat_simps => fn flat_simps => fn Ps =>
      (fn {context = ctxt, prems = _} =>
        mk_flat_shape_unflat_shape_tac ctxt shape_induct_thm
        (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) shape_vals) Label_exhaust_thm
        Label_split_thms Label_distinct_thms shape_case_thms shape_inject_thms
        invar_shape_depth_iff_thms unflat_simps flat_simps invar_shape_simps F_pred_map_thms
        F_map_comp_thms F_map_cong_pred_thms F_map_ident_thms F_pred_mono_strong_thms))
      unflat_shape_simpss flat_shape_simpss flat_shape_unflat_shape_subgoalss;

    val flat_shape_unflat_shape_thmss =
      let
        val raw_thms = @{map 2} (fn goal => fn tac =>
          Goal.prove_sorry lthy flat_shape_unflat_shape_vars [] goal tac
          |> Thm.close_derivation)
          flat_shape_unflat_shape_goals flat_shape_unflat_shape_tactics;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (fn thm => thm RS spec RS mp)) raw_thmss
      end;

    (* unflat_shape_flat_shape *)

    fun mk_unflat_shape_flat_shape invar shape_F flat_shape unflat_shape =
      lambda shape_F (HOLogic.mk_all (name_of u0, depthT,
      HOLogic.mk_imp (invar $ u0 $ shape_F,
      HOLogic.mk_eq (unflat_shape $ u0 $ (flat_shape $ shape_F), shape_F))));
    val unflat_shape_flat_shape_subgoalss = @{map 4} (@{map 4} mk_unflat_shape_flat_shape)
      invar_shape_F_valss shape_F_valss flat_shape_valss unflat_shape_valss;
    val unflat_shape_flat_shape_goalss = @{map 2} (@{map 2} Sfn) shape_F_valss unflat_shape_flat_shape_subgoalss;
    val unflat_shape_flat_shape_goals = map mk_Trueprop_list_conj unflat_shape_flat_shape_goalss;
    val unflat_shape_flat_shape_varss = map (map name_of) shape_F_valss;

    val unflat_shape_flat_shape_tactics = @{map 4} (fn flat_simps => fn unflat_simps => fn Ps => fn us =>
      (fn {context = ctxt, prems = _} =>
        mk_unflat_shape_flat_shape_tac ctxt shape_induct_thm
        (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) us) shape_inject_thms unflat_simps flat_simps
        invar_shape_simps F_pred_map_thms F_map_comp_thms F_pred_True_thms shape_case_thms
        F_map_cong_pred_thms F_map_ident_thms F_pred_mono_strong_thms Label_split_thms))
      flat_shape_simpss unflat_shape_simpss unflat_shape_flat_shape_subgoalss shape_F_valss;

    val unflat_shape_flat_shape_thmss =
      let
        val raw_thms = @{map 3} (fn vars => fn goal => fn tac =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          unflat_shape_flat_shape_varss unflat_shape_flat_shape_goals unflat_shape_flat_shape_tactics;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (fn thm => thm RS spec RS mp)) raw_thmss
      end;

    (* invar_shape_flat_shape *)

    fun mk_invar_shape_flat_shape label invar invar_F shape_F flat =
      lambda shape_F (HOLogic.mk_all (name_of u0, depthT,
      HOLogic.mk_imp (invar_F $ u0 $shape_F,
      invar $ (snoc_val $ label $ u0) $ (flat $ shape_F))));
    val invar_shape_flat_shape_subgoalss = @{map 4} (fn label =>
      @{map 4} (mk_invar_shape_flat_shape label) invar_shape_vals)
      label_ctors invar_shape_F_valss shape_F_valss flat_shape_valss;
    val invar_shape_flat_shape_goalss =
      @{map 2} (@{map 2} Sfn) shape_F_valss invar_shape_flat_shape_subgoalss;
    val invar_shape_flat_shape_goals = map mk_Trueprop_list_conj invar_shape_flat_shape_goalss;
    val invar_shape_flat_shape_varss = map (map name_of) shape_F_valss;

    val invar_shape_flat_shape_tactics = @{map 3} (fn flat_simps => fn Ps => fn us =>
      (fn {context = ctxt, prems = _} =>
        mk_invar_shape_flat_shape_tac ctxt shape_induct_thm
        (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) us) flat_simps
        invar_shape_simps F_pred_map_thms F_pred_True_thms F_pred_mono_strong_thms
        Label_split_thms F_pred_cong_thms))
      flat_shape_simpss invar_shape_flat_shape_subgoalss shape_F_valss;

    val invar_shape_flat_shape_thmss =
      let
        val raw_thms = @{map 3} (fn vars => fn goal => fn tac =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          invar_shape_flat_shape_varss invar_shape_flat_shape_goals invar_shape_flat_shape_tactics;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (fn thm => thm RS spec RS mp)) raw_thmss
      end;

    (* invar_shape_unflat_shape *)

    fun mk_invar_shape_unflat_shape label invar shape invar_F unflat =
      lambda shape (HOLogic.mk_all (name_of u0, depthT,
      HOLogic.mk_imp (invar $ (snoc_val $ label $ u0) $ shape,
      invar_F $ u0 $ (unflat $ u0 $ shape))));
    val invar_shape_unflat_shape_subgoalss = @{map 3} (fn label =>
      @{map 4} (mk_invar_shape_unflat_shape label) invar_shape_vals shape_vals)
      label_ctors invar_shape_F_valss unflat_shape_valss;
    val invar_shape_unflat_shape_goalss = map (@{map 2} Sfn
      shape_vals) invar_shape_unflat_shape_subgoalss;
    val invar_shape_unflat_shape_goals = map mk_Trueprop_list_conj invar_shape_unflat_shape_goalss;
    val invar_shape_unflat_shape_vars = map name_of shape_vals;

    val invar_shape_unflat_shape_tactics = @{map 2} (fn unflat_simps => fn Ps =>
      (fn {context = ctxt, prems = _} =>
        mk_invar_shape_unflat_shape_tac ctxt shape_induct_thm
        (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) shape_vals) unflat_simps
        invar_shape_simps F_pred_map_thms F_pred_mono_strong_thms Label_split_thms
        Label_distinct_thms))
      unflat_shape_simpss invar_shape_unflat_shape_subgoalss;

    val invar_shape_unflat_shape_thmss =
      let
        val raw_thms = @{map 2} (fn goal => fn tac =>
          Goal.prove_sorry lthy invar_shape_unflat_shape_vars [] goal tac
          |> Thm.close_derivation)
          invar_shape_unflat_shape_goals invar_shape_unflat_shape_tactics;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (fn thm => thm RS spec RS mp)) raw_thmss
      end;

    (* cimmerse_shape_inj *)

    val cimmerse_shape_inj_thms =
      let
        val mk_invar_shapes = @{map 2} (fn invar_shape => fn sh =>
          invar_shape $ (mk_rev u0) $ sh) invar_shape_shape_vals;
        val invar_shape1s = mk_invar_shapes shape_shape_vals;
        val invar_shape2s = mk_invar_shapes shape_shape_vals';
        val cimmerse_shape_eqs = @{map 3} (fn cimmerse_shape => fn sh1 => fn sh2 =>
            HOLogic.mk_eq (cimmerse_shape $ u0 $ sh1, cimmerse_shape $ u0 $ sh2))
          cimmerse_shape_vals shape_shape_vals shape_shape_vals';
        val eqs = @{map 2} (curry HOLogic.mk_eq) shape_shape_vals shape_shape_vals';
        fun mk_goal invar1 invar2 cimmerse_eq eq sh1 sh2 =
          Library.foldr1 HOLogic.mk_imp [invar1, invar2, cimmerse_eq, eq]
          |> list_all_free [sh1, sh2];
        val goals = @{map 6} mk_goal invar_shape1s invar_shape2s cimmerse_shape_eqs eqs
          shape_shape_vals shape_shape_vals';
      in
        @{map 4} (fn goal => fn flat_shape_vals => fn cimmerse_simps => fn shape_inj_map_strong =>
          Goal.prove_sorry lthy [dest_Free u0 |> fst] [] (HOLogic.mk_Trueprop goal)
            (fn {context = ctxt, prems = _} => mk_cimmerse_shape_inj_tac ctxt
               (Thm.cterm_of ctxt (lambda u0 goal)) (Thm.cterm_of ctxt u0)
               (map (Thm.cterm_of ctxt) flat_shape_vals) Label_exhaust_thm Label_case_thms
               shape_inject_thms invar_map_shape_closed_thms
               (flat invar_shape_unflat_shape_thmss) (flat flat_shape_unflat_shape_thmss)
               cimmerse_simps shape_inj_map_strong)
          |> Thm.close_derivation
          |> (fn thm => thm RS spec RS spec RS mp RS mp RS mp))
        goals (flat_shape_shape_valss |> transpose) cimmerse_shape_simpss shape_inj_map_strong_thms
      end;

    (* Characteristic Theorems *)

    (* map_flat_shape *)

    val F_A'ss = @{map 2} (@{map 2} (flip mk_T_of_bnf A's)) Dsss Fss |> transpose;
    val shape's = map (mk_T_of_bnf resDs A's o #fp_bnf) shape_sugars;
    val shape_F'ss = map (fn F_Ts =>
      map (mk_T_of_bnf resDs F_Ts o #fp_bnf) shape_sugars) F_A'ss;
    val map_shapes = map (mk_map_of_bnf resDs As A's o #fp_bnf) shape_sugars;
    val map_shape_Fss = @{map 2} (fn F_Ts => fn F_T's =>
      map (mk_map_of_bnf resDs F_Ts F_T's o #fp_bnf) shape_sugars) F_Ass F_A'ss;
    val flat_shape'_Tss = map (@{map 2} (flip (curry (op -->))) shape's) shape_F'ss;
    val flat_shape'_valss = @{map 2} (@{map 2} mk_const_val) flat_shapess flat_shape'_Tss;

    fun mk_map_flat_shape map_Fs map_S flat_a shape_F flat_c map_S_F =
      lambda shape_F (HOLogic.mk_eq (list_comb (map_S, f_vals) $ (flat_a $ shape_F),
      flat_c $ (list_comb (map_S_F, map (fn map_F =>
      list_comb (map_F, f_vals)) map_Fs) $ shape_F)));
    val map_flat_shape_subgoalss = @{map 5} (fn map_Fs =>
      @{map 5} (mk_map_flat_shape map_Fs) map_shapes)
      map_Fss flat_shape_valss shape_F_valss
      flat_shape'_valss map_shape_Fss;
    val map_flat_shape_goalss = @{map 2} (@{map 2} Sfn)
      shape_F_valss map_flat_shape_subgoalss;
    val map_flat_shape_goals = map mk_Trueprop_list_conj map_flat_shape_goalss;
    val map_flat_shape_varss = map (fn shape_As => map name_of (shape_As @ f_vals)) shape_F_valss;

    val map_flat_shape_tactics = @{map 3} (fn flat_simps => fn Ps => fn us =>
      (fn {context = ctxt, prems = _} =>
        mk_map_flat_shape_tac ctxt shape_induct_thm
        (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) us) shape_inject_thms
        shape_map_thms flat_simps F_map_comp_thms F_map_cong_thms))
      flat_shape_simpss map_flat_shape_subgoalss shape_F_valss;

    val map_flat_shape_thmss =
      let
        val raw_thms = @{map 3} (fn vars => fn goal => fn tac =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          map_flat_shape_varss map_flat_shape_goals map_flat_shape_tactics;
      in
        map split_conj_thm raw_thms
      end;

    (* set_flat_shape *)

    val set_shapess = map (mk_sets_of_bnf (rep_par resDs) (rep_par As) o #fp_bnf) shape_sugars;
    val set_shape_Fsss = map (fn F_Ts =>
      map (mk_sets_of_bnf (rep_par resDs) (rep_par F_Ts) o #fp_bnf) shape_sugars) F_Ass;

    fun mk_set_flat_shapes set_Fss set_shapes flat shape_F set_shape_Fs =
      @{map 2} (fn set_shape => fn set_Fs =>
      lambda shape_F (HOLogic.mk_eq (set_shape $ (flat $ shape_F),
      Library.foldl1 mk_union (@{map 2} (fn set_shape_F => fn set_F =>
      mk_UNION (set_shape_F $ shape_F) set_F) set_shape_Fs set_Fs))))
      set_shapes (transpose set_Fss);
    val set_flat_shape_subgoalsss = @{map 4} (fn set_Fss =>
      @{map 4} (mk_set_flat_shapes set_Fss) set_shapess)
      set_Fsss flat_shape_valss shape_F_valss set_shape_Fsss;
    val set_flat_shape_subgoalsss = transpose (map transpose set_flat_shape_subgoalsss);
    val set_flat_shape_goalsss = map (@{map 2} (@{map 2} Sfn)
      shape_F_valss) set_flat_shape_subgoalsss;
    val set_flat_shape_goalss = map (map mk_Trueprop_list_conj) set_flat_shape_goalsss;
    val set_flat_shape_varss = map (map name_of) shape_F_valss;

    val set_flat_shape_tacticss = map (@{map 3} (fn flat_simps => fn us => fn Ps =>
      (fn {context = ctxt, prems = _} =>
        mk_set_flat_shape_tac ctxt shape_induct_thm
        (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) us) shape_set_thms flat_simps
        F_set_map_thms))
      flat_shape_simpss shape_F_valss) set_flat_shape_subgoalsss;

    val set_flat_shape_thmsss =
      let
        val raw_thmss = @{map 2} (@{map 3} (fn vars => fn tac => fn goal =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          set_flat_shape_varss) set_flat_shape_tacticss set_flat_shape_goalss;
      in
       map (map split_conj_thm) raw_thmss
      end;

    (* rel_flat_shape *)

    val ((shape_F'_valss, Q_vals), _) = lthy
      |> mk_Freess "s_f'" shape_F'ss
      ||>> mk_Frees "Q" (@{map 2} mk_pred2T As A's);
    val invar_shape_F_T'ss = map (map (fn shape => depthT --> shape --> HOLogic.boolT)) shape_F'ss;
    val invar_shape_F'_valss = map (@{map 2} mk_const_val invar_shapes) invar_shape_F_T'ss;
    val rel_shapes = map (mk_rel_of_bnf resDs As A's o #fp_bnf) shape_sugars;
    val rel_shape_Fss = @{map 2} (fn F_Ts => fn F_T's =>
      map (mk_rel_of_bnf resDs F_Ts F_T's o #fp_bnf) shape_sugars) F_Ass F_A'ss;
    val flat_shape_transfer_thmss = @{map 2} (fn defs => map (fn const =>
      derive_transfer_by_transfer_prover lthy As A's Q_vals const defs F_map_transfer_thms))
      flat_shape_defss flat_shape_valss;

    fun mk_rel_flat_shape_raw rel_Fs rel_T shape shape' invar invar' flat flat' rel_T_F =
      lambda shape (HOLogic.mk_all (name_of shape', fastype_of shape',
      HOLogic.mk_all (name_of u0, depthT,
      HOLogic.mk_imp (invar $ u0 $ shape,
      HOLogic.mk_imp (invar' $ u0 $ shape',
      HOLogic.mk_imp (list_comb (rel_T, Q_vals) $ (flat $ shape) $ (flat' $ shape'),
      list_comb (rel_T_F, map (fn rel_F =>
      list_comb (rel_F, Q_vals)) rel_Fs) $ shape $ shape'))))));
    fun mk_rel_flat_shape rel_Fs rel_T shape shape' invar invar' flat flat' rel_T_F =
      Logic.mk_implies (HOLogic.mk_Trueprop (invar $ u0 $ shape),
      Logic.mk_implies (HOLogic.mk_Trueprop (invar' $ u0 $ shape'),
      mk_Trueprop_eq (list_comb (rel_T, Q_vals) $ (flat $ shape) $ (flat' $ shape'),
      list_comb (rel_T_F, map (fn rel_F =>
      list_comb (rel_F, Q_vals)) rel_Fs) $ shape $ shape')));

    val rel_flat_shape_raw_subgoalss = @{map 8} (fn rel_Fs =>
      @{map 8} (mk_rel_flat_shape_raw rel_Fs) rel_shapes)
      rel_Fss shape_F_valss shape_F'_valss invar_shape_F_valss
      invar_shape_F'_valss flat_shape_valss flat_shape'_valss rel_shape_Fss;
    val rel_flat_shape_raw_goalss = @{map 2} (@{map 2} Sfn)
      shape_F_valss rel_flat_shape_raw_subgoalss;
    val rel_flat_shape_raw_goals = map mk_Trueprop_list_conj rel_flat_shape_raw_goalss;
    val rel_flat_shape_raw_varss = map (fn shapes => map name_of (shapes @ Q_vals)) shape_F_valss;

    val rel_flat_shape_raw_tactics = @{map 2} (fn Ps => fn us =>
      (fn {context = ctxt, prems = _} =>
        mk_rel_flat_shape_raw_tac ctxt shape_induct_thm
        (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) us) invar_shape_simps
        (flat_shape_simpss |> flat) shape_rel_inject_thms invar_shape_depth_iff_thms
        F_rel_map_thms F_pred_set_thms F_rel_mono_strong_thms Label_split_thms))
      rel_flat_shape_raw_subgoalss shape_F_valss;

    val rel_flat_shape_raw_thmss =
      let
        val raw_thms = @{map 3} (fn vars => fn goal => fn tac =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          rel_flat_shape_raw_varss rel_flat_shape_raw_goals rel_flat_shape_raw_tactics;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (fn thm => thm RS spec RS spec RS mp RS mp RS mp)) raw_thmss
      end;

    val rel_flat_shape_goalss = @{map 8} (fn rel_Fs =>
      @{map 8} (mk_rel_flat_shape rel_Fs) rel_shapes)
      rel_Fss shape_F_valss shape_F'_valss invar_shape_F_valss
      invar_shape_F'_valss flat_shape_valss flat_shape'_valss rel_shape_Fss;
    val rel_flat_shape_varsss = @{map 2} (@{map 2} (fn shape => fn shape' =>
        map name_of (Q_vals @ [u0, shape, shape'])))
      shape_F_valss shape_F'_valss;

    val rel_flat_shape_thmss = @{map 4} (@{map 4} (fn transfer => fn vars => fn goal => fn rel_flat_shape_raw =>
      Goal.prove_sorry lthy vars [] goal (fn {context = ctxt, prems = _} =>
        mk_rel_final_tac ctxt rel_flat_shape_raw transfer)
      |> Thm.close_derivation)) flat_shape_transfer_thmss
      rel_flat_shape_varsss rel_flat_shape_goalss rel_flat_shape_raw_thmss;

    (* shape result *)

    val shape_res = {
        shape_sugars = shape_sugars,
        label_sugar = label_sugar,
        labelT = labelT,
        snoc_val = snoc_val,

        invar_shape_vals = invar_shape_vals,
        invar_shape_simps = invar_shape_simps,
        flat_shape_valss = flat_shape_valss,
        flat_shape_simpss = flat_shape_simpss,
        flat_shape_transfer_thmss = flat_shape_transfer_thmss,
        unflat_shape_valss = unflat_shape_valss,
        unflat_shape_simpss = unflat_shape_simpss,
        immerse_shape_valss = immerse_shapess,
        cimmerse_shape_vals = cimmerse_shape_vals,
        cimmerse_shape_simpss = cimmerse_shape_simpss,

        invar_shape_depth_iff_thms = invar_shape_depth_iff_thms,
        invar_shape_depth_iff1_thms = invar_shape_depth_iff1_thms,
        invar_map_shape_closed_thms = invar_map_shape_closed_thms,
        invar_shape_flat_shape_thmss = invar_shape_flat_shape_thmss,
        invar_shape_unflat_shape_thmss = invar_shape_unflat_shape_thmss,
        flat_shape_unflat_shape_thmss = flat_shape_unflat_shape_thmss,
        unflat_shape_flat_shape_thmss = unflat_shape_flat_shape_thmss,
        cimmerse_shape_inj_thms = cimmerse_shape_inj_thms,
        map_flat_shape_thmss = map_flat_shape_thmss,
        set_flat_shape_thmsss = set_flat_shape_thmsss,
        rel_flat_shape_thmss = rel_flat_shape_thmss
      };

  in (shape_res, lthy) end

fun construct_nu_fp fp mixfixes map_bs rel_bs pred_bs set_boss bs resBs resDs specs absT_infos lthy =
  let
    (* specs = [(G1, [[F1A, F2A], [F1B, F2B]]), (G2, [[E3A], [E3B]])] *)

    val nr_occs = map (length o hd o snd) specs;
    val nr_occ = Integer.sum nr_occs;
    val nr_param = (length o snd o hd) specs;
    val nr_Ts = length specs;
    val dummy_occs = map_index (fn (i, occs) => replicate occs i) nr_occs;
    fun rep_occs xs = @{map 2} replicate nr_occs xs;
    (* Restores occurences out of a list of length nr_Ts *)
    fun rest_occs xs = flat (rep_occs xs);
    (* Restores occurences out of a list of length nr_occ *)
    fun rest_occs2 xs = Library.unflat dummy_occs xs;
    (* Restores occurences out of a nested list of length nr_occ x nr_Ts *)
    fun rest_occs3 xss = rest_occs2 (@{map 2} nth xss (flat dummy_occs))
    fun rep_par x = replicate nr_param x;

    val Dss = map (fst o fst) specs;
    val Gs = map (snd o fst) specs;
    val F_infoss = fold (@{map 2} (flip append)) (map snd specs) (rep_par []);
    val Dsss = map (map fst) F_infoss;
    val Fss = map (map snd) F_infoss;
    val deads = fold (union (op =)) (Dss @ flat Dsss) resDs;
    val names_lthy = fold Variable.declare_typ deads lthy;

    val ((As, A's), _) = names_lthy
      |> mk_TFrees nr_param
      ||>> mk_TFrees nr_param;

    (*val internals = Config.get lthy bnf_internals;*)
    val b_names = map Binding.name_of bs;
    val b_name = mk_common_name b_names;
    val b = Binding.name b_name;

    fun mk_internal_of_b qualified name =
      Binding.prefix_name (name ^ "_") #> qualified ? Binding.prefix true b_name #> Binding.concealed;
    fun mk_internal_b qualified name = mk_internal_of_b qualified name b;
    fun mk_internal_bs qualified name = map (mk_internal_of_b qualified name) bs;
    (*val external_bs = @{map 2} (Binding.prefix false) b_names bs
      |> not internals ? map Binding.concealed;*)

    (* destructors *)

    fun mk_unLeaves Ds As shape_As shape_ctor_sugars F_shape_Ass lthy =
      let
        val undefineds = map mk_undefined As;
        val lambda_undefinedss = @{map 2} (fn undef => map (flip absdummy undef))
          undefineds F_shape_Ass;
        val (shape_vals, _) = mk_Frees "x" shape_As lthy;
      in
        @{map 4} (fn T => fn sugar => fn lambda_undefineds => fn x =>
          lambda x ((list_comb ((mk_case (Ds @ As) T (#casex sugar)) $ (mk_id_lambda T),
          lambda_undefineds)) $ x))
          As shape_ctor_sugars lambda_undefinedss shape_vals
      end;

    fun mk_unCons raws Ds As raw_ctor_sugars G_shape_raws =
      let
        fun mk_eq raw sugar T =
          lambda raw ((mk_case (Ds @ As) T (#casex (sugar))) $ (mk_id_lambda T) $ raw);
      in
        @{map 3} mk_eq raws raw_ctor_sugars G_shape_raws
      end;


    (* shape and label *)

    val (shape_res, lthy) = define_shape bs resDs As A's Dsss Fss lthy;

    val shape_sugars = #shape_sugars shape_res;
    val shape_As = map (mk_T_of_bnf resDs As o #fp_bnf) shape_sugars;
    val shape_ctor_sugars = @{map 2} (target_ctr_sugar_of_fp_sugar lthy) shape_As shape_sugars;
    val shape_ctorss = map #ctrs shape_ctor_sugars;
    val shape_Leaves = map hd shape_ctorss;
    val shape_Nodess = map tl shape_ctorss;
    val shape_Nodess' = transpose shape_Nodess;
    val F_Ass = @{map 2} (@{map 2} (flip mk_T_of_bnf As)) Dsss Fss |> transpose;
    val F_A'ss = @{map 2} (@{map 2} (flip mk_T_of_bnf A's)) Dsss Fss |> transpose;
    val shape_Fss = map (fn F_As => map (fn sugar =>
      mk_T_of_bnf resDs F_As (#fp_bnf sugar)) shape_sugars) F_Ass;
    val F_shape_Ass = @{map 2} (@{map 2} (flip mk_T_of_bnf shape_As)) Dsss Fss;

    val labelT = #labelT shape_res;
    val label_sugar = #label_sugar shape_res;
    val label_ctor_sugar = target_ctr_sugar_of_fp_sugar lthy labelT label_sugar;
    val label_ctors = #ctrs label_ctor_sugar;
    val depthT = HOLogic.listT (#labelT shape_res);
    val depth_Nil = HOLogic.nil_const labelT;
    val depth_Cons = HOLogic.cons_const labelT;
    val snoc_val = #snoc_val shape_res;

    val invar_shape_vals = #invar_shape_vals shape_res;
    val invar_shape_simps = #invar_shape_simps shape_res;
    val flat_shape_valss = #flat_shape_valss shape_res;
    val flat_shape_transfer_thmss = #flat_shape_transfer_thmss shape_res;
    val unflat_shape_valss = #unflat_shape_valss shape_res;
    val cimmerse_shape_vals = #cimmerse_shape_vals shape_res;

    val cimmerse_shape_simpss = #cimmerse_shape_simpss shape_res;
    val invar_shape_depth_iff1_thms = #invar_shape_depth_iff1_thms shape_res;
    val invar_map_shape_closed_thms = #invar_map_shape_closed_thms shape_res;
    val invar_shape_flat_shape_thms = flat (#invar_shape_flat_shape_thmss shape_res);
    val invar_shape_unflat_shape_thms = flat (#invar_shape_unflat_shape_thmss shape_res);
    val flat_shape_unflat_shape_thms = flat (#flat_shape_unflat_shape_thmss shape_res);
    val unflat_shape_flat_shape_thms = flat (#unflat_shape_flat_shape_thmss shape_res);
    val map_flat_shape_thms = flat (#map_flat_shape_thmss shape_res);
    val set_flat_shape_thms = flat (flat (#set_flat_shape_thmsss shape_res));
    val rel_flat_shape_thms = flat (#rel_flat_shape_thmss shape_res);


    (* raw *)

    val raw_bs = mk_internal_bs false rawN;
    val raw_names = map (Local_Theory.full_name lthy) raw_bs;
    val raw_As = map (Type o rpair (resDs @ As)) raw_names;
    val G_Ts = @{map 2} (flip mk_T_of_bnf (shape_As @ rest_occs raw_As)) Dss Gs;
    val construct_fp = case_fp fp BNF_LFP.construct_lfp BNF_GFP.construct_gfp;

    val raw_ctor_specs = map (fn G_T =>
      (((Binding.empty, mk_internal_b true raw_CtorN), [(mk_internal_b true raw_DtorN, G_T)]), NoSyn)) G_Ts;
    val raw_specs = @{map 2} (fn ctor => fn b =>
      (((((mk_default_tyargs resDs As, b), NoSyn), [ctor]), NoBind), [])) raw_ctor_specs raw_bs;
    val plugins = Plugin_Name.make_filter lthy (K (curry (op =) transfer_plugin));

    val lthy = lthy
      |> Local_Theory.open_target |> snd
      |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
      |> co_datatypes fp construct_fp ((plugins, discs_sels), raw_specs)
      |> Local_Theory.close_target
      handle EMPTY_DATATYPE _ => raise EMPTY_DATATYPE "";

    val raw_sugars = map (the o fp_sugar_of lthy) raw_names;
    val raw_ctor_sugars = @{map 2} (target_ctr_sugar_of_fp_sugar lthy) raw_As raw_sugars;
    val raw_ctors = map (hd o #ctrs) raw_ctor_sugars;
    val raw_dtors = map (hd o hd o #selss) raw_ctor_sugars;

    (* Types *)

    val raw_sugars = map
        (reorder_deads_in_sugar resDs As (hd raw_ctors |> fastype_of |> range_type))
      raw_sugars;
    val raw_F_Ass = map (fn F_As => map (mk_T_of_bnf resDs F_As o #fp_bnf) raw_sugars) F_Ass;
    val G_shape_raws = @{map 2} (flip mk_T_of_bnf (shape_As @ rest_occs raw_As)) Dss Gs;
    val G_shape_F_raw_Fss = @{map 2} (fn shapes => fn raws =>
      @{map 2} (flip mk_T_of_bnf (shapes @ rest_occs raws)) Dss Gs) shape_Fss raw_F_Ass;

    (* Terms *)

    val ((((((((raw_vals, raw_F_valss), G_vals), f_vals), Q_vals), u0), x), xs), _) = lthy
      |> mk_Frees "r" raw_As
      ||>> mk_Freess "r_f" raw_F_Ass
      ||>> mk_Frees "g" G_shape_raws
      ||>> mk_Frees "f" (@{map 2} (curry op -->) As A's)
      ||>> mk_Frees "Q" (@{map 2} mk_pred2T As A's)
      ||>> yield_singleton (mk_Frees "u0") depthT
      ||>> yield_singleton (mk_Frees "l") labelT
      ||>> yield_singleton (mk_Frees "ls") depthT;

    val raw_F_ctor_sugarss = map (@{map 2} (flip (target_ctr_sugar_of_fp_sugar lthy)) raw_sugars) raw_F_Ass
    val raw_F_ctorss = map (map (hd o #ctrs)) raw_F_ctor_sugarss;
    val raw_F_dtorss = map (map (hd o hd o #selss)) raw_F_ctor_sugarss;
    val map_Fss = @{map 2} (@{map 2} (fn Ds => mk_map_of_bnf Ds As A's)) Dsss Fss |> transpose;
    val set_Fsss = @{map 2} (@{map 2} (fn Ds => mk_sets_of_bnf (rep_par Ds) (rep_par As))) Dsss Fss
      |> transpose;
    val rel_Fss = @{map 2} (@{map 2} (fn Ds => mk_rel_of_bnf Ds As A's)) Dsss Fss |> transpose;

    (* Invariant for raw *)

    val invar_raw_Ts = map (fn raw => depthT --> raw --> HOLogic.boolT) raw_As;
    val invar_raw_bs = mk_internal_bs true invar_rawN;
    val invar_raw_terms = mk_Frees_of invar_raw_bs invar_raw_Ts;
    val pred_Gs = @{map 2} (flip mk_pred_of_bnf (shape_As @ rest_occs raw_As)) Dss Gs;
    fun mk_invar_raw invar raw_Cons G_val pred_G = Logic.mk_implies
      (HOLogic.mk_Trueprop (list_comb (list_comb (pred_G, map (Sfn u0) invar_shape_vals),
        @{map 2} (fn invar => fn label =>
        invar $ (depth_Cons $ label $ u0)) (rest_occs invar_raw_terms) label_ctors) $ G_val),
      HOLogic.mk_Trueprop (invar $ u0 $ (raw_Cons $ G_val)));

    val inductive_flags = {quiet_mode = true, verbose = false, alt_name = Binding.empty,
      coind = (fp = Greatest_FP), no_elim = false, no_ind = (fp <> Greatest_FP),
      skip_mono = false};

    val invar_raw_intros = @{map 4} mk_invar_raw invar_raw_terms raw_ctors G_vals pred_Gs;
    val invar_raw_decls = @{map 2} (rpair NoSyn oo pair) invar_raw_bs invar_raw_Ts;

    val (invar_result_raw, (lthy, lthy_old)) = lthy
      |> Local_Theory.open_target |> snd
      |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
      |> Inductive.add_inductive_i inductive_flags invar_raw_decls []
        (map (pair Binding.empty_atts) invar_raw_intros) (map pred_mono_of_bnf Gs)
      ||> `Local_Theory.close_target;

    val phi = Proof_Context.export_morphism lthy_old lthy;
    val invar_result = Inductive.transform_result phi invar_result_raw;
    val invar_raws = #preds invar_result;
    val invar_raw_vals = @{map 2} mk_const_val invar_raws invar_raw_Ts;
    val invar_raw_coinduct_thm = #induct invar_result;

    val invar_raw_F_Tss = map (map (fn raw_F => depthT --> raw_F --> HOLogic.boolT)) raw_F_Ass;
    val invar_raw_F_valss = map (@{map 2} mk_const_val invar_raws) invar_raw_F_Tss;

    (* flat raw and unflat raw *)

    val dual_primrec = case_fp fp (primrec false []) (primcorec false []);
    val dual_primrec_transfer = case_fp fp
      (primrec false [BNF_LFP_Rec_Sugar.Transfer_Option])
      (primcorec false [BNF_GFP_Rec_Sugar.Transfer_Option]);

    fun mk_flat_unflat_raw lthy =
      let
        val flat_raw_Tss = map (@{map 2} (flip (curry (op -->))) raw_As) raw_F_Ass;
        val flat_raw_bss = map (fn j => map (mk_internal_b true o mk_flat_rawN j)
          (1 upto nr_Ts)) (1 upto nr_occ);
        val unflat_raw_Tss = map (@{map 2}
          (fn raw => fn raw_F => depthT --> raw --> raw_F) raw_As) raw_F_Ass;
        val unflat_raw_bss = map (fn j => map (mk_internal_b true o mk_unflat_rawN j)
          (1 upto nr_Ts)) (1 upto nr_occ);
        val flat_raw_termss = mk_Freess_of flat_raw_bss flat_raw_Tss;
        val unflat_raw_termss = mk_Freess_of unflat_raw_bss unflat_raw_Tss;
        val (g_F_valss, _) = lthy
          |> mk_Freess "g_F" G_shape_F_raw_Fss;
        val map_Gss = @{map 2} (fn shape_Fs => fn raw_Fs => @{map 2} (fn Ds =>
            mk_map_of_bnf Ds (shape_Fs @ rest_occs raw_Fs) (shape_As @ rest_occs raw_As)) Dss Gs)
          shape_Fss raw_F_Ass;
        val raw_varss = case_fp fp (@{map 2} (@{map 2} Sfn) g_F_valss raw_F_ctorss) raw_F_valss;
        val g_varss = case_fp fp g_F_valss (@{map 2} (@{map 2} Sfn) raw_F_valss raw_F_dtorss);

        fun mk_flat flat_shapes flat_raws Cons flat_raw r g map_G =
          mk_Trueprop_eq (flat_raw $ r,
          Cons $ (list_comb (list_comb (map_G, flat_shapes), rest_occs flat_raws) $ g))

        val flat_eqss = @{map 6} (fn flat_shape_vals => fn flat_raw_vals =>
          @{map 5} (mk_flat flat_shape_vals flat_raw_vals) raw_ctors)
          flat_shape_valss flat_raw_termss flat_raw_termss raw_varss g_varss map_Gss;
        val flat_declarationss = @{map 2} (@{map 2} mk_decl) flat_raw_bss flat_raw_Tss;

        (* unflat raw *)
        val map_Gss = @{map 2} (fn shape_Fs => fn raw_Fs => @{map 2} (fn Ds =>
            mk_map_of_bnf Ds (shape_As @ rest_occs raw_As) (shape_Fs @ rest_occs raw_Fs)) Dss Gs)
          shape_Fss raw_F_Ass;
        val raw_vars = case_fp fp (@{map 2} Sfn G_vals raw_ctors) raw_vals;
        val g_vars = case_fp fp G_vals (@{map 2} Sfn raw_vals raw_dtors);

        fun mk_unflat unflat_shapes unflat_raws r g unflat_raw Cons_F map_G =
          mk_Trueprop_eq (unflat_raw $ u0 $ r,
          Cons_F $ (list_comb (list_comb (map_G,
          map (Sfn u0) unflat_shapes),
          @{map 2} (fn unflat => fn label =>
          unflat $ (depth_Cons $ label $ u0)) (rest_occs unflat_raws) label_ctors) $ g));

        val unflat_eqss = @{map 5} (fn unflat_shape_vals => fn unflat_raw_vals =>
          @{map 5} (mk_unflat unflat_shape_vals unflat_raw_vals) raw_vars g_vars)
          unflat_shape_valss unflat_raw_termss unflat_raw_termss raw_F_ctorss map_Gss;
        val unflat_declarationss = @{map 2} (@{map 2} mk_decl) unflat_raw_bss unflat_raw_Tss;

        fun mk_flat_unflat flat_decls flat_eqs flat_Ts unflat_decls unflat_eqs unflat_Ts
            (ress, lthy) =
          let
            val ((flat_raw_free, flat_raw_defs_free, flat_raw_simps_free), (lthy, lthy_old)) = lthy
              |> Local_Theory.open_target |> snd
              |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
              |> dual_primrec_transfer flat_decls
                (map (fn eq => ((Binding.empty_atts, eq), [], [])) flat_eqs)
              ||> `Local_Theory.close_target;

            val phi = Proof_Context.export_morphism lthy_old lthy;
            val flat_raws = map (Morphism.term phi) flat_raw_free;
            val flat_raw_simps = map (Morphism.fact phi) flat_raw_simps_free |> flat;
            val flat_raw_defs = (Morphism.fact phi) flat_raw_defs_free;
            val flat_raw_vals = @{map 2} mk_const_val flat_raws flat_Ts;

            val ((unflat_raw_free, _, unflat_raw_simps_free), (lthy, lthy_old)) = lthy
              |> Local_Theory.open_target |> snd
              |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
              |> dual_primrec unflat_decls (map (fn eq => ((Binding.empty_atts, eq), [], [])) unflat_eqs)
              ||> `Local_Theory.close_target;

            val phi = Proof_Context.export_morphism lthy_old lthy;
            val unflat_raws = map (Morphism.term phi) unflat_raw_free;
            val unflat_raw_simps = map (Morphism.fact phi) unflat_raw_simps_free |> flat;
            val unflat_raw_vals = @{map 2} mk_const_val unflat_raws unflat_Ts;
          in
            (ress @ [(flat_raws, flat_raw_simps, flat_raw_defs, flat_raw_vals, unflat_raws,
              unflat_raw_simps, unflat_raw_vals)], lthy)
          end;

        val (ress, lthy) = @{fold 6} mk_flat_unflat flat_declarationss flat_eqss flat_raw_Tss
          unflat_declarationss unflat_eqss unflat_raw_Tss ([], lthy);
      in
        (@{split_list 7} ress, lthy)
      end;

    val ((flat_rawss, flat_raw_simpss, flat_raw_defss, flat_raw_valss,
      _, unflat_raw_simpss, unflat_raw_valss), lthy) = mk_flat_unflat_raw lthy

    val raw_shape_subst = As ~~ shape_As;
    val shape_shape_As = map (Term.typ_subst_atomic raw_shape_subst) shape_As;
    val raw_shape_As = map (Term.typ_subst_atomic raw_shape_subst) raw_As;
    val (raw_shape_vals, _) = lthy
      |>  mk_Frees "r" raw_shape_As;
    val invar_raw_shape_vals = map (Term.subst_atomic_types raw_shape_subst) invar_raw_vals;
    val unflat_raw_shape_valss =
      map (map (Term.subst_atomic_types raw_shape_subst)) unflat_raw_valss;
    fun mk_immerse_raws raw_sugars Nodes =
      let
        val (Ts, Us) = map_split (dest_funT o fastype_of) Nodes;
        val map_Nodes = map (fn sugar =>
          list_comb (mk_map_of_bnf resDs Ts Us (#fp_bnf sugar), Nodes)) raw_sugars;
      in
        @{map 2} (fn map => fn unflat_raw =>
          HOLogic.mk_comp (map, unflat_raw $ (mk_rev u0))
          |> absfree (dest_Free u0))
        map_Nodes
      end;
    val (immerse_rawss', immerse_rawss) =
      @{map 2} (mk_immerse_raws raw_sugars) shape_Nodess' unflat_raw_shape_valss
      |> `transpose;
    val cimmerse_raw_bs = mk_internal_bs true cimmerse_rawN;
    val cimmerse_raw_Ts = map (fn raw => depthT --> raw --> raw) raw_shape_As;
    val cimmerse_raw_terms = mk_Frees_of cimmerse_raw_bs cimmerse_raw_Ts;
    fun mk_cimmerse_raw_eqs cimmerse immerse_raws sh =
      let
        val cimmerse_nil = mk_Trueprop_eq (cimmerse $ depth_Nil $ sh, sh);
        val T = fastype_of sh;
        val case_Label = mk_case [] (depthT --> T --> T) (#casex label_ctor_sugar);
        val cimmerse_cons = mk_Trueprop_eq (cimmerse $ (depth_Cons $ x $ xs) $ sh,
          cimmerse $ xs $ (list_comb (case_Label, immerse_raws) $ x $ xs $ sh));
      in
        [cimmerse_nil, cimmerse_cons]
      end;
    val cimmerse_raw_eqss =
      @{map 3} mk_cimmerse_raw_eqs cimmerse_raw_terms immerse_rawss' raw_shape_vals;
    val cimmerse_raw_decls = @{map 2} mk_decl cimmerse_raw_bs cimmerse_raw_Ts;

    val ((cimmerse_rawss_free, _, cimmerse_raw_simpss_free), (lthy, lthy_old)) = lthy
      |> Local_Theory.open_target |> snd
      |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
      |> @{fold_map 2} (fn decl => fn eqs =>
          primrec false [] [decl] (map (fn eq => ((Binding.empty_atts, eq), [], [])) eqs))
        cimmerse_raw_decls cimmerse_raw_eqss
      |>> @{split_list 3}
      ||> `Local_Theory.close_target;

    val phi = Proof_Context.export_morphism lthy_old lthy;
    val cimmerse_raws = map (Morphism.term phi) (flat cimmerse_rawss_free);
    val cimmerse_raw_simpss = map (Morphism.fact phi) (flat cimmerse_raw_simpss_free);
    val cimmerse_raw_vals = @{map 2} mk_const_val cimmerse_raws cimmerse_raw_Ts;

    (* inductive raw Witness *)

    val wit_raw_bs = mk_internal_bs true wit_rawN;

    val decode = nth (rest_occs (0 upto (nr_Ts - 1)));
    val gen_shapes =
      let
        val F_witsss = @{map 2} (@{map 2} (fn Ds => fn F => mk_wits_of_bnf
          (replicate (nwits_of_bnf F) Ds) (replicate (nwits_of_bnf F) shape_As) F)) Dsss Fss;

        fun gen_shapes label_ids xys i =
          let
            fun wits_of_shape_with_invar [] i = Lazy.lazy (fn () =>
                [([i], nth shape_Leaves i $ nth xys i)])
              | wits_of_shape_with_invar (label :: depth) i = Lazy.lazy (fn () =>
                let
                  val F_witss = nth F_witsss i;
                  val F_wits = nth F_witss label;
                  val args = map (wits_of_shape_with_invar depth) (0 upto nr_param - 1);
                  val Node = nth (nth shape_Nodess i) label;
                in
                  maps (fn wit as (I, _) =>
                    fold (map_product wit_apply) (map (Lazy.force o nth args) I) [wit]
                    |> minimize_wits
                    |> map (apsnd (fn t => Node $ t)))
                  F_wits
                  |> minimize_wits
                end);
          in
            wits_of_shape_with_invar label_ids i
            |> Lazy.force
            |> map snd
          end;
      in gen_shapes end;
    val wit_raw_trmss = mk_ctor_witss decode gen_shapes nr_param Dss As (shape_As @ rest_occs raw_As)
      Gs raw_ctors lthy;
    val wit_raw_trms = @{map 2} (fn xs => K (hd xs) handle Empty => mk_undefined) wit_raw_trmss raw_As; (* take any witness *)
    val wit_raw_freess = map (fn trm => map (fn free =>
      (Free free, mk_undefined (snd free))) (Variable.add_frees lthy trm [])) wit_raw_trms;
    val wit_raw_argss = map (map Free o fst o strip_abs) wit_raw_trms;
    val wit_rawTs = @{map 2} (fn raw => fn args =>
      map (snd o dest_Free) args ---> raw) raw_As wit_raw_argss;

    val wit_raw_eqs = @{map 2} subst_free wit_raw_freess wit_raw_trms;
    val wit_raw_decls = map (rpair NoSyn) wit_raw_bs;

    val ((wit_raws, wit_raw_defs), lthy) =
      @{fold 2} (fn eq => fn decl => fn (res, lthy) =>
      let
        val ((wit_raw_free, (_, wit_raw_def_free)), (lthy, lthy_old)) =
          lthy
          |> Local_Theory.open_target |> snd
          |> Local_Theory.map_background_naming (mk_internal true lthy Name_Space.concealed)
          |> Local_Theory.define (decl, (Binding.empty_atts, eq))
          ||> `Local_Theory.close_target;

        val phi = Proof_Context.export_morphism lthy_old lthy;
        val wit_raw = Morphism.term phi wit_raw_free;
        val wit_raw_def = Morphism.thm phi wit_raw_def_free;
      in (res @ [(wit_raw, wit_raw_def)], lthy) end)
      wit_raw_eqs wit_raw_decls ([], lthy)
      |> apfst @{split_list 2};
    val wit_raw_vals = @{map 2} mk_const_val wit_raws wit_rawTs;

    (* Most useful theorems *)

    val F_pred_map_thms = map (map pred_map_of_bnf) Fss |> flat;
    val F_pred_set_thms = map (map pred_set_of_bnf) Fss |> flat;
    val F_set_map_thms = map (map set_map_of_bnf) Fss |> flat |> flat;
    val F_wit_thms = map (map wit_thms_of_bnf) Fss |> flat |> flat;
    val G_pred_map_thms = map pred_map_of_bnf Gs;
    val G_pred_set_thms = map pred_set_of_bnf Gs;
    val G_pred_True_thms = map pred_True_of_bnf Gs;
    val G_map_comp_thms = map map_comp_of_bnf Gs;
    val G_map_cong_thms = map map_cong_of_bnf Gs;
    val G_map_cong_pred_thms = map map_cong_pred_of_bnf Gs;
    val G_map_ident_thms = map map_ident_of_bnf Gs;
    val G_map_transfer_thms = map map_transfer_of_bnf Gs;
    val G_set_map_thms = map set_map_of_bnf Gs |> flat;
    val G_rel_map_thms = map rel_map_of_bnf Gs |> flat;
    val G_rel_refl_thms = map rel_refl_of_bnf Gs;
    val G_rel_cong_thms = map rel_cong_of_bnf Gs;
    val G_rel_mono_strong_thms = map rel_mono_strong_of_bnf Gs;
    val G_pred_rel_thms = map pred_rel_of_bnf Gs;
    val G_pred_cong_thms = map pred_cong_of_bnf Gs;
    val G_pred_mono_strong_thms = map pred_mono_strong_of_bnf Gs;
    val G_wit_thms = map wit_thms_of_bnf Gs |> flat;
    val shape_case_thms = map #case_thms shape_ctor_sugars |> flat;
    val shape_map_thms = map (#map_thms o #fp_bnf_sugar) shape_sugars |> flat;
    val shape_set_thms = map (#set_thms o #fp_bnf_sugar) shape_sugars |> flat;
    val shape_rel_inject_thms = map (#rel_injects o #fp_bnf_sugar) shape_sugars |> flat;
    val Label_exhaust_thm = #exhaust label_ctor_sugar;
    val Label_distinct_thms = #distincts label_ctor_sugar;
    val Label_split_thms = [#split label_ctor_sugar, #split_asm label_ctor_sugar];
    val Label_case_thms = #case_thms label_ctor_sugar;
    val raw_exhaust_thms = map #exhaust raw_ctor_sugars;
    val raw_split_thmss = map (fn sugar => [#split sugar, #split_asm sugar]) raw_ctor_sugars;
    val raw_inject_thms = map #injects raw_ctor_sugars |> flat;
    val raw_case_thms = map #case_thms raw_ctor_sugars |> flat;
    val raw_sel_thms = map #sel_thmss raw_ctor_sugars |> flat |> flat;
    val raw_map_thms = map (#map_thms o #fp_bnf_sugar) raw_sugars |> flat;
    val raw_set_thms = map (#set_thms o #fp_bnf_sugar) raw_sugars |> flat;
    val raw_rel_inject_thms = map (#rel_injects o #fp_bnf_sugar) raw_sugars |> flat;
    val raw_induct_thm =
      hd raw_sugars |> co_induct_of o #common_co_inducts o the o #fp_co_induct_sugar;
    val raw_rel_coinduct_thm =
      hd raw_sugars |> co_induct_of o #common_rel_co_inducts o the o #fp_co_induct_sugar;
    val raw_set_induct_thms = hd raw_sugars |> #common_set_inducts o the o #fp_co_induct_sugar;

    val invar_raw_simps = postproc_thms lthy G_shape_raws NONE raw_ctors NONE (fn cts => [NONE, #1 cts])
      (@{thms ex_simps simp_thms(39,40)} @ raw_inject_thms) (#eqs invar_result);
    val (flat_raw_simps, flat_raw_simpss) = `flat (@{map 3} (fn Gs => fn ctors =>
      fp = Greatest_FP ? postproc_thms lthy Gs NONE ctors NONE (fn cts => [#1 cts]) raw_sel_thms)
      G_shape_F_raw_Fss raw_F_ctorss flat_raw_simpss);
    val (unflat_raw_simps, unflat_raw_simpss) = `flat (map (fp = Greatest_FP ?
        postproc_thms lthy G_shape_raws NONE raw_ctors NONE (fn cts => [NONE, #1 cts]) raw_sel_thms)
      unflat_raw_simpss);

    (* ~~~~Lemmas~~~~ *)

    (* invar_wit_raw *)

    fun mk_invar_wit_raw invar wit arguments =
      HOLogic.mk_Trueprop (invar $ depth_Nil $ list_comb (wit, arguments));
    val invar_wit_raw_goals = @{map 3} mk_invar_wit_raw invar_raw_vals wit_raw_vals wit_raw_argss;
    val invar_wit_raw_varss = map (map name_of) wit_raw_argss;

    val invar_wit_raw_tactic = fn {context = ctxt, prems = _} =>
      case_fp fp
        (mk_invar_wit_raw_tac ctxt wit_raw_defs invar_raw_simps invar_shape_simps Label_case_thms
          F_pred_map_thms F_pred_set_thms F_set_map_thms F_wit_thms
          G_pred_map_thms G_pred_set_thms G_set_map_thms G_wit_thms)
        (Skip_Proof.cheat_tac ctxt 1);

    val invar_wit_raw_thms = @{map 2} (fn goal => fn vars =>
      Goal.prove_sorry lthy vars [] goal invar_wit_raw_tactic
      |> Thm.close_derivation)
      invar_wit_raw_goals invar_wit_raw_varss;

    (* invar_map_raw_closed *)

    val raw_A's = map (mk_T_of_bnf resDs A's o #fp_bnf) raw_sugars;
    val invar_raw_T's = map (fn raw => depthT --> raw --> HOLogic.boolT) raw_A's;
    val invar_raw'_vals = @{map 2} mk_const_val invar_raws invar_raw_T's;
    val map_raws = map (mk_map_of_bnf resDs As A's o #fp_bnf) raw_sugars;

    fun mk_invar_map_raw_closed f_vals invar invar' raw map_T =
      (raw, invar', u0, list_comb (map_T, f_vals) $ raw, invar, u0, raw);
    fun mk_invar_map_raw_closed_goal terms =
      let
        val Ps = map (fn (x, pred1, arg_l1, arg_r1, pred2, arg_l2, arg_r2) =>
          lambda x (HOLogic.mk_all (name_of u0, fastype_of u0,
            HOLogic.mk_eq (pred1 $ arg_l1 $ arg_r1, pred2 $ arg_l2 $ arg_r2)))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms Ps);
        val xs = map #1 terms;
        val vars = map name_of (f_vals @ xs);
        val tactic = fn {context = ctxt, prems = _} =>
          mk_invar_map_raw_closed_tac ctxt raw_induct_thm
          (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) xs) raw_map_thms
          invar_raw_simps G_pred_map_thms invar_map_shape_closed_thms G_pred_cong_thms;
      in (goal, vars, tactic) end
    fun mk_invar_map_raw_closed_cogoal terms =
      let
        val subgoals = map (fn (x, pred1, arg_l1, arg_r1, pred2, arg_l2, arg_r2) =>
          lambda x (HOLogic.mk_eq (pred1 $ arg_l1 $ arg_r1, pred2 $ arg_l2 $ arg_r2))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms subgoals);
        val ((((ls1, rs1), ls2), rs2), _) = lthy
          |> mk_Frees "left1" (map (fastype_of o #3) terms)
          ||>> mk_Frees "right1" (map (fastype_of o #4) terms)
          ||>> mk_Frees "left2" (map (fastype_of o #6) terms)
          ||>> mk_Frees "right2" (map (fastype_of o #7) terms);
        fun mk_P x l r arg_l arg_r pred = lambda l (lambda r (
          HOLogic.mk_exists (name_of u0, fastype_of u0,
          HOLogic.mk_exists (name_of x, fastype_of x,
          HOLogic.mk_conj (HOLogic.mk_conj (HOLogic.mk_eq (l, arg_l),
            HOLogic.mk_eq (r, arg_r)), pred)))));
        val (Ps, Q_vals) = @{map 5} (fn (x, pred1, arg_l1, arg_r1, pred2, arg_l2, arg_r2) =>
          fn l1 => fn r1 => fn l2 => fn r2 =>
            (mk_P x l2 r2 arg_l2 arg_r2 (pred1 $ arg_l1 $ arg_r1),
            mk_P x l1 r1 arg_l1 arg_r1 (pred2 $ arg_l2 $ arg_r2))) terms ls1 rs1 ls2 rs2
          |> @{split_list 2};
        val vars = name_of u0 :: map (name_of o #1) terms @ map name_of f_vals;

        val tactic = fn {context = ctxt, prems = _} =>
          mk_invar_map_raw_closed_cotac ctxt (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) Q_vals)
          invar_raw_coinduct_thm raw_exhaust_thms raw_inject_thms raw_map_thms invar_raw_simps
          invar_map_shape_closed_thms G_pred_map_thms G_pred_mono_strong_thms;
      in (goal, vars, tactic) end;

    val invar_map_raw_closed_terms = @{map 4} (mk_invar_map_raw_closed f_vals)
      invar_raw_vals invar_raw'_vals raw_vals map_raws;
    val invar_map_raw_closed_goal = case_fp fp mk_invar_map_raw_closed_goal
      mk_invar_map_raw_closed_cogoal invar_map_raw_closed_terms;

    val invar_map_raw_closed_thms =
      let
        val raw_thm = (fn (goal, vars, tac) =>
          Goal.prove_sorry lthy vars [] goal tac) invar_map_raw_closed_goal
          |> Thm.close_derivation;
      in
        case_fp fp (map (fn thm => thm RS spec)) I (split_conj_thm raw_thm)
      end;


    (* Lemmas about flat and unflat *)

    (* invar_raw_flat_raw *)

    fun mk_invar_raw_flat_raw label invar invar_F raw_F flat =
      (raw_F, invar_F $ u0 $ raw_F, invar, snoc_val $ label $ u0, flat $ raw_F);
    fun mk_invar_raw_flat_raw_goal terms =
      let
        val Ps = map (fn (x, asm, pred, arg_l, arg_r) =>
          lambda x (HOLogic.mk_all (name_of u0, fastype_of u0,
          HOLogic.mk_imp (asm, pred $ arg_l $ arg_r)))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms Ps);
        val xs = map #1 terms;
        val vars = map name_of xs;
        val tactic = fn {context = ctxt, prems = _} =>
          mk_invar_raw_flat_raw_tac ctxt raw_induct_thm
          (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) xs) flat_raw_simps
          invar_raw_simps invar_shape_flat_shape_thms G_pred_map_thms G_pred_mono_strong_thms;
      in (goal, vars, tactic) end
    fun mk_invar_raw_flat_raw_cogoal terms =
      let
        val subgoals = map (fn (x, asm, pred, arg_l, arg_r) =>
          lambda x (HOLogic.mk_imp (asm, pred $ arg_l $ arg_r))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms subgoals);
        val ((ls, rs), _) = lthy
          |> mk_Frees "left" (map (fastype_of o #4) terms)
          ||>> mk_Frees "right" (map (fastype_of o #5) terms);
        val Ps = @{map 3} (fn (x, asm, _, arg_l, arg_r) => fn l => fn r =>
          lambda l (lambda r (
          HOLogic.mk_exists (name_of u0, fastype_of u0,
          HOLogic.mk_exists (name_of x, fastype_of x,
          HOLogic.mk_conj (HOLogic.mk_conj (HOLogic.mk_eq (l, arg_l),
            HOLogic.mk_eq (r, arg_r)), asm)))))) terms ls rs;
        val vars = name_of u0 :: map (name_of o #1) terms;

        val tactic = fn {context = ctxt, prems = _} =>
          mk_invar_raw_flat_raw_cotac ctxt (map (Thm.cterm_of lthy) Ps) invar_raw_coinduct_thm
          raw_exhaust_thms raw_inject_thms raw_map_thms raw_sel_thms invar_raw_simps flat_raw_simps
          invar_shape_flat_shape_thms G_pred_map_thms G_pred_mono_strong_thms;
      in (goal, vars, tactic) end;
    val invar_raw_flat_raw_termss = @{map 4} (fn label =>
      @{map 4} (mk_invar_raw_flat_raw label) invar_raw_vals)
      label_ctors invar_raw_F_valss raw_F_valss flat_raw_valss;
    val invar_raw_flat_raw_goals = map (case_fp fp
      mk_invar_raw_flat_raw_goal mk_invar_raw_flat_raw_cogoal) invar_raw_flat_raw_termss;

    val invar_raw_flat_raw_thmss =
      let
        val raw_thms = map (fn (goal, vars, tac) =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          invar_raw_flat_raw_goals;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (case_fp fp (fn thm => thm RS spec RS mp) (fn thm => thm RS mp))) raw_thmss
      end;

    (* invar_raw_unflat_raw *)

    fun mk_invar_raw_unflat_raw label invar raw invar_F unflat =
      (raw, invar $ (snoc_val $ label $ u0) $ raw, invar_F, u0, unflat $ u0 $ raw);
    fun mk_invar_raw_unflat_raw_goal terms =
      let
        val Ps = map (fn (x, asm, pred, arg_l, arg_r) =>
          lambda x (HOLogic.mk_all (name_of u0, fastype_of u0,
          HOLogic.mk_imp (asm, pred $ arg_l $ arg_r)))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms Ps);
        val xs = map #1 terms;
        val vars = map name_of xs;
        val tactic = fn {context = ctxt, prems = _} =>
          mk_invar_raw_unflat_raw_tac ctxt raw_induct_thm
          (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) xs) unflat_raw_simps
          invar_raw_simps invar_shape_unflat_shape_thms G_pred_map_thms G_pred_mono_strong_thms;
      in (goal, vars, tactic) end
    fun mk_invar_raw_unflat_raw_cogoal terms =
      let
        val subgoals = map (fn (x, asm, pred, arg_l, arg_r) =>
          lambda x (HOLogic.mk_imp (asm, pred $ arg_l $ arg_r))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms subgoals);
        val ((ls, rs), _) = lthy
          |> mk_Frees "left" (map (fastype_of o #4) terms)
          ||>> mk_Frees "right" (map (fastype_of o #5) terms);
        val Ps = @{map 3} (fn (x, asm, _, arg_l, arg_r) => fn l => fn r =>
          lambda l (lambda r (
          HOLogic.mk_exists (name_of u0, fastype_of u0,
          HOLogic.mk_exists (name_of x, fastype_of x,
          HOLogic.mk_conj (HOLogic.mk_conj (HOLogic.mk_eq (l, arg_l),
            HOLogic.mk_eq (r, arg_r)), asm)))))) terms ls rs;
        val vars = name_of u0 :: map (name_of o #1) terms;

        val tactic = fn {context = ctxt, prems = _} =>
          mk_invar_raw_unflat_raw_cotac ctxt (map (Thm.cterm_of lthy) Ps) invar_raw_coinduct_thm
          raw_exhaust_thms raw_inject_thms raw_map_thms raw_sel_thms invar_raw_simps unflat_raw_simps
          invar_shape_unflat_shape_thms G_pred_map_thms G_pred_mono_strong_thms;
      in (goal, vars, tactic) end;
    val invar_raw_unflat_raw_termss = @{map 3} (fn label =>
      @{map 4} (mk_invar_raw_unflat_raw label) invar_raw_vals raw_vals)
      label_ctors invar_raw_F_valss unflat_raw_valss;
    val invar_raw_unflat_raw_goals = map (case_fp fp
      mk_invar_raw_unflat_raw_goal mk_invar_raw_unflat_raw_cogoal) invar_raw_unflat_raw_termss;

    val invar_raw_unflat_raw_thmss =
      let
        val raw_thms = map (fn (goal, vars, tac) =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          invar_raw_unflat_raw_goals;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (case_fp fp (fn thm => thm RS spec RS mp) (fn thm => thm RS mp))) raw_thmss
      end;

    (* flat_raw_unflat_raw *)

    fun mk_flat_raw_unflat_raw label raw invar flat unflat =
      (raw, invar $ (snoc_val $ label $ u0) $ raw, flat $ (unflat $ u0 $ raw), raw);
    fun mk_flat_raw_unflat_raw_goal terms =
      let
        val Ps = map (fn (x, asm, left, right) =>
          lambda x (HOLogic.mk_all (name_of u0, fastype_of u0,
          HOLogic.mk_imp (asm, HOLogic.mk_eq (left, right))))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms Ps);
        val xs = map #1 terms;
        val vars = map name_of xs;

        val tactic = fn {context = ctxt, prems = _} =>
          mk_flat_raw_unflat_raw_tac ctxt raw_induct_thm
          (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) xs) unflat_raw_simps
          flat_raw_simps invar_raw_simps raw_inject_thms flat_shape_unflat_shape_thms G_pred_map_thms
          G_map_comp_thms G_map_cong_pred_thms G_map_ident_thms G_pred_mono_strong_thms;
      in (goal, vars, tactic) end;
    fun mk_flat_raw_unflat_raw_cogoal terms =
      let
        val subgoals = map (fn (x, asm, left, right) =>
          lambda x (HOLogic.mk_imp (asm, HOLogic.mk_eq (left, right)))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms subgoals);
        val (_, _, xs, ys) = @{split_list 4} terms;
        val ((ls, rs), _) = lthy
          |> mk_Frees "left" (map fastype_of xs)
          ||>> mk_Frees "right" (map fastype_of ys);
        val Ps = @{map 3} (fn (x, asm, left, right) => fn l => fn r =>
          lambda l (lambda r (
          HOLogic.mk_exists (name_of u0, fastype_of u0,
          HOLogic.mk_exists (name_of x, fastype_of x,
          HOLogic.mk_conj (HOLogic.mk_conj (HOLogic.mk_eq (l, left),
            HOLogic.mk_eq (r, right)), asm)))))) terms ls rs;
        val vars = map name_of (u0 :: ys);

        val tactic = fn {context = ctxt, prems = _} =>
          mk_flat_raw_unflat_raw_cotac ctxt (map (Thm.cterm_of lthy) Ps)
          (map (Thm.cterm_of lthy) xs) (map (Thm.cterm_of lthy) ys) raw_induct_thm
          raw_exhaust_thms raw_sel_thms flat_raw_simps unflat_raw_simps invar_raw_simps
          flat_shape_unflat_shape_thms G_rel_map_thms G_pred_rel_thms G_rel_mono_strong_thms;
      in (goal, vars, tactic) end;

    val flat_raw_unflat_raw_termss = @{map 3} (fn label =>
      @{map 4} (mk_flat_raw_unflat_raw label) raw_vals invar_raw_vals)
      label_ctors flat_raw_valss unflat_raw_valss;
    val flat_raw_unflat_raw_goals = map (case_fp fp
      mk_flat_raw_unflat_raw_goal mk_flat_raw_unflat_raw_cogoal) flat_raw_unflat_raw_termss;

    val flat_raw_unflat_raw_thmss =
      let
        val raw_thms = map (fn (goal, vars, tac) =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          flat_raw_unflat_raw_goals;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (case_fp fp (fn thm => thm RS spec RS mp) (fn thm => thm RS mp))) raw_thmss
      end;

    (* unflat_raw_flat_raw *)

    fun mk_unflat_raw_flat_raw invar_F raw_F flat unflat =
      (raw_F, invar_F $ u0 $ raw_F, unflat $ u0 $ (flat $ raw_F), raw_F);

    fun mk_unflat_raw_flat_raw_goal terms =
      let
        val Ps = map (fn (x, asm, left, right) =>
          lambda x (HOLogic.mk_all (name_of u0, fastype_of u0,
          HOLogic.mk_imp (asm, HOLogic.mk_eq (left, right))))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms Ps);
        val xs = map #1 terms;
        val vars = map name_of xs;

        val tactic = fn {context = ctxt, prems = _} =>
          mk_unflat_raw_flat_raw_tac ctxt raw_induct_thm
          (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) xs) unflat_raw_simps flat_raw_simps
          invar_raw_simps raw_inject_thms unflat_shape_flat_shape_thms G_pred_map_thms
          G_map_comp_thms G_map_cong_pred_thms G_map_ident_thms G_pred_mono_strong_thms;
      in (goal, vars, tactic) end;
    fun mk_unflat_raw_flat_raw_cogoal terms =
      let
        val subgoals = map (fn (x, asm, left, right) =>
          lambda x (HOLogic.mk_imp (asm, HOLogic.mk_eq (left, right)))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms subgoals);
        val (_, _, xs, ys) = @{split_list 4} terms;
        val ((ls, rs), _) = lthy
          |> mk_Frees "left" (map fastype_of xs)
          ||>> mk_Frees "right" (map fastype_of ys);
        val Ps = @{map 3} (fn (x, asm, left, right) => fn l => fn r =>
          lambda l (lambda r (
          HOLogic.mk_exists (name_of u0, fastype_of u0,
          HOLogic.mk_exists (name_of x, fastype_of x,
          HOLogic.mk_conj (HOLogic.mk_conj (HOLogic.mk_eq (l, left),
            HOLogic.mk_eq (r, right)), asm)))))) terms ls rs;
        val vars = map name_of (u0 :: ys);

        val tactic = fn {context = ctxt, prems = _} =>
          mk_unflat_raw_flat_raw_cotac ctxt (map (Thm.cterm_of lthy) Ps)
          (map (Thm.cterm_of lthy) xs) (map (Thm.cterm_of lthy) ys) raw_induct_thm
          raw_exhaust_thms raw_sel_thms flat_raw_simps unflat_raw_simps invar_raw_simps
          unflat_shape_flat_shape_thms G_rel_map_thms G_pred_rel_thms G_rel_mono_strong_thms;
      in (goal, vars, tactic) end;

    val unflat_raw_flat_raw_termss = @{map 4} (@{map 4} mk_unflat_raw_flat_raw)
      invar_raw_F_valss raw_F_valss flat_raw_valss unflat_raw_valss;
    val unflat_raw_flat_raw_goals = map (case_fp fp
      mk_unflat_raw_flat_raw_goal mk_unflat_raw_flat_raw_cogoal) unflat_raw_flat_raw_termss;

    val unflat_raw_flat_raw_thmss =
      let
        val raw_thms = map (fn (goal, vars, tac) =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          unflat_raw_flat_raw_goals;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (case_fp fp (fn thm => thm RS spec RS mp) (fn thm => thm RS mp))) raw_thmss
      end;

    (* invar_cimmerse_raw *)

    val invar_cimmerse_raw_thms =
      let
        val mk_invar_raws = @{map 2} (fn invar_raw => fn sh =>
          invar_raw $ (mk_rev u0) $ sh) invar_raw_shape_vals;
        val invar_raws = mk_invar_raws raw_shape_vals;
        val invar_cimmerse_raws = @{map 3} (fn invar_raw => fn cimmerse_raw => fn sh =>
           invar_raw $ depth_Nil $ (cimmerse_raw $ u0 $ sh))
          invar_raw_shape_vals cimmerse_raw_vals raw_shape_vals;
        fun mk_goal invar invar_cimmerse sh =
          Library.foldr1 HOLogic.mk_imp [invar, invar_cimmerse]
          |> list_all_free [sh];
        val goals = @{map 3} mk_goal invar_raws invar_cimmerse_raws raw_shape_vals;
      in
        @{map 2} (fn goal => fn cimmerse_simps =>
          Goal.prove_sorry lthy [dest_Free u0 |> fst] [] (HOLogic.mk_Trueprop goal)
            (fn {context = ctxt, prems = _} => mk_invar_cimmerse_raw_tac ctxt
               (Thm.cterm_of ctxt (lambda u0 goal)) (Thm.cterm_of ctxt u0)
               Label_exhaust_thm Label_case_thms invar_map_raw_closed_thms
               (flat invar_raw_unflat_raw_thmss) cimmerse_simps)
          |> Thm.close_derivation
          |> (fn thm => thm RS spec RS mp))
        goals cimmerse_raw_simpss
      end;

    val unConss = mk_unCons raw_vals resDs As raw_ctor_sugars G_shape_raws;
    val unConss_shape = map (Term.subst_atomic_types raw_shape_subst) unConss;
    val unCons_cimmerse_raw_thms =
      let
        val map_argTs = shape_shape_As @ rest_occs raw_shape_As;
        val maps = @{map 2} (fn Ds => mk_map_of_bnf Ds map_argTs map_argTs) Dss Gs;
        val ids = map HOLogic.id_const shape_shape_As;
        val mapls = map (fn mapx =>
            list_comb (list_comb (mapx, ids),
              map (fn t => t $ depth_Nil) (flat (rest_occs3 immerse_rawss))))
          maps;
        val rargs = map (fn t => t $ u0) cimmerse_shape_vals @
          @{map 2} (fn t => fn u => t $ (snoc_val $ u $ u0))
          (rest_occs cimmerse_raw_vals) label_ctors;
        val maprs = map (fn map => list_comb (map, rargs)) maps;
        fun mk_goal mapl mapr unCons cimmerse r =
          HOLogic.mk_eq (mapl $ (unCons $ (cimmerse $ u0 $ r)), mapr $ (unCons $ r))
          |> list_all_free [r];
        val goals = @{map 5} mk_goal mapls maprs unConss_shape cimmerse_raw_vals raw_shape_vals;
      in
        @{map 5} (fn goal => fn raw_splits => fn raw_map => fn G_map_comp => fn G_map_cong =>
          Goal.prove_sorry lthy [dest_Free u0 |> fst] [] (HOLogic.mk_Trueprop goal)
            (fn {context = ctxt, prems = _} => mk_unCons_cimmerse_raw_tac ctxt
              (Thm.cterm_of ctxt (lambda u0 goal)) (Thm.cterm_of ctxt u0)
              Label_distinct_thms Label_case_thms Label_split_thms unflat_raw_simps
              (flat cimmerse_raw_simpss) (flat cimmerse_shape_simpss) raw_inject_thms
              raw_splits raw_map G_map_comp G_map_cong)
          |> Thm.close_derivation
          |> (fn thm => thm RS spec))
        goals raw_split_thmss raw_map_thms G_map_comp_thms G_map_cong_thms
      end;


    (* Characteristic Theorems *)

    (* map_flat_raw *)

    val raw_F'ss = map (fn F_Ts =>
      map (mk_T_of_bnf resDs F_Ts o #fp_bnf) raw_sugars) F_A'ss;
    val map_raws = map (mk_map_of_bnf resDs As A's o #fp_bnf) raw_sugars;
    val map_raw_Fss = @{map 2} (fn F_Ts => fn F_T's => map
      (mk_map_of_bnf resDs F_Ts F_T's o #fp_bnf) raw_sugars) F_Ass F_A'ss;
    val flat_raw'_Tss = map (@{map 2} (flip (curry (op -->))) raw_A's) raw_F'ss;
    val flat_raw'_valss = @{map 2} (@{map 2} mk_const_val) flat_rawss flat_raw'_Tss;

    fun mk_map_flat_raw map_Fs map_raw flat_a raw_F flat_c map_raw_F =
      (raw_F, list_comb (map_raw, f_vals) $ (flat_a $ raw_F),
      flat_c $ (list_comb (map_raw_F, map (fn map_F =>
      list_comb (map_F, f_vals)) map_Fs) $ raw_F));

    fun mk_map_flat_raw_goal terms =
      let
        val Ps = map (fn (x, left, right) => lambda x (HOLogic.mk_eq (left, right))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms Ps);
        val xs = map #1 terms;
        val vars = map name_of (xs @ f_vals);

        val tactic = fn {context = ctxt, prems = _} =>
          mk_map_flat_raw_tac ctxt raw_induct_thm
          (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) xs) raw_map_thms flat_raw_simps
          G_map_comp_thms map_flat_shape_thms G_map_cong_thms;
      in (goal, vars, tactic) end;
    fun mk_map_flat_raw_cogoal terms =
      let
        val subgoals = map (fn (_, left, right) => HOLogic.mk_eq (left, right)) terms;
        val goal = mk_Trueprop_list_conj subgoals;
        val (raw_vals, xs, ys) = @{split_list 3} terms;
        val ((ls, rs), _) = lthy
          |> mk_Frees "left" (map fastype_of xs)
          ||>> mk_Frees "right" (map fastype_of ys);
        val Ps = @{map 3} (fn (x, left, right) => fn l => fn r =>
          lambda l (lambda r (
          HOLogic.mk_exists (name_of u0, fastype_of u0,
          HOLogic.mk_exists (name_of x, fastype_of x,
          HOLogic.mk_conj (HOLogic.mk_eq (l, left),
            HOLogic.mk_eq (r, right))))))) terms ls rs;
        val vars = map name_of (raw_vals @ f_vals);

        val tactic = fn {context = ctxt, prems = _} =>
          mk_map_flat_raw_cotac ctxt (map (Thm.cterm_of lthy) Ps)
          (map (Thm.cterm_of lthy) xs) (map (Thm.cterm_of lthy) ys) raw_induct_thm
          raw_exhaust_thms raw_sel_thms raw_map_thms flat_raw_simps
          map_flat_shape_thms G_rel_map_thms G_rel_refl_thms;
      in (goal, vars, tactic) end;

    val map_flat_raw_termss = @{map 5} (fn map_Fs =>
      @{map 5} (mk_map_flat_raw map_Fs) map_raws)
      map_Fss flat_raw_valss raw_F_valss flat_raw'_valss map_raw_Fss;
    val map_flat_raw_goals = map (case_fp fp
      mk_map_flat_raw_goal mk_map_flat_raw_cogoal) map_flat_raw_termss;

    val map_flat_raw_thmss =
      let
        val raw_thms = map (fn (goal, vars, tac) =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          map_flat_raw_goals;
      in
        map split_conj_thm raw_thms
      end;

    (* set_flat_raw *)

    val set_rawss = map (mk_sets_of_bnf (rep_par resDs) (rep_par As) o #fp_bnf) raw_sugars;
    val set_raw_Fsss = map (fn F_Ts =>
      map (mk_sets_of_bnf (rep_par resDs) (rep_par F_Ts) o #fp_bnf) raw_sugars) F_Ass;

    fun mk_set_flat_raw set_Fs set_raw flat raw_F set_raw_Fs =
      (raw_F, set_raw, (flat $ raw_F), Library.foldl1 mk_union
      (@{map 2} (fn set_raw_F => fn set_F =>
      mk_UNION (set_raw_F $ raw_F) set_F) set_raw_Fs set_Fs), set_Fs);
    fun mk_set_flat_raws set_Fss set_raws flat raw_F set_raw_Fs =
      @{map 2} (fn set_Fs => fn set_raw => mk_set_flat_raw set_Fs set_raw flat raw_F set_raw_Fs) set_Fss set_raws;

    fun mk_set_flat_raw_goal terms =
      let
        val Ps = map (fn (x, pred, arg, right, _) =>
          lambda x (HOLogic.mk_eq (pred $ arg,  right))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms Ps);
        val xs = map #1 terms;
        val vars = map name_of xs;

        val tactic = fn {context = ctxt, prems = _} =>
          mk_set_flat_raw_tac ctxt raw_induct_thm
          (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) xs) raw_set_thms flat_raw_simps
          G_set_map_thms set_flat_shape_thms;
      in (goal, vars, tactic) end;
    fun mk_set_flat_raw_cogoal induct_thm terms =
      let
        val subgoals = map (fn (_, pred, arg, right, _) => HOLogic.mk_eq (pred $ arg, right)) terms;
        val goal = mk_Trueprop_list_conj subgoals;
        val (raw_vals, preds, args, _, set_Fss) = @{split_list 5} terms;
        val (((raw'_vals, elem_vals), f_elem_vals), _) = lthy
          |> mk_Frees "raw'" (map fastype_of args)
          ||>> mk_Frees "elem" (map (HOLogic.dest_setT o range_type o fastype_of) preds)
          ||>> mk_Frees "f_elem" (map (domain_type o fastype_of) (hd set_Fss));
        val lemma = list_all_free raw'_vals (Library.foldr1 HOLogic.mk_conj (@{map 3}
          (fn (raw, pred, arg, right, _) => fn raw' => fn elem =>
          mk_Ball (pred $ raw') (lambda elem (HOLogic.mk_all (name_of raw, fastype_of raw,
          HOLogic.mk_imp (HOLogic.mk_eq (arg, raw'), HOLogic.mk_mem (elem, right))))))
          terms raw'_vals elem_vals));
        val Pss = @{map 2} (fn f => fn set_F => map (fn (raw, pred, arg, _, _) =>
          lambda f (lambda raw (mk_Ball (set_F $ f) (
          lambda (hd elem_vals) (HOLogic.mk_mem (hd elem_vals, pred $ arg))))))
          terms) f_elem_vals (hd set_Fss);
        val vars = map name_of (raw_vals @ f_vals);

        val tactic = fn {context = ctxt, prems = _} =>
          mk_set_flat_raw_cotac ctxt (Thm.cterm_of lthy lemma) (map (map (Thm.cterm_of lthy)) Pss)
          (map (Thm.cterm_of lthy o #1) terms) induct_thm raw_set_induct_thms nr_occ
          raw_exhaust_thms raw_inject_thms raw_set_thms flat_raw_simps
          set_flat_shape_thms G_set_map_thms;
      in (goal, vars, tactic) end;

    val set_flat_raw_termsss = @{map 4} (fn set_Fss =>
      @{map 4} (mk_set_flat_raws (transpose set_Fss)) set_rawss)
      set_Fsss flat_raw_valss raw_F_valss set_raw_Fsss
      |> map transpose |> transpose;
    val set_flat_raw_goalss = case_fp fp (map (map mk_set_flat_raw_goal))
      (@{map 2} (fn thm => map (mk_set_flat_raw_cogoal thm)) raw_set_induct_thms) set_flat_raw_termsss;

    val set_flat_raw_thmsss =
      let
        val raw_thms = map (map (fn (goal, vars, tac) =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation))
          set_flat_raw_goalss;
      in
        map (map split_conj_thm) raw_thms
      end;

    (* rel_flat_raw_raw *)

    val (raw_F'_valss, _) = lthy
      |> mk_Freess "raw_f'" raw_F'ss;
    val invar_raw_F_T'ss = map (map (fn shape => depthT --> shape --> HOLogic.boolT)) raw_F'ss;
    val invar_raw_F'_valss = map (@{map 2} mk_const_val invar_raws) invar_raw_F_T'ss;
    val rel_raws = map (mk_rel_of_bnf resDs As A's o #fp_bnf) raw_sugars;
    val rel_raw_Fss = @{map 2} (fn F_As => fn F_A's =>
      map (mk_rel_of_bnf resDs F_As F_A's o #fp_bnf) raw_sugars) F_Ass F_A'ss;
    val flat_raw_transfer_thmss = @{map 2} (fn defs => map (fn const =>
        derive_transfer_by_transfer_prover lthy As A's Q_vals const defs
          (flat flat_shape_transfer_thmss @ G_map_transfer_thms)))
      flat_raw_defss flat_raw_valss;

    fun mk_rel_flat_raw_raw rel_Fs rel_T raw raw' invar invar' flat flat' rel_T_F =
      (raw, raw', invar $ u0 $ raw, invar' $ u0 $ raw',
      list_comb (rel_T, Q_vals) $ (flat $ raw) $ (flat' $ raw'),
      list_comb (rel_T_F, map (fn rel_F =>
      list_comb (rel_F, Q_vals)) rel_Fs) $ raw $ raw', map (fn rel_F =>
      list_comb (rel_F, Q_vals)) rel_Fs);
    fun mk_rel_flat_raw rel_Fs rel_T raw raw' invar invar' flat flat' rel_T_F =
      Logic.mk_implies (HOLogic.mk_Trueprop (invar $ u0 $ raw),
      Logic.mk_implies (HOLogic.mk_Trueprop (invar' $ u0 $ raw'),
      mk_Trueprop_eq (list_comb (rel_T, Q_vals) $ (flat $ raw) $ (flat' $ raw'),
      list_comb (rel_T_F, map (fn rel_F =>
      list_comb (rel_F, Q_vals)) rel_Fs) $ raw $ raw')));

    fun mk_rel_flat_raw_goal terms =
      let
        val Ps = map (fn (raw, raw', asm1, asm2, left, right, _) =>
          lambda raw (HOLogic.mk_all (name_of raw', fastype_of raw',
          HOLogic.mk_all (name_of u0, depthT, HOLogic.mk_imp (asm1,
          HOLogic.mk_imp (asm2, HOLogic.mk_imp (left, right))))))) terms;
        val goal = mk_Trueprop_list_conj (@{map 2} (Sfn o #1) terms Ps);
        val xs = map #1 terms;
        val vars = map name_of (xs @ Q_vals);

        val tactic = fn {context = ctxt, prems = _} =>
          mk_rel_flat_raw_raw_tac ctxt raw_induct_thm
          (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) xs) raw_exhaust_thms
          invar_raw_simps flat_raw_simps raw_rel_inject_thms G_rel_map_thms G_pred_set_thms
          G_rel_mono_strong_thms rel_flat_shape_thms;
      in (goal, vars, tactic) end;
    fun mk_rel_flat_raw_cogoal terms =
      let
        val subgoals = map (fn (_, _, asm1, asm2, left, right, _) =>
          HOLogic.mk_imp (asm1, HOLogic.mk_imp (asm2, HOLogic.mk_imp (left, right)))) terms;
        val goal = mk_Trueprop_list_conj subgoals;
        val (raw_vals, raw'_vals, _, _, _, _, predss) = @{split_list 7} terms;
        val Ps = @{map 3} (fn (_, _, asm1, asm2, left, _, _) => fn raw => fn raw' =>
          lambda raw (lambda raw' (
          HOLogic.mk_exists (name_of u0, fastype_of u0,
          HOLogic.mk_conj (HOLogic.mk_conj (asm1, asm2), left))))) terms raw_vals raw'_vals;
        val vars = map name_of (raw_vals @ raw'_vals @ Q_vals);

        val tactic = fn {context = ctxt, prems = _} =>
          mk_rel_flat_raw_raw_cotac ctxt (map (Thm.cterm_of lthy) Ps) (map (Thm.cterm_of lthy) (hd predss))
          (map (Thm.cterm_of lthy) raw_vals) (map (Thm.cterm_of lthy) raw'_vals) raw_rel_coinduct_thm
          raw_exhaust_thms raw_sel_thms raw_rel_inject_thms flat_raw_simps invar_raw_simps
          rel_flat_shape_thms G_rel_map_thms G_pred_set_thms G_rel_refl_thms G_rel_cong_thms G_rel_mono_strong_thms;
      in (goal, vars, tactic) end;

    val rel_flat_raw_raw_termss = @{map 8} (fn rel_Fs =>
      @{map 8} (mk_rel_flat_raw_raw rel_Fs) rel_raws)
      rel_Fss raw_F_valss raw_F'_valss invar_raw_F_valss
      invar_raw_F'_valss flat_raw_valss flat_raw'_valss rel_raw_Fss;
    val rel_flat_raw_raw_goals = map (case_fp fp
      mk_rel_flat_raw_goal mk_rel_flat_raw_cogoal) rel_flat_raw_raw_termss;

    val rel_flat_raw_raw_thmss =
      let
        val raw_thms = map (fn (goal, vars, tac) =>
          Goal.prove_sorry lthy vars [] goal tac
          |> Thm.close_derivation)
          rel_flat_raw_raw_goals;
        val raw_thmss = map split_conj_thm raw_thms;
      in
        map (map (case_fp fp (fn thm => thm RS spec RS spec) I
          #> (fn thm => thm RS mp RS mp RS mp))) raw_thmss
      end;

    val rel_flat_raw_goalss = @{map 8} (fn rel_Fs =>
      @{map 8} (mk_rel_flat_raw rel_Fs) rel_raws)
      rel_Fss raw_F_valss raw_F'_valss invar_raw_F_valss
      invar_raw_F'_valss flat_raw_valss flat_raw'_valss rel_raw_Fss;
    val rel_flat_raw_varsss = @{map 2} (@{map 2} (fn raw => fn raw' =>  map name_of (Q_vals @ [u0, raw, raw'])))
      raw_F_valss raw_F'_valss;

    val rel_flat_raw_thmss = @{map 4} (@{map 4} (fn transfer => fn vars => fn goal => fn rel_flat_raw_raw =>
      Goal.prove_sorry lthy vars [] goal (fn {context = ctxt, prems = _} =>
        mk_rel_final_tac ctxt rel_flat_raw_raw transfer)
      |> Thm.close_derivation)) flat_raw_transfer_thmss
      rel_flat_raw_varsss rel_flat_raw_goalss rel_flat_raw_raw_thmss;


    (* Type definition *)

    fun mk_set_of_t invar raw rawT =
      HOLogic.mk_Collect (name_of raw, rawT, invar $ depth_Nil $ raw);
    val invar_raw_appl_vals = @{map 3} mk_set_of_t invar_raw_vals raw_vals raw_As;
    fun mk_wit_raw_appl wit arguments =
      Thm.cterm_of lthy (list_comb (wit, map (mk_undefined o snd o dest_Free) arguments));
    val wit_raw_appl_cterms = @{map 2} mk_wit_raw_appl wit_raw_vals wit_raw_argss;
    val typedef_tactics = @{map 2} (fn wit_appl => fn invar_wit =>
      fn ctxt => mk_typedef_tac ctxt wit_appl invar_wit)
      wit_raw_appl_cterms invar_wit_raw_thms;

    val params = map dest_TFree (fill_None (map (AList.lookup (op =)
      (map (`(fst o dest_TFree)) resDs) o fst) resBs) As);

    val (Ts, lthy) = @{fold 4} (fn b => fn mx => fn invar => fn tac => fn (res, lthy) =>
      typedef (b, params, mx) invar NONE (case_fp fp tac (fn ctxt => Skip_Proof.cheat_tac ctxt 1)) lthy
      |>> append res o single)
      bs mixfixes invar_raw_appl_vals typedef_tactics ([], lthy);

    val T_infos = map (fst o snd) Ts;
    val T_thms = map (snd o snd) Ts;
    val T_As = map #abs_type T_infos;
    val T_F_restss = @{map 2} (fn T_A => map (fn F_As =>
      Term.typ_subst_atomic (As ~~ F_As) T_A)) T_As (rest_occs2 F_Ass);


    (* Constructor and Selector *)

    val raw_F_restss = rest_occs3 raw_F_Ass;
    val G_T_Fs = @{map 2} (flip mk_T_of_bnf (As @ flat T_F_restss)) Dss Gs;
    val T_Abs_Ts = @{map 2} (curry op -->) raw_As T_As;
    val T_Abs_vals = @{map 2} (Const oo pair o #Abs_name) T_infos T_Abs_Ts;
    val T_Rep_Ts = @{map 2} (curry op -->) T_As raw_As;
    val T_Rep_vals = @{map 2} (Const oo pair o #Rep_name) T_infos T_Rep_Ts;
    val T_Abs_Fss = @{map 2} (@{map 2} (curry op -->)) raw_F_restss T_F_restss;
    val T_Abs_F_valss = @{map 2} (@{map 2} (Const oo pair o #Abs_name)) (rep_occs T_infos) T_Abs_Fss;
    val T_Rep_Fss = @{map 2} (@{map 2} (curry op -->)) T_F_restss raw_F_restss;
    val T_Rep_F_valss = @{map 2} (@{map 2} (Const oo pair o #Rep_name)) (rep_occs T_infos) T_Rep_Fss;

    val map_G1s = @{map 2} (fn Ds => mk_map_of_bnf Ds (As @ flat T_F_restss) (shape_As @ rest_occs raw_As)) Dss Gs;
    val map_G2s = @{map 2} (fn Ds => mk_map_of_bnf Ds (shape_As @ rest_occs raw_As) (As @ flat T_F_restss)) Dss Gs;
    val ctor_T_Ts = @{map 2} (curry op -->) G_T_Fs T_As;
    val ctor_T_bs = mk_internal_bs true ctor_T_N;
    val dtor_T_Ts = @{map 2} (curry op -->) T_As G_T_Fs;
    val dtor_T_bs = mk_internal_bs true dtor_T_N;
    val ((t_vals, G_T_F_vals), _) = lthy
      |> mk_Frees "t" T_As
      ||>> mk_Frees "g_T_F" G_T_Fs;

    fun mk_T_ctor_eq leaves flats reps abs cons map_G g =
      lambda g (abs $ (cons $ (list_comb (list_comb (map_G, leaves),
      @{map 2} (curry HOLogic.mk_comp) flats reps) $ g)));

    val T_ctor_eqs = @{map 4} (mk_T_ctor_eq shape_Leaves (flat (rest_occs3 flat_raw_valss)) (flat T_Rep_F_valss))
      T_Abs_vals raw_ctors map_G1s G_T_F_vals;
    val T_ctor_decls = map (rpair NoSyn) ctor_T_bs;

    fun mk_T_dtor_eq unLeaves absts unflats rep unCons map_G t =
      lambda t (list_comb (list_comb (map_G, unLeaves),
      @{map 2} (fn abst => fn unflat =>
      HOLogic.mk_comp(abst, unflat $ depth_Nil)) absts unflats) $
      (unCons $ (rep $ t)));

    val unLeaves = mk_unLeaves resDs As shape_As shape_ctor_sugars F_shape_Ass lthy;
    val T_dtor_eqs = @{map 4} (mk_T_dtor_eq unLeaves (flat T_Abs_F_valss) (flat (rest_occs3 unflat_raw_valss)))
      T_Rep_vals unConss map_G2s t_vals;
    val T_dtor_decls = map (rpair NoSyn) dtor_T_bs;

    val ((ctor_Ts, T_ctor_defs, dtor_Ts, T_dtor_defs), lthy) =
      @{fold 4} (fn T_ctor_eq => fn T_ctor_decl => fn T_dtor_eq => fn T_dtor_decl => fn (res, lthy) =>
      let
        val ((T_ctor_free, (_, T_ctor_def_free)), (lthy, lthy_old)) =
          lthy
          |> Local_Theory.open_target |> snd
          |> Local_Theory.define (T_ctor_decl, (Binding.empty_atts, T_ctor_eq))
          ||> `Local_Theory.close_target;

        val phi = Proof_Context.export_morphism lthy_old lthy;
        val T_ctor = Morphism.term phi T_ctor_free;
        val T_ctor_def = Morphism.thm phi T_ctor_def_free;

        val ((T_dtor_free, (_, T_dtor_def_free)), (lthy, lthy_old)) =
          lthy
          |> Local_Theory.open_target |> snd
          |> Local_Theory.define (T_dtor_decl, (Binding.empty_atts, T_dtor_eq))
          ||> `Local_Theory.close_target;

        val phi = Proof_Context.export_morphism lthy_old lthy;
        val T_dtor = Morphism.term phi T_dtor_free;
        val T_dtor_def = Morphism.thm phi T_dtor_def_free;
      in (res @ [(T_ctor, T_ctor_def, T_dtor, T_dtor_def)], lthy) end)
      T_ctor_eqs T_ctor_decls T_dtor_eqs T_dtor_decls ([], lthy)
      |> apfst @{split_list 4};
    val ctor_T_vals = @{map 2} mk_const_val ctor_Ts ctor_T_Ts;
    val dtor_T_vals = @{map 2} mk_const_val dtor_Ts dtor_T_Ts;
    val dtor_T_T's = map (typ_subst_atomic (As ~~ A's)) dtor_T_Ts;
    val dtor_T'_vals = @{map 2} mk_const_val dtor_Ts dtor_T_T's;

    (* Lift BNF *)

    fun lift_specs set_bos = @{map 2} (rpair o apsnd SOME)
      (hd T_As |> dest_Type |> snd |> map dest_TFree) set_bos;

    val lift_args = @{map 6} (fn T => fn map_b => fn rel_b => fn pred_b => fn set_bos => fn wits =>
        ((((([BNF_Lift.No_Warn_Wits, BNF_Lift.Plugins_Option (K (K false))],
          lift_specs set_bos), fst T), SOME wits),
        SOME (#type_definition (snd (snd T)))), (map_b, rel_b, pred_b)))
      Ts map_bs rel_bs pred_bs set_boss wit_raw_trmss;

    val invar_wit_thmss = map (fn wits => replicate (length wits)
      (fn {context = ctxt, prems = _} =>
        mk_invar_wit_tac ctxt wit_raw_defs invar_raw_simps invar_shape_simps
        F_pred_map_thms F_pred_set_thms F_set_map_thms F_wit_thms
        G_pred_map_thms G_pred_set_thms G_set_map_thms G_wit_thms)) wit_raw_trmss;
    val wit_set_thmsss = map (fn wits => replicate (length wits) (rep_par
      (fn {context = ctxt, prems = _} =>
        mk_wit_set_tac ctxt F_wit_thms G_wit_thms raw_set_thms
        shape_set_thms))) wit_raw_trmss;
    val lift_tacss = @{map 3} (fn thm => fn invar_wits => fn set_witss =>
      replicate 2 (fn {context = ctxt, prems = _} => mk_lift_tac ctxt thm) @
      invar_wits @ flat set_witss) invar_map_raw_closed_thms invar_wit_thmss wit_set_thmsss;

    val lthy = @{fold 2} (fn args => fn tacs => fn lthy =>
      BNF_Lift.lift_bnf args tacs lthy) lift_args lift_tacss lthy;

    val T_bnfs = map (the o bnf_of lthy o fst) Ts;
    val T_map_defs = map map_def_of_bnf T_bnfs;
    val T_set_defs = map set_defs_of_bnf T_bnfs |> flat;
    val T_rel_defs = map rel_def_of_bnf T_bnfs;
    val T_bnfs = map (reorder_deads resDs As (hd T_As)) T_bnfs;


    (* Constructor is Bijection *)

    (* un_T_T *)

    fun mk_un_T_o_T dtor_T ctor_T T =
      mk_Trueprop_eq (HOLogic.mk_comp (dtor_T, ctor_T), HOLogic.id_const T);

    val un_T_o_T_goals = @{map 3} mk_un_T_o_T dtor_T_vals ctor_T_vals G_T_Fs;

    val un_T_o_T_tactics = @{map 3} (fn T_ctor_def => fn T_dtr_def => fn T_Abs_inverse =>
      (fn {context = ctxt, prems = _} =>
        mk_un_T_o_T_tac ctxt T_ctor_def T_dtr_def T_Abs_inverse
        (map #Rep T_thms) (map #Rep_inverse T_thms) invar_raw_simps raw_case_thms
        invar_shape_simps shape_case_thms invar_map_raw_closed_thms (flat invar_raw_flat_raw_thmss)
        (flat unflat_raw_flat_raw_thmss) G_pred_map_thms G_pred_True_thms G_map_comp_thms
        G_map_ident_thms G_pred_cong_thms G_map_cong_thms))
      T_ctor_defs T_dtor_defs (map #Abs_inverse T_thms);

    val un_T_o_T_thms = @{map 2} (fn goal => fn tac =>
        Goal.prove_sorry lthy [] [] goal tac
        |> Thm.close_derivation)
        un_T_o_T_goals un_T_o_T_tactics;

    (* T_un_T *)

    fun mk_T_o_un_T dtor_T ctor_T T =
      mk_Trueprop_eq (HOLogic.mk_comp (ctor_T, dtor_T), HOLogic.id_const T);

    val T_o_un_T_goals = @{map 3} mk_T_o_un_T dtor_T_vals ctor_T_vals T_As;

    val T_o_un_T_tactics = @{map 6} (fn T_ctor_def => fn T_dtor_def => fn T_Rep =>
      fn T_Rep_inject => fn raw_exhaust => fn cT_Rep =>
      (fn {context = ctxt, prems = _} =>
        mk_T_o_un_T_tac ctxt T_ctor_def T_dtor_def T_Rep T_Rep_inject
        raw_exhaust cT_Rep (map #Abs_inverse T_thms) (map #Rep T_thms) (map #Rep_inverse T_thms)
        invar_raw_simps raw_case_thms raw_inject_thms invar_shape_simps shape_case_thms
        invar_map_raw_closed_thms (flat invar_raw_flat_raw_thmss) (flat invar_raw_unflat_raw_thmss)
        (flat flat_raw_unflat_raw_thmss) G_pred_map_thms G_pred_True_thms G_map_comp_thms
        invar_shape_depth_iff1_thms G_map_cong_pred_thms G_map_ident_thms G_pred_cong_thms
        G_pred_mono_strong_thms))
      T_ctor_defs T_dtor_defs (map #Rep T_thms) (map #Rep_inject T_thms)
      raw_exhaust_thms (map (Thm.cterm_of lthy) T_Rep_vals);

    val T_o_un_T_thms = @{map 2} (fn goal => fn tac =>
        Goal.prove_sorry lthy [] [] goal tac
        |> Thm.close_derivation)
        T_o_un_T_goals T_o_un_T_tactics;

    val T_un_T_thms = map (fn thm => thm RS @{thm pointfree_idE}) T_o_un_T_thms;
    val un_T_T_thms = map (fn thm => thm RS @{thm pointfree_idE}) un_T_o_T_thms;

    val bij_un_T_thms =
      @{map 2} (fn thm1 => fn thm2 => @{thm o_bij} OF [thm1, thm2]) T_o_un_T_thms un_T_o_T_thms;
    val inj_un_T_thms = map (fn thm => thm RS @{thm bij_is_inj}) bij_un_T_thms;
    val un_T_inject_thms = map (fn thm => thm RS @{thm inj_eq}) inj_un_T_thms;
(*    val surj_un_T_thms = map (fn thm => thm RS @{thm bij_is_surj}) bij_un_T_thms;
    val un_T_nchotomy_thms = map (fn thm => thm RS @{thm surjD}) surj_un_T_thms;
    val un_T_exhaust_thms = map (fn thm => thm RS exE) un_T_nchotomy_thms; *)

    val bij_T_thms =
      @{map 2} (fn thm1 => fn thm2 => @{thm o_bij} OF [thm1, thm2]) un_T_o_T_thms T_o_un_T_thms;
    val inj_T_thms = map (fn thm => thm RS @{thm bij_is_inj}) bij_T_thms;
    val T_inject_thms = map (fn thm => thm RS @{thm inj_eq}) inj_T_thms;
(*     val surj_T_thms = map (fn thm => thm RS @{thm bij_is_surj}) bij_T_thms;
    val T_nchotomy_thms = map (fn thm => thm RS @{thm surjD}) surj_T_thms;
    val T_exhaust_thms = map (fn thm => thm RS exE) T_nchotomy_thms; *)

    (* Characteristic Theorems *)

    (* map_T_eq *)

    val T_A's = map (Term.typ_subst_atomic (As ~~ A's)) T_As;
    val T_F'_restss = @{map 2} (fn T_A => map (fn F_A's =>
      Term.typ_subst_atomic (As ~~ F_A's) T_A)) T_As (rest_occs2 F_A'ss);
    val G_A'_Ts = @{map 2} (flip mk_T_of_bnf (A's @ flat T_F'_restss)) Dss Gs;
    val T_ctor'_vals = @{map 2} mk_const_val ctor_Ts (@{map 2} (curry op -->) G_A'_Ts T_A's);
    val map_Ts = map (mk_map_of_bnf resDs As A's) T_bnfs;
    val map_T_Fss =  @{map 3} (@{map 3} (fn F_Ts => fn F_T's =>
      mk_map_of_bnf resDs F_Ts F_T's)) (rest_occs2 F_Ass) (rest_occs2 F_A'ss) (rep_occs T_bnfs);
    val map_Gs = map (mk_map_of_bnf resDs (As @ flat T_F_restss) (A's @ flat T_F'_restss)) Gs;

    fun mk_map_T_eq f_vals map_Fss map_T_Fs map_T ctor_T g ctor_T' map_G =
      mk_Trueprop_eq (list_comb (map_T, f_vals) $ (ctor_T $ g),
      ctor_T' $ (list_comb (list_comb (map_G, f_vals),
      @{map 2} (fn map_T_F => fn map_Fs =>
      list_comb (map_T_F, map (list_comb o rpair f_vals) map_Fs))
      map_T_Fs map_Fss) $ g));

    val map_T_eq_goals = @{map 5} (mk_map_T_eq f_vals map_Fss (flat map_T_Fss))
      map_Ts ctor_T_vals G_T_F_vals T_ctor'_vals map_Gs;
    val map_T_eq_varss = map (fn g => map name_of (g :: f_vals)) G_T_F_vals;

    val map_T_eq_tactics = map (fn T_ctor_def =>
      (fn {context = ctxt, prems = _} =>
        mk_map_T_eq_tac ctxt T_map_defs T_ctor_def (map #Abs_inverse T_thms)
        raw_map_thms shape_map_thms (flat map_flat_raw_thmss) (map #Rep T_thms)
        invar_raw_simps invar_shape_simps invar_map_raw_closed_thms (flat invar_raw_flat_raw_thmss)
        G_pred_map_thms G_pred_True_thms G_map_comp_thms G_pred_cong_thms G_map_cong_thms))
      T_ctor_defs;

    val map_T_eq_thms = @{map 3} (fn vars => fn goal => fn tac =>
      Goal.prove_sorry lthy vars [] goal tac
      |> Thm.close_derivation)
      map_T_eq_varss map_T_eq_goals map_T_eq_tactics;

    (* set_T_eq *)

    val set_Tss = map (mk_sets_of_bnf (rep_par resDs) (rep_par As)) T_bnfs |> transpose;
    val set_T_Fss = @{map 2} (fn F_As => mk_sets_of_bnf (rep_par resDs) (rep_par F_As))
      F_Ass (rest_occs T_bnfs) ;
    val set_G_T_Fss = @{map 2} (fn Ds => mk_sets_of_bnf (replicate (nr_param + nr_occ) Ds)
      (replicate (nr_param + nr_occ) (As @ flat T_F_restss))) Dss Gs;
    val set_G_T_F1ss = map (take nr_param) set_G_T_Fss;
    val set_G_T_F2ss = map (drop nr_param) set_G_T_Fss;
    val set_Fsss = set_Fsss |> map transpose |> transpose
    fun mk_set_T_eq_sub A set_F F_A set_T_F set_G_T_F g =
      mk_Union A $ (mk_image set_F $ (mk_Union F_A $
      (mk_image set_T_F $ (set_G_T_F $ g))));
    fun mk_set_T_eq A F_Ass set_Fss set_T_Fss g ctor_T set_G_T_F2s set_T set_G_T_F1 =
      mk_Trueprop_eq (set_T $ (ctor_T $ g), mk_union (set_G_T_F1 $ g,
      Library.foldl1 mk_union (@{map 4} (fn set_G_T_F2 =>
      @{map 3} (fn set_F => fn F_A => fn set_T_F =>
      mk_set_T_eq_sub A set_F F_A set_T_F set_G_T_F2 g))
      set_G_T_F2s set_Fss F_Ass set_T_Fss |> flat)))

    val set_T_eq_goalss = @{map 4} (fn A => fn set_Fss => @{map 5} (mk_set_T_eq A F_Ass set_Fss set_T_Fss)
      G_T_F_vals ctor_T_vals set_G_T_F2ss) As set_Fsss set_Tss (transpose set_G_T_F1ss);
    val set_T_eq_varss = map (single o name_of) G_T_F_vals;

    val set_T_eq_tactics = map (fn T_ctor_def =>
      (fn {context = ctxt, prems = _} =>
        mk_set_T_eq_tac ctxt T_set_defs T_ctor_def
        (map #Abs_inverse T_thms) raw_set_thms shape_set_thms (flat (flat set_flat_raw_thmsss))
        (map #Rep T_thms) invar_raw_simps invar_shape_simps (flat invar_raw_flat_raw_thmss)
        G_pred_map_thms G_pred_True_thms G_set_map_thms G_pred_cong_thms))
      T_ctor_defs;

    val set_T_eq_thmss = map (@{map 3} (fn vars => fn tac => fn goal =>
      Goal.prove_sorry lthy vars [] goal tac
      |> Thm.close_derivation)
      set_T_eq_varss set_T_eq_tactics) set_T_eq_goalss;

    (* rel_T_eq *)

    val (G_T_F'_vals, _) = lthy
      |> mk_Frees "g_T_F'" G_A'_Ts
    val rel_Ts = map (mk_rel_of_bnf resDs As A's) T_bnfs;
    val rel_T_Fss =  @{map 3} (@{map 3} (fn F_Ts => fn F_T's =>
      mk_rel_of_bnf resDs F_Ts F_T's)) (rest_occs2 F_Ass) (rest_occs2 F_A'ss) (rep_occs T_bnfs);
    val rel_Gs = map (mk_rel_of_bnf resDs (As @ flat T_F_restss) (A's @ flat T_F'_restss)) Gs;

    fun mk_rel_T_eq Q_vals rel_Fss rel_T_Fs rel_T ctor_T g ctor_T' g' rel_G =
      mk_Trueprop_eq (list_comb (rel_T, Q_vals) $ (ctor_T $ g) $ (ctor_T' $ g'),
      (list_comb (list_comb (rel_G, Q_vals),
      @{map 2} (fn rel_T_F => fn rel_Fs =>
      list_comb (rel_T_F, map (list_comb o rpair Q_vals) rel_Fs))
      rel_T_Fs rel_Fss) $ g $ g'));

    val rel_T_eq_goals = @{map 6} (mk_rel_T_eq Q_vals rel_Fss (flat rel_T_Fss))
      rel_Ts ctor_T_vals G_T_F_vals T_ctor'_vals G_T_F'_vals rel_Gs;
    val rel_T_eq_varss = @{map 2} (fn g => fn g' => map name_of ([g, g'] @ Q_vals)) G_T_F_vals G_T_F'_vals;

    val rel_T_eq_tactics = map (fn T_ctor_def =>
      fn {context = ctxt, prems = _} =>
        mk_rel_T_eq_tac ctxt T_rel_defs T_ctor_def
        (map #Abs_inverse T_thms) raw_rel_inject_thms shape_rel_inject_thms
        (transpose rel_flat_raw_thmss) (map #Rep T_thms) invar_raw_simps invar_shape_simps
        invar_map_raw_closed_thms (flat invar_raw_flat_raw_thmss)
        G_pred_map_thms G_pred_True_thms G_rel_map_thms G_pred_cong_thms)
      T_ctor_defs;

    val rel_T_eq_thms = @{map 3} (fn vars => fn goal => fn tac =>
      Goal.prove_sorry lthy vars [] goal tac
      |> Thm.close_derivation)
      rel_T_eq_varss rel_T_eq_goals rel_T_eq_tactics;


    (* postprocessing co-theorems*)

    val map_T_eq_thms = (fp = Greatest_FP ?
      postproc_thms lthy T_As (SOME T_A's) dtor_T_vals (SOME dtor_T'_vals) (fn cts => #3 cts :: rep_par NONE @ [#1 cts])
      (T_un_T_thms @ un_T_T_thms) o map (fn thm => thm RS arg_cong)) map_T_eq_thms;
    val set_T_eq_thmss = map (fp = Greatest_FP ?
        postproc_thms lthy T_As NONE dtor_T_vals NONE (single o #1) T_un_T_thms)
      set_T_eq_thmss;
    val rel_T_eq_thms = (fp = Greatest_FP ?
      postproc_thms lthy T_As (SOME T_A's) dtor_T_vals (SOME dtor_T'_vals) (fn cts => rep_par NONE @ [#1 cts, #2 cts])
      T_un_T_thms) rel_T_eq_thms;


    (* Creating the final result *)

    val fp_res =
      {Ts = T_As, bnfs = T_bnfs, ctors = ctor_Ts, dtors = dtor_Ts,
       absT_infos = absT_infos, pre_bnfs = Gs,
       xtor_un_folds = [], xtor_co_recs = [], xtor_co_induct = TrueI,
       dtor_ctors = un_T_T_thms, ctor_dtors = T_un_T_thms, ctor_injects = T_inject_thms,
       dtor_injects = un_T_inject_thms, xtor_maps = map_T_eq_thms,
       xtor_map_unique = TrueI,
       xtor_setss = transpose set_T_eq_thmss,
       xtor_rels = rel_T_eq_thms,
       xtor_un_fold_thms = [], xtor_co_rec_thms = [], xtor_un_fold_unique = TrueI,
       xtor_co_rec_unique = TrueI, xtor_un_fold_o_maps = [], xtor_co_rec_o_maps = [],
       xtor_un_fold_transfers = [], xtor_co_rec_transfers = [], xtor_rel_co_induct = TrueI,
       dtor_set_inducts = []};

    val nu_fp_res =
      {As = As,
       Fss = Fss,
       Dsss = Dsss,
       Gs = Gs,
       Dss = Dss,
       resDs = resDs,
       raw_sugars = raw_sugars,
       nr_occs = nr_occs,

       invar_raws = invar_raw_vals,
       invar_raw_simps = invar_raw_simps,
       flat_rawss = flat_raw_valss,
       flat_raw_simpss = flat_raw_simpss,
       unflat_rawss = unflat_raw_valss,
       unflat_raw_simpss = unflat_raw_simpss,
       cimmerse_raws = cimmerse_raw_vals,
       cimmerse_raw_simpss = cimmerse_raw_simpss,

       T_infos = map snd Ts,
       T_ctor_defs = T_ctor_defs,
       T_dtor_defs = T_dtor_defs,

       shape_res = shape_res,
       fp_res = fp_res,

       invar_map_raw_closed_thms = invar_map_raw_closed_thms,

       flat_raw_unflat_raw_thmss = flat_raw_unflat_raw_thmss,
       unflat_raw_flat_raw_thmss = unflat_raw_flat_raw_thmss,
       invar_raw_flat_raw_thmss = invar_raw_flat_raw_thmss,
       invar_raw_unflat_raw_thmss = invar_raw_unflat_raw_thmss,

       invar_cimmerse_raw_thms = invar_cimmerse_raw_thms,
       unCons_cimmerse_raw_thms = unCons_cimmerse_raw_thms,

       map_flat_raw_thmss = map_flat_raw_thmss,
       set_flat_raw_thmsss = set_flat_raw_thmsss,
       rel_flat_raw_thmss = rel_flat_raw_thmss} : nu_fp_result;

    val (nu_fp_res, lthy) = note_nu_fp_result_thms
      (if Config.get lthy bnf_internals then Note_All else Dont_Note) I b fp nu_fp_res lthy;
  in
    (nu_fp_res, lthy)
  end;

val _ =
  Outer_Syntax.local_theory @{command_keyword nonuniform_datatype}
    "define nonuniform inductive datatypes"
    (parse_nonuniform_co_datatype_cmd Least_FP (construct_nu_fp Least_FP));

val _ = Theory.setup (fp_antiquote_setup @{binding nonuniform_datatype});

val _ =
  Outer_Syntax.local_theory @{command_keyword nonuniform_codatatype}
    "define nonuniform coinductive datatypes"
    (parse_nonuniform_co_datatype_cmd Greatest_FP (construct_nu_fp Greatest_FP));

val _ = Theory.setup (fp_antiquote_setup @{binding nonuniform_codatatype});

end;
