(*  Title:      HOL/Tools/BNF/bnf_nu_fp_def_sugar.ML
    Author:     Jasmin Blanchette, Inria, LORIA, MPII
    Copyright   2016

Sugared nonuniform datatype and codatatype constructions.
Inspired by "bnf_fp_def_sugar.ML".
*)

signature BNF_NU_FP_DEF_SUGAR =
sig
  type nu_fp_sugar =
    {T: typ,
     Xs: typ list,
     real_Ts: typ list,
     fp_sugar: BNF_FP_Def_Sugar.fp_sugar,
     nu_fp_res: BNF_NU_FP_Util.nu_fp_result,
     live_arg_bnfss: BNF_Def.bnf list list,
     live_arg_absT_infoss: BNF_Comp.absT_info list list,
     combined_absT_info: BNF_Comp.absT_info}

  val morph_nu_fp_sugar: morphism -> nu_fp_sugar -> nu_fp_sugar
  val transfer_nu_fp_sugar: theory -> nu_fp_sugar -> nu_fp_sugar
  val nu_fp_sugar_of: Proof.context -> string -> nu_fp_sugar option
  val nu_fp_sugar_of_global: theory -> string -> nu_fp_sugar option
  val nu_fp_sugars_of: Proof.context -> nu_fp_sugar list
  val nu_fp_sugars_of_global: theory -> nu_fp_sugar list
  val nu_fp_sugars_interpretation: string -> (nu_fp_sugar list -> local_theory -> local_theory) ->
    theory -> theory
  val interpret_nu_fp_sugars: (string -> bool) -> nu_fp_sugar list -> local_theory -> local_theory
  val register_nu_fp_sugars_raw: nu_fp_sugar list -> local_theory -> local_theory
  val register_nu_fp_sugars: (string -> bool) -> nu_fp_sugar list -> local_theory -> local_theory

  val nonuniform_co_datatypes: BNF_Util.fp_kind ->
    (mixfix list -> binding list -> binding list ->
     binding list -> binding option list list -> binding list -> (string * sort) list -> typ list ->
     ((typ list * BNF_Def.bnf) * (typ list * BNF_Def.bnf) list list) list ->
     BNF_Comp.absT_info list -> local_theory -> BNF_NU_FP_Util.nu_fp_result * local_theory) ->
    ((string -> bool) * bool)
    * ((((((binding option * (typ * sort)) list * binding) * mixfix)
      * ((binding, binding * typ) Ctr_Sugar.ctr_spec * mixfix) list)
        * (binding * binding * binding)) * term list) list ->
    local_theory -> local_theory
  val nonuniform_co_datatype_cmd: BNF_Util.fp_kind ->
    (mixfix list -> binding list -> binding list -> binding list -> binding option list list ->
     binding list -> (string * sort) list -> typ list ->
     ((typ list * BNF_Def.bnf) * (typ list * BNF_Def.bnf) list list) list ->
     BNF_Comp.absT_info list -> local_theory -> BNF_NU_FP_Util.nu_fp_result * local_theory) ->
    ((Proof.context -> Plugin_Name.filter) * bool)
    * ((((((binding option * (string * string option)) list * binding) * mixfix)
         * ((binding, binding * string) Ctr_Sugar.ctr_spec * mixfix) list)
        * (binding * binding * binding)) * string list) list ->
    Proof.context -> local_theory

  val parse_nonuniform_co_datatype_cmd: BNF_Util.fp_kind ->
    (mixfix list -> binding list -> binding list -> binding list -> binding option list list ->
     binding list -> (string * sort) list -> typ list ->
     ((typ list * BNF_Def.bnf) * (typ list * BNF_Def.bnf) list list) list ->
     BNF_Comp.absT_info list -> local_theory -> BNF_NU_FP_Util.nu_fp_result * local_theory) ->
    (local_theory -> local_theory) parser
end;

structure BNF_NU_FP_Def_Sugar : BNF_NU_FP_DEF_SUGAR =
struct

open Ctr_Sugar
open BNF_Util
open BNF_Tactics
open BNF_Def
open BNF_Comp
open BNF_FP_Util
open BNF_FP_Def_Sugar
open BNF_FP_Rec_Sugar_Transfer
open BNF_NU_FP_Util
open BNF_NU_FP_Sugar_Tactics

val argN = "arg";

type nu_fp_sugar =
  {T: typ,
   Xs: typ list,
   real_Ts: typ list,
   fp_sugar: BNF_FP_Def_Sugar.fp_sugar,
   nu_fp_res: BNF_NU_FP_Util.nu_fp_result,
   live_arg_bnfss: BNF_Def.bnf list list,
   live_arg_absT_infoss: BNF_Comp.absT_info list list,
   combined_absT_info: BNF_Comp.absT_info};

fun morph_nu_fp_sugar phi ({T, Xs, real_Ts, fp_sugar, nu_fp_res, live_arg_bnfss,
    live_arg_absT_infoss, combined_absT_info} : nu_fp_sugar) =
  {T = Morphism.typ phi T,
   Xs = map (Morphism.typ phi) Xs,
   real_Ts = map (Morphism.typ phi) real_Ts,
   fp_sugar = morph_fp_sugar phi fp_sugar,
   nu_fp_res = morph_nu_fp_result phi nu_fp_res,
   live_arg_bnfss = map (map (morph_bnf phi)) live_arg_bnfss,
   live_arg_absT_infoss = map (map (morph_absT_info phi)) live_arg_absT_infoss,
   combined_absT_info = morph_absT_info phi combined_absT_info};

val transfer_nu_fp_sugar = morph_nu_fp_sugar o Morphism.transfer_morphism;

structure Data = Generic_Data
(
  type T = nu_fp_sugar Symtab.table;
  val empty = Symtab.empty;
  val extend = I;
  fun merge data : T = Symtab.merge (K true) data;
);

fun nu_fp_sugar_of_generic context =
  Option.map (transfer_nu_fp_sugar (Context.theory_of context)) o Symtab.lookup (Data.get context);

fun nu_fp_sugars_of_generic context =
  Symtab.fold (cons o transfer_nu_fp_sugar (Context.theory_of context) o snd) (Data.get context) [];

val nu_fp_sugar_of = nu_fp_sugar_of_generic o Context.Proof;
val nu_fp_sugar_of_global = nu_fp_sugar_of_generic o Context.Theory;

val nu_fp_sugars_of = nu_fp_sugars_of_generic o Context.Proof;
val nu_fp_sugars_of_global = nu_fp_sugars_of_generic o Context.Theory;

structure NU_FP_Sugar_Plugin = Plugin(type T = nu_fp_sugar list);

fun nu_fp_sugars_interpretation name f =
  NU_FP_Sugar_Plugin.interpretation name (fn nu_fp_sugars => fn lthy =>
    f (map (transfer_nu_fp_sugar (Proof_Context.theory_of lthy)) nu_fp_sugars) lthy);

val interpret_nu_fp_sugars = NU_FP_Sugar_Plugin.data;

val register_nu_fp_sugars_raw =
  fold (fn nu_fp_sugar as {T = Type (s, _), ...} =>
    Local_Theory.declaration {syntax = false, pervasive = true}
      (fn phi => Data.map (Symtab.update (s, morph_nu_fp_sugar phi nu_fp_sugar))));

fun register_nu_fp_sugars plugins nu_fp_sugars =
  register_nu_fp_sugars_raw nu_fp_sugars #> interpret_nu_fp_sugars plugins nu_fp_sugars;

fun interpret_bnfs_register_nu_fp_sugars plugins Ts BTs Xss real_Tss fp pre_bnfs live_arg_bnfsss
    pre_absT_infos live_arg_absT_infosss combined_absT_infos fp_nesting_bnfs live_nesting_bnfs
    nu_fp_res ctrXs_Tsss ctor_iff_dtors ctr_defss ctr_sugars map_thmss rel_injectss rel_distinctss
    map_disc_iffss map_selsss rel_selss rel_intross rel_casess pred_injectss set_thmss set_selsssss
    set_introsssss set_casess ctr_transferss case_transferss disc_transferss sel_transferss =
  let
    val fp_res = #fp_res nu_fp_res;
    val fp_bnfs = #bnfs fp_res;

    val fp_sugars =
      map_index (fn (kk, T) =>
        {T = T, BT = nth BTs kk, X = hd (nth Xss kk), fp = fp, fp_res = fp_res, fp_res_index = kk,
         pre_bnf = nth pre_bnfs kk, absT_info = nth pre_absT_infos kk, fp_bnf = nth fp_bnfs kk,
         fp_nesting_bnfs = fp_nesting_bnfs, live_nesting_bnfs = live_nesting_bnfs,
         fp_ctr_sugar =
           {ctrXs_Tss = nth ctrXs_Tsss kk,
            ctor_iff_dtor = nth ctor_iff_dtors kk,
            ctr_defs = nth ctr_defss kk,
            ctr_sugar = nth ctr_sugars kk,
            ctr_transfers = nth ctr_transferss kk,
            case_transfers = nth case_transferss kk,
            disc_transfers = nth disc_transferss kk,
            sel_transfers = nth sel_transferss kk},
         fp_bnf_sugar =
           {map_thms = nth map_thmss kk,
            map_disc_iffs = nth map_disc_iffss kk,
            map_selss = nth map_selsss kk,
            rel_injects = nth rel_injectss kk,
            rel_distincts = nth rel_distinctss kk,
            rel_sels = nth rel_selss kk,
            rel_intros = nth rel_intross kk,
            rel_cases = nth rel_casess kk,
            pred_injects = nth pred_injectss kk,
            set_thms = nth set_thmss kk,
            set_selssss = nth set_selsssss kk,
            set_introssss = nth set_introsssss kk,
            set_cases = nth set_casess kk},
         fp_co_induct_sugar = NONE}) Ts;

    val nu_fp_sugars =
      map_index (fn (kk, T) =>
        {T = T, Xs = nth Xss kk, real_Ts = nth real_Tss kk, fp_sugar = nth fp_sugars kk,
         nu_fp_res = nu_fp_res, live_arg_bnfss = nth live_arg_bnfsss kk,
         live_arg_absT_infoss = nth live_arg_absT_infosss kk,
         combined_absT_info = nth combined_absT_infos kk}) Ts;
  in
    register_fp_sugars_raw fp_sugars
    #> register_nu_fp_sugars_raw nu_fp_sugars
    #> fold (interpret_bnf plugins) fp_bnfs
    #> interpret_fp_sugars plugins fp_sugars
    #> interpret_nu_fp_sugars plugins nu_fp_sugars
  end;

fun combined_absT_info_of ctxt pre_live live_As dead_As real_Ts live_arg_absT_infosss fp_bnfs
    pre_bnf ({abs = pre_abs0, rep = pre_rep0, abs_inverse = pre_abs_inverse,
      type_definition = pre_type_definition, ...} : absT_info) =
  let
    val pre_As = dead_As @ live_As;

    fun mk_map_abs fp_bnf live_arg_absT_infos =
      let
        val abss = map (mk_ctor pre_As o #abs) live_arg_absT_infos;
        val repTs = map (domain_type o fastype_of) abss;
        val absTs = map (range_type o fastype_of) abss;
      in
        list_comb (mk_map_of_bnf dead_As repTs absTs fp_bnf, abss)
      end;

    fun mk_map_rep fp_bnf live_arg_absT_infos =
      let
        val reps = map (mk_dtor pre_As o #rep) live_arg_absT_infos;
        val absTs = map (domain_type o fastype_of) reps;
        val repTs = map (range_type o fastype_of) reps;
      in
        list_comb (mk_map_of_bnf dead_As absTs repTs fp_bnf, reps)
      end;

    val ids = map HOLogic.id_const live_As;
    val map_abss = flat (map2 (map o mk_map_abs) fp_bnfs live_arg_absT_infosss);
    val (repTs, absTs) = split_list (map (dest_funT o fastype_of) map_abss);
    val abs_map = list_comb (mk_map pre_live (pre_As @ repTs) (pre_As @ absTs) (map_of_bnf pre_bnf),
      ids @ map_abss);
    val pre_abs = mk_ctor (pre_As @ real_Ts) pre_abs0;
    val abs = HOLogic.mk_comp (abs_map, pre_abs);

    val map_reps = flat (map2 (map o mk_map_rep) fp_bnfs live_arg_absT_infosss);
    val rep_map = list_comb (mk_map pre_live (pre_As @ absTs) (pre_As @ repTs) (map_of_bnf pre_bnf),
      ids @ map_reps);
    val pre_rep = mk_dtor (pre_As @ real_Ts) pre_rep0;
    val rep = HOLogic.mk_comp (pre_rep, rep_map);

    val (absT, repT) = dest_funT (fastype_of rep);

    val type_definition =
      let
        val univ = HOLogic.mk_UNIV repT;
        val type_def = Const (@{const_name type_definition},
          fastype_of rep --> fastype_of abs --> fastype_of univ --> HOLogic.boolT);

        val goal = HOLogic.mk_Trueprop (type_def $ rep $ abs $ univ);

        val live_arg_abs_inverses =
          flat (flat (map (map (map #abs_inverse)) live_arg_absT_infosss));
        val live_arg_type_definitions =
          flat (flat (map (map (map #type_definition)) live_arg_absT_infosss));
        val live_arg_rep_inverses =
          map (fn thm => thm RS @{thm type_definition.Rep_inverse}) live_arg_type_definitions;
        val pre_rep_inverse = pre_type_definition RS @{thm type_definition.Rep_inverse};
      in
        Goal.prove_sorry ctxt [] [] goal (fn {context = ctxt, prems = _} =>
          mk_combined_type_definition_tac ctxt live_arg_abs_inverses live_arg_rep_inverses
            pre_abs_inverse pre_rep_inverse (map_ident0_of_bnf pre_bnf) (map_comp_of_bnf pre_bnf)
            (map map_ident0_of_bnf fp_bnfs) (map map_comp_of_bnf fp_bnfs))
        |> Thm.close_derivation
      end;

    val abs_inject = type_definition RS @{thm type_definition.Abs_inject[OF _ UNIV_I UNIV_I]};
    val abs_inverse = type_definition RS @{thm type_definition.Abs_inverse[OF _ UNIV_I]};
  in
    {absT = absT, repT = repT, abs = abs, rep = rep, abs_inject = abs_inject,
     abs_inverse = abs_inverse, type_definition = type_definition}
  end;

fun define_nonuniform_co_datatypes prepare_plugins prepare_constraint prepare_typ prepare_term fp
    construct_nu_fp ((raw_plugins, discs_sels0), specs) lthy =
  let
    val plugins = prepare_plugins lthy raw_plugins;
    val discs_sels = discs_sels0 orelse fp = Greatest_FP;

    val nn = length specs;
    val fp_bs = map type_binding_of_spec specs;
    val fp_b_names = map Binding.name_of fp_bs;
    val map_bs = map map_binding_of_spec specs;
    val rel_bs = map rel_binding_of_spec specs;
    val pred_bs = map pred_binding_of_spec specs;

    fun prepare_type_arg (_, (ty, c)) =
      let val TFree (s, _) = prepare_typ lthy ty in
        TFree (s, prepare_constraint lthy c)
      end;

    val Ass0 = map (map prepare_type_arg o type_args_named_constrained_of_spec) specs;
    val unsorted_Ass0 = map (map (resort_tfree_or_tvar @{sort type})) Ass0;
    val unsorted_As = Library.foldr1 (merge_type_args fp) unsorted_Ass0;
    val num_As = length unsorted_As;

    val set_boss = map (map fst o type_args_named_constrained_of_spec) specs;
    val set_boss' = transpose set_boss;

    fun add_fake_type spec =
      Typedecl.basic_typedecl {final = true}
        (type_binding_of_spec spec, num_As, Mixfix.reset_pos (mixfix_of_spec spec));

    val (fake_T_names, fake_lthy) = fold_map add_fake_type specs lthy;

    val qsoty = quote o Syntax.string_of_typ fake_lthy;

    val _ = (case Library.duplicates (op =) unsorted_As of [] => ()
      | A :: _ => error ("Duplicate type parameter " ^ qsoty A ^ " in " ^ co_prefix fp ^
          "datatype specification"));

    val bad_args =
      map (Logic.type_map (singleton (Variable.polymorphic lthy))) unsorted_As
      |> filter_out Term.is_TVar;
    val _ = null bad_args orelse
      error ("Locally fixed type argument " ^ qsoty (hd bad_args) ^ " in " ^ co_prefix fp ^
        "datatype specification");

    val mixfixes = map mixfix_of_spec specs;

    val _ = (case Library.duplicates Binding.eq_name fp_bs of [] => ()
      | b :: _ => error ("Duplicate type name declaration " ^ quote (Binding.name_of b)));

    val mx_ctr_specss = map mixfixed_ctr_specs_of_spec specs;
    val ctr_specss = map (map fst) mx_ctr_specss;
    val ctr_mixfixess = map (map snd) mx_ctr_specss;

    val disc_bindingss = map (map disc_of_ctr_spec) ctr_specss;
    val ctr_bindingss =
      map2 (fn fp_b_name => map (Binding.qualify false fp_b_name o ctr_of_ctr_spec)) fp_b_names
        ctr_specss;
    val ctr_argsss = map (map args_of_ctr_spec) ctr_specss;

    val sel_bindingsss = map (map (map fst)) ctr_argsss;
    val fake_ctr_Tsss0 = map (map (map (prepare_typ fake_lthy o snd))) ctr_argsss;
    val raw_sel_default_eqss = map sel_default_eqs_of_spec specs;

    val (As :: _) :: fake_ctr_Tsss =
      burrow (burrow (Syntax.check_typs fake_lthy)) (Ass0 :: fake_ctr_Tsss0);
    val As' = map dest_TFree As;

    val rhs_As' = fold (fold (fold Term.add_tfreesT)) fake_ctr_Tsss [];
    val _ = (case subtract (op =) As' rhs_As' of [] => ()
      | extras => error ("Extra type variables on right-hand side: " ^
          commas (map (qsoty o TFree) extras)));

    fun filter_lives xs =
      map_filter (fn (x, set_bos) => if exists is_none set_bos then NONE else SOME x)
        (xs ~~ set_boss');
    fun filter_deads xs =
      map_filter (fn (x, set_bos) => if exists is_none set_bos then SOME x else NONE)
        (xs ~~ set_boss');

    val live_As = filter_lives As;
    val dead_As = filter_deads As;

    fun check_no_fp_typ T =
      if exists_subtype (fn Type (s, _) => member (op =) fake_T_names s | _ => false) T then
        error "Nested type constructor specification not supported"
      else
        ();

    fun maybe_add_tyargs T_name (Type (s, Ts)) =
        (if s = T_name then (List.app check_no_fp_typ Ts; insert (op =) Ts) else I)
        #> fold (maybe_add_tyargs T_name) Ts
      | maybe_add_tyargs _ _ = I;

    val arg_Tsss =
      map (fn T_name => fold (fold (fold (maybe_add_tyargs T_name))) fake_ctr_Tsss []) fake_T_names
      |> map (fn [] => [As] | Tss => Tss);

    val live_arg_Tsss = map (map filter_lives) arg_Tsss;
    val live_arg_Ts = flat (flat live_arg_Tsss);

    fun check_same_dead_args [] = ()
      | check_same_dead_args (T :: Ts) =
        if forall (curry (op =) T) Ts then () else error "Cannot change dead type argument";

    val dead_arg_Tsss = map (insert (op =) dead_As o map filter_deads) arg_Tsss;
    val _ = List.app (List.app check_same_dead_args o transpose) dead_arg_Tsss;

    val _ = not (null live_As) orelse
      error ("Nonuniform " ^ co_prefix fp ^ "datatype requires at least one live type variable");

    val fp_bs_flat = map2 (fn b => map_index (fn (i, Ts) => map_index (fn (j, _) =>
          Binding.suffix_name ("_" ^ string_of_int i ^ "_" ^ string_of_int j) b) Ts))
        fp_bs live_arg_Tsss
      |> flat |> flat;

    val num_arg_variantss = map length arg_Tsss;

    val fake_Tss = map2 (map o curry Type) fake_T_names arg_Tsss;
    val fake_Ts = flat fake_Tss;

    val ((((Bs0, Cs), Es), Xss), _) = lthy
      |> fold (Variable.declare_typ o resort_tfree_or_tvar dummyS) unsorted_As
      |> mk_TFrees num_As
      ||>> mk_TFrees nn
      ||>> mk_TFrees nn
      ||>> @{fold_map 2} (variant_tfrees oo replicate) num_arg_variantss fp_b_names;

    val Xs = flat Xss;

    fun freeze_fp (T as Type (s, Ts)) =
        (case find_index (curry (op =) T) fake_Ts of
          ~1 => Type (s, map freeze_fp Ts)
        | kk => nth Xs kk)
      | freeze_fp T = T;

    val unfreeze_fp = Term.typ_subst_atomic (Xs ~~ fake_Ts);

    val ctrXs_Tsss = map (map (map freeze_fp)) fake_ctr_Tsss;
    val ctrXs_repTs = map mk_sumprodT_balanced ctrXs_Tsss;

    fun handle_bad_dead (V, X_backdrop as Type (bad_tc, _)) =
      let
        val fake_T = qsoty (unfreeze_fp V);
        val fake_T_backdrop = qsoty (unfreeze_fp X_backdrop);
      in
        if member (op =) Xs V then
          error ("Inadmissible or unsupported occurrence of type " ^ fake_T ^
            " via type constructor " ^ quote bad_tc ^ " in type expression " ^ fake_T_backdrop)
        else
          error ("Dead type variable " ^ fake_T ^ " must be declared as such")
      end;

    val (((live_arg_bnfs, live_arg_absT_infos), _), lthy) = lthy
      |> fixpoint_bnf true (Binding.qualify true argN) (K (K (K (K (K I))))) fp_bs_flat []
        (map dest_TFree dead_As) (map dest_TFree live_As) live_arg_Ts empty_comp_cache
      handle BAD_DEAD bad => handle_bad_dead bad;

    val live_arg_bnfsss = unflatt live_arg_Tsss live_arg_bnfs;
    val live_arg_absT_infosss = unflatt live_arg_Tsss live_arg_absT_infos;

    val lthy = lthy |> set_transfer_rule_attrs (map map_transfer_of_bnf live_arg_bnfs);

    val (((pre_bnfs, pre_absT_infos), _), lthy) = lthy
      |> fixpoint_bnf true I (K (K (K (K (K I))))) fp_bs [] (map dest_TFree dead_As)
        (map dest_TFree (live_As @ Xs)) ctrXs_repTs empty_comp_cache
      handle BAD_DEAD bad => handle_bad_dead bad;

    val lthy = lthy |> set_transfer_rule_attrs (map map_transfer_of_bnf pre_bnfs);

    val fp_specs = map2 (fn pre_bnf => fn arg_bnfss =>
        ((dead_As, pre_bnf), map (map (pair dead_As)) (transpose arg_bnfss)))
      pre_bnfs live_arg_bnfsss;

(* FIXME: Debug
    fun tuple_of_bnf bnf = (T_of_bnf bnf, "deads:", deads_of_bnf bnf, "lives:", lives_of_bnf bnf);

    val fp_specs' = @{map 3} (fn pre_bnf => fn arg_Tss => fn arg_bnfss =>
        ((dead_As, tuple_of_bnf pre_bnf),
         map (map (pair dead_As)) (transpose (map (map tuple_of_bnf) arg_bnfss))))
      pre_bnfs live_arg_Tsss live_arg_bnfsss;

    val _ = tracing ("fp_specs = " ^ @{make_string} fp_specs');
*)

    val (nu_fp_res as {fp_res = {bnfs = fp_bnfs as any_fp_bnf :: _, ctors = ctors0, dtors = dtors0,
           ctor_dtors, dtor_ctors, ctor_injects, xtor_maps, xtor_setss, xtor_rels, ...}, ...},
         lthy) =
      construct_nu_fp mixfixes map_bs rel_bs pred_bs set_boss fp_bs (map dest_TFree As) dead_As
        fp_specs pre_absT_infos lthy;

    val time = time lthy;
    val timer = time (Timer.startRealTimer ());

    val fp_nesting_bnfs = nesting_bnfs lthy ctrXs_Tsss Xs;
    val live_nesting_bnfs = nesting_bnfs lthy ctrXs_Tsss As;
    val fp_nested_bnfs = nesting_bnfs lthy live_arg_Tsss As;

    val pre_map_defs = map map_def_of_bnf pre_bnfs;
    val pre_set_defss = map set_defs_of_bnf pre_bnfs;
    val pre_rel_defs = map rel_def_of_bnf pre_bnfs;
    val fp_nesting_set_maps = maps set_map_of_bnf fp_nesting_bnfs;
    val fp_nesting_rel_eq_onps = map rel_eq_onp_of_bnf fp_nesting_bnfs;
    val live_nesting_map_id0s = map map_id0_of_bnf live_nesting_bnfs;
    val live_nesting_set_maps = maps set_map_of_bnf live_nesting_bnfs;
    val live_nesting_rel_eqs = map rel_eq_of_bnf live_nesting_bnfs;
    val live_nesting_rel_eq_onps = map rel_eq_onp_of_bnf live_nesting_bnfs;
    val fp_nested_rel_eq_onps = map rel_eq_onp_of_bnf fp_nested_bnfs;

    val liveness = liveness_of_fp_bnf num_As any_fp_bnf;
    val live = live_of_bnf any_fp_bnf;
    val pre_live = live + Integer.sum num_arg_variantss;

    val Bs = @{map 3} (fn alive => fn A as TFree (_, S) => fn B =>
        if alive then resort_tfree_or_tvar S B else A)
      liveness As Bs0;

    val B_ify_T = Term.typ_subst_atomic (As ~~ Bs);

    val ctors = map (mk_ctor As) ctors0;
    val dtors = map (mk_dtor As) dtors0;

    val fpTs = map (domain_type o fastype_of) dtors;
    val fpBTs = map B_ify_T fpTs;

    val real_Tss = map2 (map o (fn Type (s, _) => fn Ts => Type (s, Ts))) fpTs arg_Tsss;
    val real_Ts = flat real_Tss;

    val real_unfreeze_fp = Term.typ_subst_atomic (Xs ~~ real_Ts);

    val ctr_Tsss = map (map (map real_unfreeze_fp)) ctrXs_Tsss;
    val ns = map length ctr_Tsss;
    val kss = map (fn n => 1 upto n) ns;
    val mss = map (map length) ctr_Tsss;

    val combined_absT_infos =
      map2 (combined_absT_info_of lthy pre_live live_As dead_As real_Ts live_arg_absT_infosss
        fp_bnfs)
      pre_bnfs pre_absT_infos;

    val abss = map #abs combined_absT_infos;
    val abs_injects = map #abs_inject combined_absT_infos;
    val abs_inverses = map #abs_inverse combined_absT_infos;
    val type_definitions = map #type_definition combined_absT_infos;

    val SOME sum_bnf = bnf_of lthy @{type_name sum};
    val SOME prod_bnf = bnf_of lthy @{type_name prod};

    fun define_ctrs_dtrs_for_type_etc fp_bnf fp_b fpT C E ctor dtor ctor_dtor dtor_ctor ctor_inject
        pre_map_def pre_set_defs pre_rel_def fp_map_thm fp_set_thms fp_rel_thm n ks ms abs
        abs_inject type_definition ctr_bindings ctr_mixfixes ctr_Tss disc_bindings sel_bindingss
        raw_sel_default_eqs lthy =
      let
        val fp_b_name = Binding.name_of fp_b;

        val ((_, ctrs0, ctor_iff_dtor_thm, ctr_defs), lthy) =
          define_ctrs_dtrs_for_type fp_b_name fpT ctor dtor ctor_dtor dtor_ctor n ks abs
            ctr_bindings ctr_mixfixes ctr_Tss lthy;
        val ctr_defs' = map (unfold_thms lthy [o_apply]) ctr_defs;

        val sel_default_eqs =
          let
            val sel_Tss = map (map (curry (op -->) fpT)) ctr_Tss;
            val sel_bTs =
              flat sel_bindingss ~~ flat sel_Tss
              |> filter_out (Binding.is_empty o fst)
              |> distinct (Binding.eq_name o apply2 fst);
            val sel_default_lthy = fake_local_theory_for_sel_defaults sel_bTs lthy
          in
            map (prepare_term sel_default_lthy) raw_sel_default_eqs
          end;

        fun abs_inverse_variants_of ({abs_inverse, ...} : absT_info) =
          let val abs_inverse0 = mk_pointfree_id lthy abs_inverse in
            [abs_inverse, abs_inverse0, abs_inverse0 RS @{thm rewriteL_comp_comp},
             abs_inverse0 RS @{thm rewriteR_comp_comp}]
          end;

        fun map_comp_variants_of bnf =
          let
            val map_comp = map_comp_of_bnf bnf;
            val map_comp0' = map_comp0_of_bnf bnf RS sym;
          in
            [map_comp, map_comp0', map_comp0' RS @{thm rewriteL_comp_comp},
             map_comp0' RS @{thm rewriteR_comp_comp}]
          end;

        val extra_unfolds_map =
          map map_def_of_bnf live_arg_bnfs @
          maps abs_inverse_variants_of (pre_absT_infos @ live_arg_absT_infos) @
          maps map_comp_variants_of (fp_nesting_bnfs @ fp_bnfs);
        val extra_unfolds_set =
          maps set_defs_of_bnf live_arg_bnfs @ map map_def_of_bnf pre_bnfs @
          maps set_map_of_bnf fp_bnfs @
          maps abs_inverse_variants_of (pre_absT_infos @ live_arg_absT_infos) @
          map map_id0_of_bnf (sum_bnf :: prod_bnf :: live_nesting_bnfs) @
          @{thms UN_empty2 UN_singleton UN_simps(10) id_apply image_id};
        val extra_unfolds_rel =
          map rel_def_of_bnf live_arg_bnfs @ map map_def_of_bnf pre_bnfs @
          maps abs_inverse_variants_of (pre_absT_infos @ live_arg_absT_infos) @
          map map_id0_of_bnf (sum_bnf :: prod_bnf :: live_nesting_bnfs) @
          maps (map mk_abs_def o rel_map_of_bnf) (fp_nesting_bnfs @ fp_bnfs) @
          @{thms map_sum.simps map_prod_simp};

        fun massage_res (ctr_sugar, maps_sets_rels) =
          (maps_sets_rels, (ctor_iff_dtor_thm, ctr_defs, ctr_sugar));
      in
        (wrap_ctrs plugins fp discs_sels fp_b_name ctor_inject n ms abs_inject type_definition
           disc_bindings sel_bindingss sel_default_eqs ctrs0 ctor_iff_dtor_thm ctr_defs
         #> (fn (ctr_sugar, lthy) =>
           derive_map_set_rel_pred_thms plugins fp live As Bs C E abs_inverses ctr_defs'
             fp_nesting_set_maps fp_nesting_rel_eq_onps live_nesting_map_id0s live_nesting_set_maps
             live_nesting_rel_eqs live_nesting_rel_eq_onps fp_nested_rel_eq_onps fp_b_name fp_bnf
             fp_bnfs fpT ctor ctor_dtor dtor_ctor pre_map_def pre_set_defs pre_rel_def fp_map_thm
             fp_set_thms fp_rel_thm extra_unfolds_map extra_unfolds_set extra_unfolds_rel ctr_Tss
             ctr_sugar lthy
           |>> pair ctr_sugar)
         #>> massage_res, lthy)
      end;

    fun wrap_ctrs_derive_map_set_rel_pred_thms_for_types (wrap_one_etcs, lthy) =
      fold_map I wrap_one_etcs lthy
      |>> apsnd @{split_list 3} o apfst @{split_list 17} o split_list;

    fun interpret_bnfs_etc
        (((map_thmss, map_disc_iffss, map_selsss, rel_injectss, rel_distinctss, rel_selss,
           rel_intross, rel_casess, pred_injectss, set_thmss, set_selsssss, set_introsssss,
           set_casess, ctr_transferss, case_transferss, disc_transferss, sel_transferss),
          (ctor_iff_dtors, ctr_defss, ctr_sugars)), lthy) =
      interpret_bnfs_register_nu_fp_sugars plugins fpTs fpBTs Xss real_Tss fp pre_bnfs
        live_arg_bnfsss pre_absT_infos live_arg_absT_infosss combined_absT_infos fp_nesting_bnfs
        live_nesting_bnfs nu_fp_res ctrXs_Tsss ctor_iff_dtors ctr_defss ctr_sugars map_thmss
        rel_injectss rel_distinctss map_disc_iffss map_selsss rel_selss rel_intross rel_casess
        pred_injectss set_thmss set_selsssss set_introsssss set_casess ctr_transferss
        case_transferss disc_transferss sel_transferss lthy;

    val lthy = lthy
      |> fold2 (fn Type (s, _) => fn bnf => register_bnf_raw s bnf) fpTs fp_bnfs
      |> @{fold_map 28} define_ctrs_dtrs_for_type_etc fp_bnfs fp_bs fpTs Cs Es ctors dtors
        ctor_dtors dtor_ctors ctor_injects pre_map_defs pre_set_defss pre_rel_defs xtor_maps
        xtor_setss xtor_rels ns kss mss abss abs_injects type_definitions ctr_bindingss
        ctr_mixfixess ctr_Tsss disc_bindingss sel_bindingsss raw_sel_default_eqss
      |> wrap_ctrs_derive_map_set_rel_pred_thms_for_types
      |> interpret_bnfs_etc;

    val timer = time (timer ("Constructors, discriminators, selectors, etc., for the new " ^
      co_prefix fp ^ "datatype"));
  in
    lthy
  end;

fun nonuniform_co_datatypes fp = define_nonuniform_co_datatypes (K I) (K I) (K I) (K I) fp;

fun nonuniform_co_datatype_cmd fp construct_fp options lthy =
  define_nonuniform_co_datatypes Plugin_Name.make_filter Typedecl.read_constraint Syntax.parse_typ
    Syntax.parse_term fp construct_fp options lthy
  handle EMPTY_DATATYPE _ => error "Cannot define empty datatype";

fun parse_nonuniform_co_datatype_cmd fp construct_nu_fp =
  parse_co_datatype >> nonuniform_co_datatype_cmd fp construct_nu_fp;

end;
