(*  Title:      HOL/Tools/BNF/bnf_nu_fp_util.ML
    Author:     Dmitriy Traytel, ETH Zuerich
    Author:     Fabian Meier, ETH Zuerich
    Author:     Jasmin Blanchette, Inria, LORIA, MPII
    Copyright   2016

Shared library for the nonuniform datatype and codatatype constructions.
*)

signature BNF_NU_FP_UTIL =
sig
  val mk_rev: term -> term
  val mk_pointfree_id: Proof.context -> thm -> thm

  type shape_result =
    {shape_sugars: BNF_FP_Def_Sugar.fp_sugar list,
     label_sugar: BNF_FP_Def_Sugar.fp_sugar,
     labelT: typ,
     snoc_val: term,

     invar_shape_vals: term list,
     invar_shape_simps: thm list,
     flat_shape_valss: term list list,
     flat_shape_simpss: thm list list,
     flat_shape_transfer_thmss: thm list list,
     unflat_shape_valss: term list list,
     unflat_shape_simpss: thm list list,
     immerse_shape_valss: term list list,
     cimmerse_shape_vals: term list,
     cimmerse_shape_simpss: thm list list,

     invar_shape_depth_iff_thms: thm list,
     invar_shape_depth_iff1_thms: thm list,
     invar_map_shape_closed_thms: thm list,
     invar_shape_flat_shape_thmss: thm list list,
     invar_shape_unflat_shape_thmss: thm list list,
     flat_shape_unflat_shape_thmss: thm list list,
     unflat_shape_flat_shape_thmss: thm list list,
     cimmerse_shape_inj_thms: thm list,
     map_flat_shape_thmss: thm list list,
     set_flat_shape_thmsss: thm list list list,
     rel_flat_shape_thmss: thm list list}

  type nu_fp_result =
    {As: typ list,
     Fss: BNF_Def.bnf list list,
     Dsss: typ list list list,
     Gs: BNF_Def.bnf list,
     Dss: typ list list,
     resDs: typ list,
     raw_sugars: BNF_FP_Def_Sugar.fp_sugar list,
     nr_occs: int list,

     invar_raws: term list,
     invar_raw_simps: thm list,
     flat_rawss: term list list,
     flat_raw_simpss: thm list list,
     unflat_rawss: term list list,
     unflat_raw_simpss: thm list list,
     cimmerse_raws: term list,
     cimmerse_raw_simpss: thm list list,
     T_infos: Typedef.info list,
     T_ctor_defs: thm list,
     T_dtor_defs: thm list,

     shape_res: shape_result,
     fp_res: BNF_FP_Util.fp_result,

     invar_map_raw_closed_thms: thm list,

     flat_raw_unflat_raw_thmss: thm list list,
     unflat_raw_flat_raw_thmss: thm list list,
     invar_raw_flat_raw_thmss: thm list list,
     invar_raw_unflat_raw_thmss: thm list list,
     invar_cimmerse_raw_thms: thm list,
     unCons_cimmerse_raw_thms: thm list,

     map_flat_raw_thmss: thm list list,
     set_flat_raw_thmsss: thm list list list,
     rel_flat_raw_thmss: thm list list}

  val morph_shape_result: morphism -> shape_result -> shape_result
  val morph_nu_fp_result: morphism -> nu_fp_result -> nu_fp_result

  val mk_shapeN: int -> string;
  val mk_shape_NodeN: int -> string;
  val shape_LeafN: string;
  val mk_flat_shapeN: int -> int -> string;
  val mk_unflat_shapeN: int -> int -> string;
  val mk_cimmerse_shapeN: int -> string;
  val mk_invar_shapeN: int -> string;
  val labelN: string;
  val mk_label_CtorN: int -> string;

  val rawN: string;
  val raw_CtorN: string;
  val raw_DtorN: string;
  val mk_flat_rawN: int -> int -> string;
  val mk_unflat_rawN: int -> int -> string;
  val cimmerse_rawN: string;
  val invar_rawN: string;
  val wit_rawN: string;
  val ctor_T_N: string;
  val dtor_T_N: string;

  val note_nu_fp_result_thms: BNF_Def.fact_policy -> (binding -> binding) -> binding ->
    BNF_Util.fp_kind -> nu_fp_result -> local_theory -> nu_fp_result * local_theory
end;

structure BNF_NU_FP_Util : BNF_NU_FP_UTIL =
struct

open BNF_Util
open BNF_Tactics
open BNF_Def
open BNF_FP_Util
open BNF_FP_Def_Sugar

fun mk_rev xs =
  let
    val T = fastype_of xs;
  in
    Const (@{const_name rev}, T --> T) $ xs
  end

(*transforms f (g x) = x into f o g = id using first order matches for f and g*)
fun mk_pointfree_id ctxt thm =
  let
    val (f_o_g, x_T) = thm
      |> Thm.prop_of |> HOLogic.dest_Trueprop |> HOLogic.dest_eq
      |>> (dest_comb #> apsnd (dest_comb #> fst) #> HOLogic.mk_comp)
      ||> fastype_of;
  in
    mk_Trueprop_eq (f_o_g, HOLogic.id_const x_T)
    |> (fn goal => Goal.prove_sorry ctxt [] [] goal
      (K (rtac ctxt ext 1 THEN unfold_thms_tac ctxt ([id_apply, o_apply, thm]) THEN
          rtac ctxt refl 1)))
    |> Thm.close_derivation
  end;

type shape_result =
  {shape_sugars: BNF_FP_Def_Sugar.fp_sugar list,
   label_sugar: BNF_FP_Def_Sugar.fp_sugar,
   labelT: typ,
   snoc_val: term,

   invar_shape_vals: term list,
   invar_shape_simps: thm list,
   flat_shape_valss: term list list,
   flat_shape_simpss: thm list list,
   flat_shape_transfer_thmss: thm list list,
   unflat_shape_valss: term list list,
   unflat_shape_simpss: thm list list,
   immerse_shape_valss: term list list,
   cimmerse_shape_vals: term list,
   cimmerse_shape_simpss: thm list list,

   invar_shape_depth_iff_thms: thm list,
   invar_shape_depth_iff1_thms: thm list,
   invar_map_shape_closed_thms: thm list,
   invar_shape_flat_shape_thmss: thm list list,
   invar_shape_unflat_shape_thmss: thm list list,
   flat_shape_unflat_shape_thmss: thm list list,
   unflat_shape_flat_shape_thmss: thm list list,
   cimmerse_shape_inj_thms: thm list,
   map_flat_shape_thmss: thm list list,
   set_flat_shape_thmsss: thm list list list,
   rel_flat_shape_thmss: thm list list};


type nu_fp_result =
  {As: typ list,
   Fss: BNF_Def.bnf list list,
   Dsss: typ list list list,
   Gs: BNF_Def.bnf list,
   Dss: typ list list,
   resDs: typ list,
   raw_sugars: BNF_FP_Def_Sugar.fp_sugar list,
   nr_occs: int list,

   invar_raws: term list,
   invar_raw_simps: thm list,
   flat_rawss: term list list,
   flat_raw_simpss: thm list list,
   unflat_rawss: term list list,
   unflat_raw_simpss: thm list list,
   cimmerse_raws: term list,
   cimmerse_raw_simpss: thm list list,
   T_infos: Typedef.info list,
   T_ctor_defs: thm list,
   T_dtor_defs: thm list,

   shape_res: shape_result,
   fp_res: BNF_FP_Util.fp_result,

   invar_map_raw_closed_thms: thm list,

   flat_raw_unflat_raw_thmss: thm list list,
   unflat_raw_flat_raw_thmss: thm list list,
   invar_raw_flat_raw_thmss: thm list list,
   invar_raw_unflat_raw_thmss: thm list list,
   invar_cimmerse_raw_thms: thm list,
   unCons_cimmerse_raw_thms: thm list,

   map_flat_raw_thmss: thm list list,
   set_flat_raw_thmsss: thm list list list,
   rel_flat_raw_thmss: thm list list}

fun morph_shape_result phi {shape_sugars, label_sugar, labelT, snoc_val,
    invar_shape_vals, invar_shape_simps, flat_shape_valss, flat_shape_simpss,
    flat_shape_transfer_thmss, unflat_shape_valss, unflat_shape_simpss,
    immerse_shape_valss, cimmerse_shape_vals, cimmerse_shape_simpss,
    invar_shape_depth_iff_thms, invar_shape_depth_iff1_thms, invar_map_shape_closed_thms,
    invar_shape_flat_shape_thmss, invar_shape_unflat_shape_thmss, flat_shape_unflat_shape_thmss,
    unflat_shape_flat_shape_thmss, cimmerse_shape_inj_thms,
    map_flat_shape_thmss, set_flat_shape_thmsss, rel_flat_shape_thmss} =
  {shape_sugars = map (morph_fp_sugar phi) shape_sugars,
   label_sugar = morph_fp_sugar phi label_sugar,
   labelT = Morphism.typ phi labelT,
   snoc_val = Morphism.term phi snoc_val,

   invar_shape_vals = map (Morphism.term phi) invar_shape_vals,
   invar_shape_simps = map (Morphism.thm phi) invar_shape_simps,
   flat_shape_valss = map (map (Morphism.term phi)) flat_shape_valss,
   flat_shape_simpss = map (map (Morphism.thm phi)) flat_shape_simpss,
   flat_shape_transfer_thmss = map (map (Morphism.thm phi)) flat_shape_transfer_thmss,
   unflat_shape_valss = map (map (Morphism.term phi)) unflat_shape_valss,
   unflat_shape_simpss = map (map (Morphism.thm phi)) unflat_shape_simpss,
   immerse_shape_valss = map (map (Morphism.term phi)) immerse_shape_valss,
   cimmerse_shape_vals = map (Morphism.term phi) cimmerse_shape_vals,
   cimmerse_shape_simpss = map (map (Morphism.thm phi)) cimmerse_shape_simpss,

   invar_shape_depth_iff_thms = map (Morphism.thm phi) invar_shape_depth_iff_thms,
   invar_shape_depth_iff1_thms = map (Morphism.thm phi) invar_shape_depth_iff1_thms,
   invar_map_shape_closed_thms = map (Morphism.thm phi) invar_map_shape_closed_thms,
   invar_shape_flat_shape_thmss = map (map (Morphism.thm phi)) invar_shape_flat_shape_thmss,
   invar_shape_unflat_shape_thmss = map (map (Morphism.thm phi)) invar_shape_unflat_shape_thmss,
   flat_shape_unflat_shape_thmss = map (map (Morphism.thm phi)) flat_shape_unflat_shape_thmss,
   unflat_shape_flat_shape_thmss = map (map (Morphism.thm phi)) unflat_shape_flat_shape_thmss,
   cimmerse_shape_inj_thms = map (Morphism.thm phi) cimmerse_shape_inj_thms,
   map_flat_shape_thmss = map (map (Morphism.thm phi)) map_flat_shape_thmss,
   set_flat_shape_thmsss = map (map (map (Morphism.thm phi))) set_flat_shape_thmsss,
   rel_flat_shape_thmss = map (map (Morphism.thm phi)) rel_flat_shape_thmss
  } : shape_result;

fun morph_nu_fp_result phi {As, Fss, Dsss, Gs, Dss, resDs,
    raw_sugars, nr_occs,
    invar_raws, invar_raw_simps, flat_rawss, flat_raw_simpss, unflat_rawss, unflat_raw_simpss,
    cimmerse_raws, cimmerse_raw_simpss,
    T_infos, T_ctor_defs, T_dtor_defs,
    fp_res, shape_res, invar_map_raw_closed_thms,
    flat_raw_unflat_raw_thmss, unflat_raw_flat_raw_thmss, invar_raw_flat_raw_thmss,
    invar_raw_unflat_raw_thmss, invar_cimmerse_raw_thms, unCons_cimmerse_raw_thms,
    map_flat_raw_thmss, set_flat_raw_thmsss, rel_flat_raw_thmss} =
  {As = map (Morphism.typ phi) As,
   Fss = map (map (morph_bnf phi)) Fss,
   Dsss = map (map (map (Morphism.typ phi))) Dsss,
   Gs = map (morph_bnf phi) Gs,
   Dss = map (map (Morphism.typ phi)) Dss,
   resDs = map (Morphism.typ phi) resDs,
   raw_sugars = map (morph_fp_sugar phi) raw_sugars,
   nr_occs = nr_occs,

   invar_raws = map (Morphism.term phi) invar_raws,
   invar_raw_simps = map (Morphism.thm phi) invar_raw_simps,
   flat_rawss = map (map (Morphism.term phi)) flat_rawss,
   flat_raw_simpss = map (map (Morphism.thm phi)) flat_raw_simpss,
   unflat_rawss = map (map (Morphism.term phi)) unflat_rawss,
   unflat_raw_simpss = map (map (Morphism.thm phi)) unflat_raw_simpss,
   cimmerse_raws = map (Morphism.term phi) cimmerse_raws,
   cimmerse_raw_simpss = map (map (Morphism.thm phi)) cimmerse_raw_simpss,

   T_infos = map (Typedef.transform_info phi) T_infos,
   T_ctor_defs = map (Morphism.thm phi) T_ctor_defs,
   T_dtor_defs = map (Morphism.thm phi) T_dtor_defs,

   shape_res = morph_shape_result phi shape_res,
   fp_res = morph_fp_result phi fp_res,

   invar_map_raw_closed_thms = map (Morphism.thm phi) invar_map_raw_closed_thms,
   flat_raw_unflat_raw_thmss = map (map (Morphism.thm phi)) flat_raw_unflat_raw_thmss,
   unflat_raw_flat_raw_thmss = map (map (Morphism.thm phi)) unflat_raw_flat_raw_thmss,
   invar_raw_flat_raw_thmss = map (map (Morphism.thm phi)) invar_raw_flat_raw_thmss,
   invar_raw_unflat_raw_thmss = map (map (Morphism.thm phi)) invar_raw_unflat_raw_thmss,
   invar_cimmerse_raw_thms = map (Morphism.thm phi) invar_cimmerse_raw_thms,
   unCons_cimmerse_raw_thms = map (Morphism.thm phi) unCons_cimmerse_raw_thms,

   map_flat_raw_thmss = map (map (Morphism.thm phi))  map_flat_raw_thmss,
   set_flat_raw_thmsss = map (map (map (Morphism.thm phi)))  set_flat_raw_thmsss,
   rel_flat_raw_thmss = map (map (Morphism.thm phi))  rel_flat_raw_thmss} : nu_fp_result;

  (* Names *)

infix 6 ^#;
fun a ^# b = a ^ string_of_int b;
fun mk_shapeN i = "shape_" ^# i;
fun mk_shape_NodeN i = "node_" ^# i;
val shape_LeafN = "leaf";
fun mk_flat_shapeN i j = "flat" ^# j ^ "_shape_" ^# i;
fun mk_unflat_shapeN i j = "unflat" ^# j ^ "_shape_" ^# i;
fun mk_cimmerse_shapeN i = "cimmerse_shape_" ^# i;
fun mk_flat_shape_simpsN j = "flat" ^# j ^ "_shape_simps";
fun mk_unflat_shape_simpsN j = "unflat" ^# j ^ "_shape_simps";
val cimmerse_shape_simpsN = "cimmerse_shape_simps";
fun mk_invar_shapeN i = "invar_shape_" ^# i;
val invar_shape_simpsN = "invar_shape_simps";
val labelN = "label";
fun mk_label_CtorN i = "L_" ^# i;

val invar_shape_depth_iffN = "invar_shape_depth_iff";
val invar_map_shape_closedN = "invar_map_shape_closed";
fun mk_invar_shape_flat_shapeN i = "invar_shape_flat" ^# i ^ "_shape";
fun mk_invar_shape_unflat_shapeN i = "invar_shape_unflat" ^# i ^ "_shape";
fun mk_flat_shape_unflat_shapeN i = "flat" ^# i ^ "_shape_unflat" ^# i ^ "_shape";
fun mk_unflat_shape_flat_shapeN i = "unflat" ^# i ^ "_shape_flat" ^# i ^ "_shape";
val cimmerse_shape_injN = "cimmerse_shape_inj";
fun mk_map_flat_shapeN i = "map_flat" ^# i ^ "_shape";
fun mk_set_flat_shapeN j i = "set" ^# j ^ "_flat" ^# i ^ "_shape";
fun mk_rel_flat_shapeN i = "rel_flat" ^# i ^ "_shape";

val rawN = "raw";
val raw_CtorN = "Ctor";
val raw_DtorN = "Dtor";
fun mk_flat_rawN i j = "flat" ^# j ^ "_raw_" ^# i;
fun mk_unflat_rawN i j = "unflat" ^# j ^ "_raw_" ^# i;
val cimmerse_rawN = "cimmerse_raw";
fun mk_flat_raw_simpsN j = "flat" ^# j ^ "_raw_simps";
fun mk_unflat_raw_simpsN j = "unflat" ^# j ^ "_raw_simps";
val cimmerse_raw_simpsN = "cimmerse_raw_simps";
val invar_rawN = "invar_raw";
val invar_raw_simpsN = "invar_raw_simps";
val wit_rawN = "wit_raw";
val ctor_T_N = "Ctor";
val dtor_T_N = "Dtor";

val invar_map_raw_closedN = "invar_map_raw_closed";
fun mk_flat_raw_unflat_rawN i = "flat" ^# i ^ "_raw_unflat" ^# i ^ "_raw";
fun mk_unflat_raw_flat_rawN i = "unflat" ^# i ^ "_raw_flat" ^# i ^ "_raw";
fun mk_invar_raw_flat_rawN i = "invar_raw_flat" ^# i ^ "_raw";
fun mk_invar_raw_unflat_rawN i = "invar_raw_unflat" ^# i ^ "_raw";
val invar_cimmerse_rawN = "invar_cimmerse_raw"
val unCons_cimmerse_rawN = "unCons_cimmerse_raw"
fun mk_map_flat_rawN i = "map_flat" ^# i ^ "_raw";
fun mk_set_flat_rawN j i = "set" ^# j ^ "_flat" ^# i ^ "_raw";
fun mk_rel_flat_rawN i = "rel_flat" ^# i ^ "_raw";

val dtor_ctorN = "dtor_ctor";
val ctor_dtorN = "ctor_dtor";
val ctor_injectN = "ctor_inject";
val dtor_injectN = "dtor_inject";

val ctor_mapN = "ctor_map";
val ctor_relN = "ctor_rel";
fun mk_ctor_setN j = "ctor_set" ^# j;
val dtor_mapN = "dtor_map";
val dtor_relN = "dtor_rel";
fun mk_dtor_setN j = "dtor_set" ^# j;

fun note_nu_fp_result_thms fact_policy qualify0 b fp nu_fp_res lthy =
  let
    val qualify =
      let val qs = Binding.path_of b;
      in fold_rev (fn (s, mand) => Binding.qualify mand s) qs #> qualify0 end;

    fun note_if_note_all (noted0, lthy0) =
      let
        val shape_res = #shape_res (nu_fp_res:nu_fp_result):shape_result;
        val fp_res = #fp_res (nu_fp_res:nu_fp_result);
        fun name_thms mk_name thmss = map mk_name (1 upto (length thmss)) ~~ thmss;
        val notes =
        (* shape *)
          [(invar_shape_simpsN, #invar_shape_simps shape_res),
           (invar_shape_depth_iffN, #invar_shape_depth_iff_thms shape_res),
           (invar_map_shape_closedN, #invar_map_shape_closed_thms shape_res),
           (cimmerse_shape_simpsN, #cimmerse_shape_simpss shape_res |> flat),
           (cimmerse_shape_injN, #cimmerse_shape_inj_thms shape_res)] @
          name_thms mk_flat_shape_simpsN (#flat_shape_simpss shape_res) @
          name_thms mk_unflat_shape_simpsN (#unflat_shape_simpss shape_res) @
          name_thms mk_invar_shape_flat_shapeN (#invar_shape_flat_shape_thmss shape_res) @
          name_thms mk_invar_shape_unflat_shapeN (#invar_shape_unflat_shape_thmss shape_res) @
          name_thms mk_flat_shape_unflat_shapeN (#flat_shape_unflat_shape_thmss shape_res) @
          name_thms mk_unflat_shape_flat_shapeN (#unflat_shape_flat_shape_thmss shape_res) @
          name_thms mk_map_flat_shapeN (#map_flat_shape_thmss shape_res) @
          flat (map_index (fn (j, thmss) =>
            name_thms (mk_set_flat_shapeN j) thmss) (#set_flat_shape_thmsss shape_res)) @
          name_thms mk_rel_flat_shapeN (#rel_flat_shape_thmss shape_res) @

        (* raw *)
          [(invar_raw_simpsN, #invar_raw_simps nu_fp_res),
           (invar_map_raw_closedN, #invar_map_raw_closed_thms nu_fp_res),
           (cimmerse_raw_simpsN, #cimmerse_raw_simpss nu_fp_res |> flat),
           (invar_cimmerse_rawN, #invar_cimmerse_raw_thms nu_fp_res),
           (unCons_cimmerse_rawN, #unCons_cimmerse_raw_thms nu_fp_res)] @
          name_thms mk_flat_raw_simpsN (#flat_raw_simpss nu_fp_res) @
          name_thms mk_unflat_raw_simpsN (#unflat_raw_simpss nu_fp_res) @
          name_thms mk_invar_raw_flat_rawN (#invar_raw_flat_raw_thmss nu_fp_res) @
          name_thms mk_invar_raw_unflat_rawN (#invar_raw_unflat_raw_thmss nu_fp_res) @
          name_thms mk_flat_raw_unflat_rawN (#flat_raw_unflat_raw_thmss nu_fp_res) @
          name_thms mk_unflat_raw_flat_rawN (#unflat_raw_flat_raw_thmss nu_fp_res) @
          name_thms mk_map_flat_rawN (#map_flat_raw_thmss nu_fp_res) @
          flat (map_index (fn (j, thmss) =>
            name_thms (mk_set_flat_rawN j) thmss) (#set_flat_raw_thmsss nu_fp_res)) @
          name_thms mk_rel_flat_rawN (#rel_flat_raw_thmss nu_fp_res) @

        (* T *)
          [(dtor_ctorN, #dtor_ctors fp_res),
           (ctor_dtorN, #ctor_dtors fp_res),
           (ctor_injectN, #ctor_injects fp_res),
           (dtor_injectN, #dtor_injects fp_res),
           (case_fp fp ctor_mapN dtor_mapN, #xtor_maps fp_res),
           (case_fp fp ctor_relN dtor_relN, #xtor_rels fp_res)] @
          name_thms (case_fp fp mk_ctor_setN mk_dtor_setN) (#xtor_setss fp_res)
          |> map (fn (thmN, thms) =>
            ((qualify (Binding.qualify true (Binding.name_of b) (Binding.name thmN)), []),
             [(thms, [])]));
      in
        Local_Theory.notes notes lthy0 |>> append noted0
      end;
  in
    ([], lthy)
    |> fact_policy = Note_All ? note_if_note_all
    |>> (fn [] => nu_fp_res
        | noted => morph_nu_fp_result (substitute_noted_thm noted) nu_fp_res)
  end;

end;
